-------------------------------------------------------------------------------
-- sensor-ddl.sql : tables and synthetic data for a moving object database
-------------------------------------------------------------------------------

DROP FUNCTION IF EXISTS sensor_datagen(integer, integer);

CREATE OR REPLACE FUNCTION sensor_datagen(integer, integer)
RETURNS void AS $$
DECLARE
  scnt ALIAS FOR $1; -- number of sensors
  rcnt ALIAS FOR $2; -- number of tuples per sensor
  ts_0 time := localtime;
  x_0 integer;
  y_0 integer;
BEGIN

  -- for each sensor
  FOR i IN 1..scnt LOOP

    -- pick a random start location
    x_0 := runif(300,400)::integer;
    y_0 := runif(100,500)::integer;

    -- do a random walk from there
    INSERT INTO sensor
      SELECT loc.ts, i AS sid, location_pdf(loc)
        , NULL -- joint pdf stored in first dimension
      FROM random_walk( (ts_0, x_0, y_0, 5) , rcnt ) AS loc;

  END LOOP;

END;
$$ LANGUAGE plpgsql VOLATILE STRICT;

	COMMENT ON FUNCTION sensor_datagen(integer, integer)
	IS 'generates synthetic data from multiple sensors';

-------------------------------------------------------------------------------


DROP FUNCTION IF EXISTS loadData();

CREATE FUNCTION loadData() RETURNS void AS $$
BEGIN
	DROP TABLE IF EXISTS room;
	DROP TABLE IF EXISTS sensor;
	-------------------------------------------------------------------------------
	CREATE TABLE room (rid integer PRIMARY KEY, x_min real NOT NULL, x_max real NOT NULL, y_min real NOT NULL, y_max real NOT NULL);

	COMMENT ON TABLE room IS 'locations of rooms in a building';
	COMMENT ON COLUMN room.rid IS 'identifies the room';
	COMMENT ON COLUMN room.x_min IS 'minimum x position of room';
	COMMENT ON COLUMN room.x_max IS 'maximum x position of room';
	COMMENT ON COLUMN room.y_min IS 'minimum y position of room';
	COMMENT ON COLUMN room.y_max IS 'maximum y position of room';

	INSERT INTO room VALUES
		(1, 300,330,  30,120), -- office
		(2, 320,330, 120,500), -- hallway
		(3, 330,410, 120,225), -- room 1
		(4, 330,410, 225,325), -- room 2
		(5, 330,410, 400,530); -- room 3


	-------------------------------------------------------------------------------

	CREATE TABLE sensor
	(
		ts time NOT NULL,
		sid integer NOT NULL,

		-- uncertain location
		xpos ucon,
		ypos ucon,
		DEPENDENT (xpos, ypos)
	);

	COMMENT ON TABLE sensor IS 'uncertain version of location table';
	COMMENT ON COLUMN sensor.ts IS 'time of sensor reading';
	COMMENT ON COLUMN sensor.sid IS 'identifies the sensor';
	COMMENT ON COLUMN sensor.xpos IS 'x position of location';
	COMMENT ON COLUMN sensor.ypos IS 'y position of location';

	-------------------------------------------------------------------------------
	-- INSERT INTO sensor
	PERFORM sensor_datagen(4, 25);

	ALTER TABLE sensor ADD CONSTRAINT sensor_pkey PRIMARY KEY (ts, sid);
	CLUSTER sensor USING sensor_pkey;

	CREATE INDEX sensor_sid_idx ON sensor (sid,ts);


END;
$$ language plpgsql;
