/*
 * errors.c - server error reporting functions
 * 
 * Orion Project, http://orion.cs.purdue.edu/
 * Copyright (c) 2008 Purdue University
 */

#include "postgres.h"
#include "errors.h"

/*---------------------------------------------------------------------------*\
  Utility Functions
\*---------------------------------------------------------------------------*/

/*
 * Creates an error mark string, given the error message, the original input
 * string, and the error position.
 */
static char *errorMark(char *msg, char *str, char *pos)
{
	int32 len = 8 + strlen(msg) + (int32) (pos - str) + 2;
	char *mark = palloc(len);

	memset(mark, ' ', len - 2);
	mark[len-2] = '^';
	mark[len-1] = '\0';
	return mark;
}

/*---------------------------------------------------------------------------*\
  Text Representation
\*---------------------------------------------------------------------------*/

/*
 * Invalid PDF for a given data type.
 */
void err_pdf_type(char *sqlname, char *str)
{
	ereport(ERROR, (errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
					errmsg("invalid pdf type for %s: %s",
							sqlname, str)));
}

/*
 * Incorrect I/O syntax for a given PDF.
 */
void err_pdf_text(const char *pdfname, char *str, char *pos)
{
	char msg[50];
	sprintf(msg, "invalid pdf syntax for %s: ", pdfname);

	ereport(ERROR, (errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
					errmsg("%s%s\n%s",
							msg, str, errorMark(msg, str, pos))));
}

/*
 * Missing delimeter in PDF string.
 */
void err_pdf_delim(char dlm, char *str) {
	ereport(ERROR, (errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
					errmsg("missing '%c' delimiter in pdf: %s",
							dlm, str)));
}

/*
 * Expected NULL terminator when paring a PDF.
 */
void err_pdf_extra(char *str, char *pos) {
	char msg[50];
	sprintf(msg, "unexpected characters in pdf: ");

	ereport(ERROR, (errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
					errmsg("%s%s\n%s",
							msg, str, errorMark(msg, str, pos))));
}

/*
 * Incorrect pdf header (e.g. histogram: minv, maxv, step).
 */
void err_pdf_hdr(char *valname, char *str)
{
	ereport(ERROR, (errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
					errmsg("incorrect number of %s: %s", valname, str)));
}

/*---------------------------------------------------------------------------*\
  Argument Validation
\*---------------------------------------------------------------------------*/

/*
 * Invalid argument for a public function.
 */
void err_arg_value(char *funcname, char *valname, int32 value)
{
	ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
					errmsg("%s - invalid argument %s: '%c' (%d)",
							funcname, valname, (char) value, value)));
}

/*
 * Invalid count for variable-length data.
 */
void err_arg_limit(char *funcname, char *valname, int32 limit, int32 value)
{
	ereport(ERROR, (errcode(ERRCODE_PROGRAM_LIMIT_EXCEEDED),
					errmsg("%s - number of %s (%d) must be between 1 and %d",
							funcname, valname, value, limit)));
}

/*---------------------------------------------------------------------------*\
  Internal Errors (Bugs)
\*---------------------------------------------------------------------------*/

/*
 * Unknown (default) case in switch statements.
 */
void err_bad_case(char *funcname, char *valname, int32 value)
{
	ereport(ERROR, (errcode(ERRCODE_DATA_CORRUPTED),
					errmsg_internal("%s - unknown %s in switch: '%c' (%d)",
							funcname, valname, (char) value, value)));
}

/*
 * Place holder for incremental type and pdf development.
 */
void err_not_impl(char *funcname, char *valname, int32 value)
{
	ereport(ERROR, (errcode(ERRCODE_INTERNAL_ERROR),
					errmsg_internal("%s not implemented for %s: '%c' (%d)",
							funcname, valname, (char) value, value)));
}

/*
 * Raises an internal error with the given message and hint.
 */
void err_custom(char *message, char *hint)
{
	ereport(ERROR, (errcode(ERRCODE_INTERNAL_ERROR),
					errmsg_internal("%s", message), errhint("%s", hint)));
}
