--
-- orion.sql - installs Orion into a database
--
-- All objects are created in the orion schema.
-- To uninstall: (also drops uncertain columns)
--    DROP SCHEMA IF EXISTS orion CASCADE;
-- 
-- Orion Project, http://orion.cs.purdue.edu/
-- Copyright (c) 2008 Purdue University
--

CREATE SCHEMA orion;
COMMENT ON SCHEMA orion IS 'uncertain DBMS extensions';

SET search_path TO orion;
SET client_min_messages = warning;

-------------------------------------------------------------------------------
-- System Catalogs
-- (NOTE: keep table names in sync with orion_init function in orion_bki.c)
-------------------------------------------------------------------------------

CREATE TABLE orion_depend
(
	-- table identifier (pg_class.oid)
	relid oid NOT NULL,

	-- column number (pg_attribute.attnum)
	attnum smallint NOT NULL,

	-- dimension number (within the pdf)
	dimnum smallint NOT NULL,

	-- unique dependency set id
	depid integer NOT NULL,

	PRIMARY KEY (relid, attnum)
);

	COMMENT ON TABLE orion_depend
	IS 'dependency sets for joint pdfs';

CREATE INDEX orion_depend_index ON orion_depend(depid);

CREATE SEQUENCE orion_depend_seq;

	COMMENT ON SEQUENCE orion_depend_seq
	IS 'generates unique depids for orion_depend';

CREATE SEQUENCE history_id_seq;

	COMMENT ON SEQUENCE history_id_seq
	IS 'generates unique ids for history';

CREATE TABLE orion_catmap
(
	-- category identifier
	catid serial PRIMARY KEY,

	-- category display name
	catname name NOT NULL UNIQUE
);

	COMMENT ON TABLE orion_catmap
	IS 'maps category ids to labels';

	COMMENT ON SEQUENCE orion_catmap_catid_seq
	IS 'generates unique ids for category names';

-------------------------------------------------------------------------------
-- Uncertain Type
-------------------------------------------------------------------------------

CREATE FUNCTION uncertain_in(cstring) RETURNS uncertain
AS '' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION uncertain_in(cstring)
	IS 'throws error; use uncertain_parse instead';

CREATE FUNCTION uncertain_out(uncertain) RETURNS cstring
AS '' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION uncertain_out(uncertain)
	IS 'output function for uncertain data types';

CREATE TYPE uncertain
(
	INPUT = uncertain_in,
	OUTPUT = uncertain_out,
	STORAGE = extended
);

	COMMENT ON TYPE uncertain
	IS 'arbitrary joint probability distribution';

CREATE FUNCTION uncertain_parse(cstring, cstring) RETURNS uncertain
AS '' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION uncertain_parse(cstring, cstring)
	IS 'input function for uncertain data types';

CREATE FUNCTION uncertain_print(uncertain) RETURNS cstring
AS '' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION uncertain_print(uncertain)
	IS 'outputs the internal uncertain structure';

-------------------------------------------------------------------------------
-- Pseudo-Types
-------------------------------------------------------------------------------

CREATE FUNCTION ucon_in(cstring) RETURNS ucon
AS 'uncertain_in' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION ucon_in(cstring)
	IS 'pseudo-type input; calls uncertain_in';

CREATE FUNCTION ucon_out(ucon) RETURNS cstring
AS 'uncertain_out' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION ucon_out(ucon)
	IS 'pseudo-type output; calls uncertain_out';

CREATE TYPE ucon
(
  INPUT = ucon_in,
  OUTPUT = ucon_out,
  STORAGE = extended
);

	COMMENT ON TYPE ucon
	IS 'uncertain continuous numeric';

CREATE CAST (ucon AS uncertain)
WITHOUT FUNCTION AS IMPLICIT;

CREATE CAST (uncertain AS ucon)
WITHOUT FUNCTION AS IMPLICIT;


CREATE FUNCTION udis_in(cstring) RETURNS udis
AS 'uncertain_in' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION udis_in(cstring)
	IS 'pseudo-type input; calls uncertain_in';

CREATE FUNCTION udis_out(udis) RETURNS cstring
AS 'uncertain_out' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION udis_out(udis)
	IS 'pseudo-type output; calls uncertain_out';

CREATE TYPE udis
(
  INPUT = udis_in,
  OUTPUT = udis_out,
  STORAGE = extended
);

	COMMENT ON TYPE udis
	IS 'uncertain discrete numeric';

CREATE CAST (udis AS uncertain)
WITHOUT FUNCTION AS IMPLICIT;

CREATE CAST (uncertain AS udis)
WITHOUT FUNCTION AS IMPLICIT;


CREATE FUNCTION uord_in(cstring) RETURNS uord
AS 'uncertain_in' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION uord_in(cstring)
	IS 'pseudo-type input; calls uncertain_in';

CREATE FUNCTION uord_out(uord) RETURNS cstring
AS 'uncertain_out' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION uord_out(uord)
	IS 'pseudo-type output; calls uncertain_out';

CREATE TYPE uord
(
  INPUT = uord_in,
  OUTPUT = uord_out,
  STORAGE = extended
);

	COMMENT ON TYPE uord
	IS 'uncertain ordered categorical';

CREATE CAST (uord AS uncertain)
WITHOUT FUNCTION AS IMPLICIT;

CREATE CAST (uncertain AS uord)
WITHOUT FUNCTION AS IMPLICIT;


CREATE FUNCTION unom_in(cstring) RETURNS unom
AS 'uncertain_in' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION unom_in(cstring)
	IS 'pseudo-type input; calls uncertain_in';

CREATE FUNCTION unom_out(unom) RETURNS cstring
AS 'uncertain_out' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION unom_out(unom)
	IS 'pseudo-type output; calls uncertain_out';

CREATE TYPE unom
(
  INPUT = unom_in,
  OUTPUT = unom_out,
  STORAGE = extended
);

	COMMENT ON TYPE unom
	IS 'uncertain unordered categorical';

CREATE CAST (unom AS uncertain)
WITHOUT FUNCTION AS IMPLICIT;

CREATE CAST (uncertain AS unom)
WITHOUT FUNCTION AS IMPLICIT;

-------------------------------------------------------------------------------
-- Inquiry Functions
-------------------------------------------------------------------------------

CREATE FUNCTION u_lower(uncertain, integer) RETURNS real
AS '' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION u_lower(uncertain, integer)
	IS 'lower bound of a pdf dimension';

CREATE FUNCTION u_upper(uncertain, integer) RETURNS real
AS '' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION u_upper(uncertain, integer)
	IS 'upper bound of a pdf dimension';

CREATE FUNCTION u_pdf(uncertain, real[]) RETURNS real
AS '' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION u_pdf(uncertain, real[])
	IS 'pdf value at a given point';

CREATE FUNCTION u_cdf(uncertain, real[]) RETURNS real
AS '' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION u_cdf(uncertain, real[])
	IS 'cdf value at a given point';

CREATE FUNCTION u_prob(uncertain, real[], real[]) RETURNS real
AS '' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION u_prob(uncertain, real[], real[])
	IS 'probability within a region';

CREATE FUNCTION u_mass(uncertain) RETURNS real
AS '' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION u_mass(uncertain)
	IS 'total probability mass (cdf at highest point)';

CREATE FUNCTION u_nonzero(uncertain) RETURNS bool
AS $$ SELECT (u_mass($1) > 0); $$ LANGUAGE SQL IMMUTABLE STRICT;

	COMMENT ON FUNCTION u_nonzero(uncertain)
	IS 'true if total prob of the uncertain is nonzero';

-------------------------------------------------------------------------------
-- Sampling Functions
-------------------------------------------------------------------------------

CREATE FUNCTION u_real2hist(real) RETURNS uncertain
AS '' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION u_real2hist(real)
	IS 'constructs a pdf over the given value';

CREATE FUNCTION u_sample(uncertain, integer) RETURNS uncertain
AS '' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION u_sample(uncertain, integer)
	IS 'approximates or enumerates a distribution';

CREATE FUNCTION u_reduce(uncertain, integer) RETURNS uncertain
AS '' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION u_reduce(uncertain, integer)
	IS 'evaluates products/floors -> canonical multigram';

-------------------------------------------------------------------------------
-- PDF Operations
-------------------------------------------------------------------------------

CREATE FUNCTION u_floor(uncertain, float4[][]) RETURNS uncertain
AS '' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION u_floor(uncertain, float4[][])
	IS 'floors a specified region of a distribution';

CREATE FUNCTION u_floor_ex(uncertain, int2[], float4[][]) RETURNS uncertain
AS '' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION u_floor_ex(uncertain, int2[], float4[][])
	IS 'floors a specified region of a distribution';

CREATE FUNCTION u_margin(uncertain, int2[]) RETURNS uncertain
AS '' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION u_margin(uncertain, int2[])
	IS 'marginalizes a distribution over a subset of attributes';

CREATE FUNCTION u_product(uncertain, uncertain) RETURNS uncertain
AS '' LANGUAGE internal IMMUTABLE STRICT;

	COMMENT ON FUNCTION u_product(uncertain, uncertain)
	IS 'constructs the joint distribution of the given pdfs';

-------------------------------------------------------------------------------
-- Aggregate functions
-- (NOTE: These do not have possible worlds semantics)
-------------------------------------------------------------------------------

CREATE FUNCTION uncertain_min_sfunc (float4, uncertain) RETURNS float4 AS $$
  SELECT float4smaller(COALESCE($1, u_lower($2,0)), u_lower($2,0));
$$ LANGUAGE SQL;

CREATE AGGREGATE min (uncertain) 
(
  sfunc = uncertain_min_sfunc,
  stype = float4
);

CREATE FUNCTION uncertain_max_sfunc (float4, uncertain) RETURNS float4 AS $$
  SELECT float4larger(COALESCE($1, u_upper($2,0)), u_upper($2,0));
$$ LANGUAGE SQL;

CREATE AGGREGATE max (uncertain) 
(
  sfunc = uncertain_max_sfunc,
  stype = float4
);

CREATE FUNCTION uncertain_avg_sfunc (float8[], uncertain) RETURNS float8[] AS $$
  SELECT  CAST(ARRAY[($1)[1] + (u_lower($2,0) + u_upper($2,0))/2, ($1)[2] + 1]
          AS float8[]);
$$ LANGUAGE SQL STRICT;

CREATE FUNCTION uncertain_avg_ffunc(float8[]) RETURNS float8 AS $$
  SELECT CASE WHEN ($1)[2]=0 THEN NULL ELSE (($1)[1])/(($1)[2]) END;
$$ LANGUAGE SQL STRICT;

CREATE AGGREGATE avg(uncertain)
(
  sfunc = uncertain_avg_sfunc,
  finalfunc = uncertain_avg_ffunc,
  stype = float8[],
  initcond = '{0,0}'
);

CREATE FUNCTION uncertain_sum_sfunc(float8, uncertain) RETURNS float8 AS $$
  SELECT COALESCE($1,0.) + (u_lower($2,0) + u_upper($2,0))/2;
$$ LANGUAGE SQL;

CREATE AGGREGATE sum(uncertain)
(
  sfunc = uncertain_sum_sfunc,
  stype = float8
);

-------------------------------------------------------------------------------
-- 1D Operators (mapped to pdf operations via query rewriting)
-------------------------------------------------------------------------------

CREATE FUNCTION _u_lt_uc(uncertain, float4) RETURNS bool
AS $$ SELECT TRUE; $$ LANGUAGE SQL IMMUTABLE STRICT;

	COMMENT ON FUNCTION _u_lt_uc(uncertain, float4)
	IS 'operator function that always returns true';

CREATE FUNCTION _u_lt_cu(float4, uncertain) RETURNS bool
AS $$ SELECT TRUE; $$ LANGUAGE SQL IMMUTABLE STRICT;

	COMMENT ON FUNCTION _u_lt_cu(float4, uncertain)
	IS 'operator function that always returns true';

CREATE FUNCTION _u_gt_uc(uncertain, float4) RETURNS bool
AS $$ SELECT TRUE; $$ LANGUAGE SQL IMMUTABLE STRICT;

	COMMENT ON FUNCTION _u_gt_uc(uncertain, float4)
	IS 'operator function that always returns true';

CREATE FUNCTION _u_gt_cu(float4, uncertain) RETURNS bool
AS $$ SELECT TRUE; $$ LANGUAGE SQL IMMUTABLE STRICT;

	COMMENT ON FUNCTION _u_gt_cu(float4, uncertain)
	IS 'operator function that always returns true';

CREATE OPERATOR <
(
	LEFTARG = uncertain,
	RIGHTARG = float4,
	PROCEDURE = _u_lt_uc,
	COMMUTATOR = >=
);

	COMMENT ON OPERATOR < (uncertain, float4)
	IS 'pdf < val with nonzero probability';

CREATE OPERATOR <
(
	LEFTARG = float4,
	RIGHTARG = uncertain,
	PROCEDURE = _u_lt_cu,
	COMMUTATOR = >=
);

	COMMENT ON OPERATOR < (float4, uncertain)
	IS 'val < pdf with nonzero probability';

CREATE OPERATOR >
(
	LEFTARG = uncertain,
	RIGHTARG = float4,
	PROCEDURE = _u_gt_uc,
	COMMUTATOR = <=
);

	COMMENT ON OPERATOR > (uncertain, float4)
	IS 'pdf > val with nonzero probability';

CREATE OPERATOR >
(
	LEFTARG = float4,
	RIGHTARG = uncertain,
	PROCEDURE = _u_gt_cu,
	COMMUTATOR = <=
);

	COMMENT ON OPERATOR > (float4, uncertain)
	IS 'val > pdf with nonzero probability';

CREATE FUNCTION _u_gt_uu(uncertain, uncertain) RETURNS bool
AS $$ SELECT TRUE; $$ LANGUAGE SQL IMMUTABLE STRICT;

	COMMENT ON FUNCTION _u_gt_uu(uncertain, uncertain)
	IS 'operator function that always returns true';

CREATE OPERATOR >
(
	LEFTARG = uncertain,
	RIGHTARG = uncertain,
	PROCEDURE = _u_gt_uu,
	COMMUTATOR = <=
);

	COMMENT ON OPERATOR > (uncertain, uncertain)
	IS 'pdf > pdf with nonzero probability';

CREATE FUNCTION _u_lt_uu(uncertain, uncertain) RETURNS bool
AS $$ SELECT TRUE; $$ LANGUAGE SQL IMMUTABLE STRICT;

	COMMENT ON FUNCTION _u_lt_uu(uncertain, uncertain)
	IS 'operator function that always returns true';

CREATE OPERATOR <
(
	LEFTARG = uncertain,
	RIGHTARG = uncertain,
	PROCEDURE = _u_lt_uu,
	COMMUTATOR = >=
);

	COMMENT ON OPERATOR < (uncertain, uncertain)
	IS 'pdf < pdf with nonzero probability';

CREATE FUNCTION _u_lte_uu(uncertain, uncertain) RETURNS bool
AS $$ SELECT TRUE; $$ LANGUAGE SQL IMMUTABLE STRICT;

	COMMENT ON FUNCTION _u_lte_uu(uncertain, uncertain)
	IS 'operator function that always returns true';

CREATE OPERATOR <=
(
	LEFTARG = uncertain,
	RIGHTARG = uncertain,
	PROCEDURE = _u_lte_uu,
	COMMUTATOR = >
);

	COMMENT ON OPERATOR <= (uncertain, uncertain)
	IS 'pdf <= pdf with nonzero probability';

CREATE FUNCTION _u_gte_uu(uncertain, uncertain) RETURNS bool
AS $$ SELECT TRUE; $$ LANGUAGE SQL IMMUTABLE STRICT;

	COMMENT ON FUNCTION _u_gte_uu(uncertain, uncertain)
	IS 'operator function that always returns true';

CREATE OPERATOR >=
(
	LEFTARG = uncertain,
	RIGHTARG = uncertain,
	PROCEDURE = _u_gte_uu,
	COMMUTATOR = <
);

	COMMENT ON OPERATOR >= (uncertain, uncertain)
	IS 'pdf >= pdf with nonzero probability';

CREATE FUNCTION _u_lte_uc(uncertain, float4) RETURNS bool
AS $$ SELECT TRUE; $$ LANGUAGE SQL IMMUTABLE STRICT;

	COMMENT ON FUNCTION _u_lte_uc(uncertain, float4)
	IS 'operator function that always returns true';

CREATE OPERATOR <=
(
	LEFTARG = uncertain,
	RIGHTARG = float4,
	PROCEDURE = _u_lte_uc,
	COMMUTATOR = >
);

	COMMENT ON OPERATOR <= (uncertain, float4)
	IS 'pdf <= constant with nonzero probability';

CREATE FUNCTION _u_gte_uc(uncertain, float4) RETURNS bool
AS $$ SELECT TRUE; $$ LANGUAGE SQL IMMUTABLE STRICT;

	COMMENT ON FUNCTION _u_gte_uc(uncertain, float4)
	IS 'operator function that always returns true';

CREATE OPERATOR >=
(
	LEFTARG = uncertain,
	RIGHTARG = float4,
	PROCEDURE = _u_gte_uc,
	COMMUTATOR = <
);

	COMMENT ON OPERATOR >= (uncertain, float4)
	IS 'pdf >= constant with nonzero probability';

CREATE FUNCTION _u_eq_uc(uncertain, float4) RETURNS bool
AS $$ SELECT TRUE; $$ LANGUAGE SQL IMMUTABLE STRICT;

	COMMENT ON FUNCTION _u_eq_uc(uncertain, float4)
	IS 'operator function that always returns true';

CREATE OPERATOR =
(
	LEFTARG = uncertain,
	RIGHTARG = float4,
	PROCEDURE = _u_eq_uc,
	COMMUTATOR = !=
);

	COMMENT ON OPERATOR = (uncertain, float4)
	IS 'pdf = constant with nonzero probability';

CREATE FUNCTION _u_neq_uc(uncertain, float4) RETURNS bool
AS $$ SELECT TRUE; $$ LANGUAGE SQL IMMUTABLE STRICT;

	COMMENT ON FUNCTION _u_neq_uc(uncertain, float4)
	IS 'operator function that always returns true';

CREATE OPERATOR !=
(
	LEFTARG = uncertain,
	RIGHTARG = float4,
	PROCEDURE = _u_neq_uc,
	COMMUTATOR = =
);

	COMMENT ON OPERATOR != (uncertain, float4)
	IS 'pdf != constant with nonzero probability';


CREATE FUNCTION _u_eq_uu(uncertain, uncertain) RETURNS bool
AS $$ SELECT TRUE; $$ LANGUAGE SQL IMMUTABLE STRICT;

	COMMENT ON FUNCTION _u_eq_uu(uncertain, uncertain)
	IS 'operator function that always returns true';

CREATE OPERATOR =
(
	LEFTARG = uncertain,
	RIGHTARG = uncertain,
	PROCEDURE = _u_eq_uu,
	COMMUTATOR = !=
);

	COMMENT ON OPERATOR = (uncertain, uncertain)
	IS 'pdf = pdf with nonzero probability';

CREATE FUNCTION _u_neq_uu(uncertain, uncertain) RETURNS bool
AS $$ SELECT TRUE; $$ LANGUAGE SQL IMMUTABLE STRICT;

	COMMENT ON FUNCTION _u_neq_uu(uncertain, uncertain)
	IS 'operator function that always returns true';

CREATE OPERATOR !=
(
	LEFTARG = uncertain,
	RIGHTARG = uncertain,
	PROCEDURE = _u_neq_uu,
	COMMUTATOR = =
);

	COMMENT ON OPERATOR != (uncertain, uncertain)
	IS 'pdf != pdf with nonzero probability';

-------------------------------------------------------------------------------

CREATE TABLE orion_present AS
  SELECT true AS present;

	COMMENT ON TABLE orion_present
	IS 'is Orion installed in this database?';

RESET search_path;
RESET client_min_messages;
