/*
 * binomial.c - 1D binomial distribution
 * 
 * Orion Project, http://orion.cs.purdue.edu/
 * Copyright (c) 2008 Purdue University
 */

#include "postgres.h"
#include "pdf.h"
#include "pdfs_io.h"
#include "uncertain.h"
#include "errors.h"
#include <math.h>

/*---------------------------------------------------------------------------*\
  Prototypes & VTable
\*---------------------------------------------------------------------------*/

typedef struct Binomial {
  PDF     pdf;
  int32   n;
  float4  p;
} Binomial;

#define TO_BINOMIAL(x) Binomial *x = (Binomial *) _ ## x

static float4 binomial_lower(PDF*, int);
static float4 binomial_upper(PDF*, int);
static float4 binomial_pdf(PDF*, float4*);
static float4 binomial_cdf(PDF*, float4*);
static float4 binomial_mass(PDF*);
static PDF*   binomial_sample(PDF*, int n);
static PDF*   binomial_parse(char*, char*);
static char*  binomial_print(PDF*);

#define PDF_NAME "binom"

const PDF_vtable Binomial_vtbl = {
  .name = PDF_NAME,
  .id = 'b',
  
  .lower = binomial_lower,
  .upper = binomial_upper,
  .pdf = binomial_pdf,
  .cdf = binomial_cdf,
  .mass = binomial_mass,
  .sample = binomial_sample,

  .parse = binomial_parse,
  .print = binomial_print
};


static float4 binomial_lower(PDF* _pdf, int d) {
  return 0;
}

static float4 binomial_upper(PDF* _pdf, int d) {
  TO_BINOMIAL(pdf);
  return pdf->n;
}

/* n = # trials, k = # successes, p = probability of one success */
static float4 real_pdf(float4 p, int32 n, int32 k) {
  float4 num=1., denom=1.;
  int32 i;

  for(i=2; i <= n-k; ++i) /* Compute (n-r)! */
    denom *= i;

  for(i=k+1; i <= n; ++i) /* Compute n!/k! */
    num *= i;

  return num/denom*pow(p,k)*pow(1-p,n-k);
}

static float4 binomial_pdf(PDF* _pdf, float4 *_x) {
  TO_BINOMIAL(pdf);
  float4 x=floor(*_x);

  if(fabsf(*_x-x) >= TOLERANCE)
    return 0;
  else
    return real_pdf(pdf->p, pdf->n, (int) x);
}

static float4 binomial_cdf(PDF *_pdf, float4 *x) {
  TO_BINOMIAL(pdf);
  float4 p = 0.;
  int i;

  for(i=0; i <= *x+TOLERANCE; ++i) {
    p += real_pdf(pdf->p, pdf->n, i);
  }

  return p;
}

static float4 binomial_mass(PDF* pdf) {
	return 1.0;
}

static PDF* binomial_sample(PDF *_pdf, int n) {
  TO_BINOMIAL(pdf);
  Histogram *hist;
  float4 *bins;
  int i;
  HistHead hdr = {
    .minv = 0,
    .maxv = pdf->n,
    .count = pdf->n+1,				/* Ignore suggested resolution */
    .type = DIM_UDIS
  };

  hist = histogram_init(1, &hdr, NULL);
  bins = BINS_PTR(hist);

  for(i=0; i <= pdf->n; ++i)	/* Setup index */
    *bins++ = i;

  bins[0] = pow(1-pdf->p,pdf->n);
  for(i=1; i <= pdf->n; ++i)  /* Compute PDF; last two terms map nC(i-1)->nCi */
    bins[i] = bins[i-1] * pdf->p / (1-pdf->p) * (pdf->n-i+1) / i;

//  for(i=1; i <= pdf->n; ++i)	/* Convert to CDF */
//    bins[i] += bins[i-1];

  return (PDF*) hist;
}

/*---------------------------------------------------------------------------*\
  I/O
\*---------------------------------------------------------------------------*/

static PDF* binomial_parse(char *str, char *dimtypes) {
	Binomial *pdf;
	char *pos = str;

	/* set the dimension type */
	dimtypes[0] = DIM_UDIS;

	/* allocate and initialize the pdf */
	pdf = palloc(sizeof(Binomial));
	pdf->pdf.vtbl = &Binomial_vtbl;
	pdf->pdf.ndim = 1;
	pdf->pdf.size = sizeof(Binomial);

	/* parse the input string */
	skipSpace(pos);
	pos = scanInt32(&pdf->n, pos, PDF_NAME, str);
	skipDelim(pos, ',')
	pos = scanFloat4(&pdf->p, pos, PDF_NAME, str);
	checkExtra(pos, str);

	if(pdf->p >= 0 && pdf->p <= 1 && pdf->n > 0)
		return (PDF*) pdf;
	else
		return NULL;
}

static char* binomial_print(PDF* _pdf) {
	char *str = palloc(28);
	TO_BINOMIAL(pdf);
	sprintf(str, "%d , %g", pdf->n, pdf->p);
	return str;
}
