/*
 * delta.c - distribution over a single point
 * 
 * Orion Project, http://orion.cs.purdue.edu/
 * Copyright (c) 2008 Purdue University
 */

#include "postgres.h"
#include "pdf.h"
#include "pdfs_io.h"
#include "uncertain.h"
#include "errors.h"
#include <math.h>

/*---------------------------------------------------------------------------*\
  Prototypes & VTable
\*---------------------------------------------------------------------------*/

typedef struct Delta {
  PDF     pdf;
  float4  value;
} Delta;

#define TO_DELTA(x) Delta *x = (Delta *) _ ## x

static float4 delta_value(PDF*, int);
static float4 delta_pdf(PDF*, float4*);
static float4 delta_cdf(PDF*, float4*);
static float4 delta_mass(PDF*);
static PDF*   delta_sample(PDF*, int n);
static PDF*   delta_parse(char*, char*);
static char*  delta_print(PDF*);

#define PDF_NAME "delta"

const PDF_vtable Delta_vtbl = {
  .name = PDF_NAME,
  .id = '^',
  
  .lower = delta_value,
  .upper = delta_value,
  .pdf = delta_pdf,
  .cdf = delta_cdf,
  .mass = delta_mass,
  .sample = delta_sample,

  .parse = delta_parse,
  .print = delta_print,
};

PDF*   delta_init(float4 x) {
  Delta *pdf = palloc(sizeof(Delta));

  pdf->pdf.vtbl = &Delta_vtbl;
  pdf->pdf.ndim = 1;
  pdf->pdf.size = sizeof(Delta);
	pdf->value = x;

  return (PDF *) pdf;
}


static float4 delta_value(PDF *pdf, int dim) {
  return ((Delta*) pdf)->value;
}

static float4 delta_pdf(PDF *pdf, float4 *v) {
  return fabsf(((Delta*) pdf)->value-*v) < TOLERANCE;
}

static float4 delta_cdf(PDF *pdf, float4 *v) {
  return ((Delta*) pdf)->value <= *v;
}

static float4 delta_mass(PDF* pdf){
	return 1.0;
}

static PDF*   delta_sample(PDF *_pdf, int n) {
  Delta *pdf = (Delta*) _pdf;
  HistHead hdr = {.minv=pdf->value, .maxv=pdf->value, .count=1, .type=DIM_UDIS};
	Histogram *h;
	float4 *bins;

	h = histogram_init(1, &hdr, NULL);
	bins = BINS_PTR(h);
	
	bins[0] = pdf->value; /* Setup index */
	bins[1] = 1.;					/* Fill bin with one */

  return (PDF *) h;
}

/*---------------------------------------------------------------------------*\
  I/O
\*---------------------------------------------------------------------------*/

static PDF* delta_parse(char *str, char *dimtypes) {
	char *pos = str;
	float4 value;

	/* set the dimension type */
	dimtypes[0] = DIM_UDIS;

	/* parse the input string */
	skipSpace(pos);
	pos = scanFloat4(&value, pos, PDF_NAME, str);
	checkExtra(pos, str);

	return delta_init(value);
}

static char* delta_print(PDF *_pdf) {
	char *str = palloc(13);
	TO_DELTA(pdf);
	sprintf(str, "%g", pdf->value);
	return str;
}
