/*
 * histogram.c - generic multidimensional histogram
 *
 * The name histogram is mainly historical: this structure technically isn't
 * a histogram since it can store both continous and discrete data, but the
 * way the data is stored (i.e. binned) shows some resemblance.
 * 
 * Orion Project, http://orion.cs.purdue.edu/
 * Copyright (c) 2008 Purdue University
 */

#include "postgres.h"
#include "pdf.h"
#include "pdfs_io.h"
#include "uncertain.h"
#include "errors.h"
#include "debug.h"
#include <math.h>
#include <float.h>
#include "executor/spi.h"

/*---------------------------------------------------------------------------*\
  Prototypes & VTable
\*---------------------------------------------------------------------------*/

#define TO_HISTOGRAM(x) Histogram *x = (Histogram *) _ ## x

float4 histogram_lower(PDF* _pdf, int dim);
float4 histogram_upper(PDF* _pdf, int dim);
float4 histogram_pdf(PDF* _pdf, float4 *x);
float4 histogram_cdf(PDF* _pdf, float4 *x);
float4 histogram_mass(PDF* _pdf);
PDF*   histogram_sample(PDF* _pdf, int n);
PDF*   histogram_parse(char* str, char*);
PDF*   histogram_parse_d(char* str, char*);
char*  histogram_print(PDF* _pdf);

#define PDF_NAME "hist"

const PDF_vtable Histogram_vtbl = {
  .name = PDF_NAME,
  .id = 'h',

  .lower = histogram_lower,
  .upper = histogram_upper,
  .pdf = histogram_pdf,
  .cdf = histogram_cdf,
  .mass = histogram_mass,
  .sample = histogram_sample,

  .parse = histogram_parse,
  .print = histogram_print,
};

const PDF_vtable Histogramd_vtbl = {
  .name = PDF_NAME "_d",
  .id = 'H',

  .lower = histogram_lower,
  .upper = histogram_upper,
  .pdf = histogram_pdf,
  .cdf = histogram_cdf,
  .mass = histogram_mass,
  .sample = histogram_sample,

  .parse = histogram_parse_d,
  .print = histogram_print,
};

/*---------------------------------------------------------------------------*\
  Helper Functions
\*---------------------------------------------------------------------------*/

/* Declarations for pdf/cdf conversion functions */
static void histogram_to_pdf_representation(Histogram *hist, float4 *bins, const int dim, const int32 strides[], uint8 is_edge[]);
static void histogram_to_cdf_representation(Histogram *hist, float4 *bins, const int dim, const int32 strides[], uint8 is_edge[]);
static void    histogram_strides(const Histogram *hist, int32 strides[]);
static inline bool		 dim_is_discrete(const HistHead *dim);
static inline int		 dim_index_size(const HistHead *dim);
static inline float4  dim_step(const HistHead *dim);
static float4* histogram_real_bin_ptr(const Histogram *hist, float4 *indices[]);
static void    histogram_steps(const Histogram*, float4 steps[]);

/*
 * Constructs a new Histogram, given the headers and optionally the bins.
 */
Histogram *histogram_init(int32 ndim, HistHead *hdrs, float4 *bins) {
	uint32 i, nbin=1, size, di_size=0; /* di_size = discrete index size */
	Histogram *hist;

	/* initialize the PDF header */
	for(i=0; i < ndim; ++i)
		nbin *= hdrs[i].count;
	
	/* compute di_size */
	for(i=0; i < ndim; ++i)
		di_size += dim_index_size(&hdrs[i]);
	
	size = sizeof(PDF) + sizeof(HistHead)*ndim + sizeof(float4)*(nbin+di_size);

	hist = palloc0(size);
	hist->pdf.vtbl = &Histogram_vtbl;
	hist->pdf.ndim = ndim;
	hist->pdf.size = size;

	/* copy the dimensions and bins */
	memcpy(hist->hdrs, hdrs, sizeof(HistHead)*ndim);

	if (bins)
		memcpy(BINS_PTR(hist)+di_size, bins, sizeof(float4)*nbin);
	
	return hist;

} /* histogram_init */

/*
 * Samples the given PDF and returns the resulting Histogram.
 */
Histogram *histogram_from_pdf(PDF *pdf, int32 count) {
	HistHead *hdrs;
	int32 i, j;
	int32 nbin; /* total number of bins */
	Histogram *hist;
	float4 *bins;
	float4 curx_f[pdf->ndim];
	int32  curx_i[pdf->ndim];
	float4 step[pdf->ndim];

	/* initialize the histogram headers */
	hdrs = palloc0(sizeof(HistHead) * pdf->ndim);
	for (i = 0; i < pdf->ndim; i++) {
		hdrs[i].minv = PDF_lower(pdf, i);
		hdrs[i].maxv = PDF_upper(pdf, i);
		hdrs[i].count = count;
		hdrs[i].type = DIM_UCON;
	}

	/* construct the histogram and get the bins */
	hist = histogram_init(pdf->ndim, hdrs, NULL);
	nbin = ipow(count, pdf->ndim);
	bins = BINS_PTR(hist);
	histogram_steps(hist, step);

	/* Assume bin i=0...count-1 holds probability of minv < x < (i+1)*step+minv */

	/* Initialize curx, the current point in space being sampled, Both floating 
	 * point and integer versions are maintained to avoid rounding errors */
	for (i = 0; i < pdf->ndim; ++i) curx_f[i] = hdrs[i].minv+step[i];
	memset(curx_i, 0, sizeof(curx_i));

	
	for (i = 0; i < nbin; i++) {				/* initialize bins by sampling the CDF */
		bins[i] = PDF_cdf(pdf, curx_f);

		j = pdf->ndim;										/* increment curx  */
		do {
			j -= 1;
			curx_i[j] = (curx_i[j]+1) % count;
			curx_f[j] = hdrs[j].minv + (curx_i[j]+1)*step[i];
		}while(curx_i[j] == 0 && j > 0);	/* was there a carry? */
	}

	return hist;
}

/*
 * Calculates the total number of histogram bins.
 */
static int32 histogram_bincnt(Histogram *hist) {
	int32 i;
	int32 count = 1;

	/* loop through each header and multiply the count */
	for (i = 0; i < hist->pdf.ndim; i++) {
		count *= hist->hdrs[i].count;
	}
	return count;
}

/*
 * Applies a floor to a histogram.
 */
static float4* histogram_apply_floor_inner(Histogram *hist, const uint8 npred, 
	float4 *preds[], const uint8 types[], float4 *bins, float4 *indices[], 
	const int dim, float4 preds_im[], const int32 strides[]);


void histogram_apply_floor(Histogram *hist, const uint8 npred, const uint8 types[], float4 *preds[]) {
	int  i;
	int32  strides[hist->pdf.ndim];
	float4 *bins;
	float4 *indices[hist->pdf.ndim];
	float4 preds_im[npred];
//	uint8	 is_edge[hist->pdf.ndim];
	int count = 1;

	bins = histogram_real_bin_ptr(hist, indices);
	histogram_strides(hist, strides);

	for(i=0; i < npred; ++i)
		preds_im[i] = 0.;

	for(i = 0; i < hist->pdf.ndim; i++) 
		count *= hist->hdrs[i].count;
//	DEBUG("count : %d", count);

//	DEBUG("bin count : %d", count);
//	DEBUG("before conversion");
//	for(i = 0; i < count; i++)
//		DEBUG("bin prob: %f", bins[i]);
	
//	histogram_to_pdf_representation(hist, bins, 0, strides, is_edge);
//	DEBUG("after conversion");
//	for(i = 0; i < count; i++)
//		DEBUG("bin prob: %f", bins[i]);
//	DEBUG("hist->pdf.ndim : %d", hist->pdf.ndim);
//	DEBUG("npred : %d", npred);
//	for(i = 0; i < npred; i++)
//		DEBUG("preds[%d][0]: %d, preds[%d][1]: %d", i, preds[i][0], i, preds[i][2]);
	histogram_apply_floor_inner(hist, npred, preds, types, bins, indices, 0, preds_im, strides);
//	for(i = 0; i < count; i++)
//		DEBUG("bin prob: %f", bins[i]);
//	histogram_to_cdf_representation(hist, bins, 0, strides, is_edge);
}

#define recurse(x) {\
			int pred; \
			for(pred=0; pred < npred; ++pred) {\
				my_preds_im[pred] = (x) * preds[pred][dim] + preds_im[pred];\
			} \
			bins = histogram_apply_floor_inner(hist, npred, preds, types, bins, indices, dim+1, my_preds_im, strides); \
}
	
static float4* histogram_apply_floor_inner(Histogram *hist, const uint8 npred, 
	float4 *preds[], const uint8 types[], float4 *bins, float4 *indices[], 
	const int dim, float4 preds_im[], const int32 strides[]) {
	float4 my_preds_im[npred];
	int i;

	if(dim == hist->pdf.ndim) {
//		DEBUG("bins : %f", bins[0]);
		int sat=0;

		for(i=0; i < npred && !sat; ++i)
#define TEST(OP) sat = (preds_im[i] OP preds[i][hist->pdf.ndim]); break;
			switch(types[i]) {
				case U_NEQ: TEST(!=);
				case U_EQ:  TEST(==);
				case U_LT:  TEST(<);
				case U_LTE: TEST(<=);
				case U_GT:  TEST(>);
				case U_GTE: TEST(>=);
			}
#undef TEST

		if(!sat)
			*bins = 0.;								/* If no predicate is satisfied, floor*/
		bins++;

	} else if(!dim_is_discrete(&hist->hdrs[dim])) { /* Continuous dimension */
//		DEBUG("flood dimension is continuous");
		float4 step = dim_step(&hist->hdrs[dim]);

		for(i=0; i < hist->hdrs[dim].count; ++i)
			recurse(hist->hdrs[dim].minv + step * (i+.5));
		
	} else {                       /* Discrete dimension */
//		DEBUG("flood dimension is discrete");
		for(i=0; i < hist->hdrs[dim].count; ++i)
			recurse(indices[dim][i]);
	}

	return bins;
}
#undef recurse


/* 
 * Marginalize histogram along given dimensions 
 */
void histogram_marginalize_inner(Histogram *h, float4 *ibins, int32 i_strides[], float4 *obins,int32 o_strides[], const uint8 *dims, uint8 len, int dim, int32 int_coord[]);

Histogram *histogram_marginalize(Histogram *h, const uint8 *dims, uint8 len) {
	int i;
	Histogram *hnew = NULL;
	HistHead hdrs[len];
	int32 int_coord[h->pdf.ndim];       // Used to detect duplicated dimensions
	int32 i_strides[h->pdf.ndim], o_strides[len];
	float4 *ibins, *obins, *i_indices[h->pdf.ndim], *o_indices[len];
	uint8	 is_edge[h->pdf.ndim];

	/* Before doing anything, see if dims is an identity mapping */
	if(len==h->pdf.ndim) {
		for(i=0; i < len; ++i)
			if(dims[i] != i)
				break;

		if(i == len)
			return h;
	}

	histogram_strides(h, i_strides);
	ibins = histogram_real_bin_ptr(h, i_indices);
	histogram_to_cdf_representation(h, ibins, 0, i_strides, is_edge);

	for(i=0; i < h->pdf.ndim; ++i)      // Initialize int_coord to last bin
		int_coord[i] = h->hdrs[i].count-1;

	for(i=0; i < len; ++i) {            // Reset int_coord for selected dims
		memcpy(&hdrs[i], &h->hdrs[dims[i]], sizeof(HistHead));
		int_coord[dims[i]] = -1;
	}

	for(i=0; i < h->pdf.ndim; ++i)      // Move bins to ignore marginalized dims
		if(int_coord[i] >= 0)
			ibins += i_strides[i] * int_coord[i];

	hnew = histogram_init(len, hdrs, NULL);
	obins = histogram_real_bin_ptr(hnew, o_indices);

	for(i=0; i < len; ++i) {            // Copy discrete indices
		int sz = dim_index_size(&hnew->hdrs[i]);
		if(sz > 0)
			memcpy(o_indices[i], i_indices[dims[i]], sizeof(float4)*sz);
	}

	histogram_strides(hnew, o_strides);
	histogram_marginalize_inner(h, ibins, i_strides, obins, o_strides, dims, len, 0, int_coord);
	histogram_to_pdf_representation(hnew, obins, 0, o_strides, is_edge);
	return hnew;
}

void histogram_marginalize_inner(Histogram *h, float4 *ibins, int32 i_strides[], float4 *obins,int32 o_strides[], const uint8 *dims, uint8 len, int dim, int32 int_coord[]) {
	int idim;

	if(dim==len) {
		*obins = *ibins;
		return;
	}

	idim = dims[dim];

	if(int_coord[idim] == -1) { /* Not duplicated */
		int i;

		for(i=0; i < h->hdrs[idim].count; ++i, ibins += i_strides[idim], obins += o_strides[dim]) {
			int_coord[idim] = i;
			histogram_marginalize_inner(h, ibins, i_strides, obins, o_strides, dims, len, dim+1, int_coord);
		}

		int_coord[idim] = -1;

	} else {									 /* Duplicated */
		float4 *src;
		int i, blocksize = sizeof(float4)*o_strides[dim];

		if(int_coord[idim] > 0) {
			memcpy(obins, obins-o_strides[dim-1],blocksize*int_coord[idim]);
			obins += o_strides[dim]*int_coord[idim];
		}

		src = obins;
		histogram_marginalize_inner(h, ibins, i_strides, obins, o_strides, dims, len, dim+1, int_coord);
		obins += o_strides[dim];

		for(i=int_coord[idim]+1; i < h->hdrs[idim].count; ++i) {
			memcpy(obins, src, blocksize);
			obins += o_strides[dim];
		}
	}
}


/* 
 * Product two histograms out into a single histogram 
 */
void histogram_product_inner(Histogram **in, float4 **bins, int nremain, float4 p);

Histogram *histogram_product(Histogram **in, uint8 nhist) {
	float4 *bins[nhist+1];
	Histogram *out;
	int ndim=0;
	int h, i;

	if(nhist==1)				/* Short-circuit in degenerate case */
		return in[0];
	
	for(h=0; h < nhist; ++h)
		ndim += in[h]->pdf.ndim;

	/* Initialize histogram */ {
		HistHead hdrs[ndim];
		int base = 0;

		/* Copy headers */
		for(h=0; h < nhist; ++h) {
			for(i=0; i < in[h]->pdf.ndim; ++i)
				memcpy(&hdrs[base+i], &in[h]->hdrs[i], sizeof(HistHead));
			base += in[h]->pdf.ndim;
		}

		out = histogram_init(ndim, hdrs, NULL);
	}

	bins[nhist] = BINS_PTR(out);

	/* Copy discrete indices */
	for(h=0; h < nhist; ++h) {
		bins[h] = BINS_PTR(in[h]);
		for(i=0; i < in[h]->pdf.ndim; ++i) {
			int sz = dim_index_size(&in[h]->hdrs[i]);
			if(sz) {
				memcpy(bins[nhist], bins[h], sizeof(float4)*sz);
				bins[nhist] += sz;
				bins[h] += sz;
			}
		}
	}

	histogram_product_inner(in, bins, nhist, 1.);

	return out;
}

void histogram_product_inner(Histogram **in, float4 **bins, int nremain, float4 p) {
	int i, nbin=1;
	float4 *mybins = bins[0];

	for(i=0; i < in[0]->pdf.ndim; ++i)
		nbin *= in[0]->hdrs[i].count;

	if(nremain==1) {	 /* Last histogram */
		for(i=0; i < nbin; ++i)
			*bins[1]++ = *mybins++ * p;
	}
	else  {
		for(i=0; i < nbin; ++i)
			histogram_product_inner(in+1, bins+1, nremain-1, *mybins++ * p);
	}
}


/* 
 * Utility methods 
 */

/* Modify bins so that they represent only the probability in that bin */
static void histogram_to_pdf_representation(Histogram *hist, float4 *bins, const int dim, const int32 strides[], uint8 is_edge[]) {
	int i;

	if(dim==hist->pdf.ndim) {
		int diag=0, sdim=0;

		for(i=0; i < hist->pdf.ndim; ++i)
			if(!is_edge[i]) {
				*bins -= bins[-strides[i]];
				diag += strides[i];
				sdim ++;
			}

		if(sdim > 1)
			*bins += (sdim-1) * bins[-diag];
		return;
	}

	bins += strides[dim]*(hist->hdrs[dim].count-1);										 /* To end */
	for(i=hist->hdrs[dim].count-1; i >= 0; --i, bins -= strides[dim]) {
		is_edge[dim] = (i == 0); /* Don't subtract previous bin in this dimension */
		histogram_to_pdf_representation(hist, bins, dim+1, strides, is_edge);
	}
}

/* Undo the to_pdf transformation */
static void histogram_to_cdf_representation(Histogram *hist, float4 *bins, const int dim, const int32 strides[], uint8 is_edge[]) {
	int i;

	if(dim==hist->pdf.ndim) {
		int diag=0, sdim=0;

		for(i=0; i < hist->pdf.ndim; ++i)
			if(!is_edge[i]) {
				*bins += bins[-strides[i]];
				diag += strides[i];
				sdim ++;
			}

		if(sdim > 1)
			*bins -= (sdim-1) * bins[-diag];
		return;
	}

	for(i=0; i < hist->hdrs[dim].count; ++i, bins += strides[dim]) {
		is_edge[dim] = (i == 0);
		histogram_to_cdf_representation(hist, bins, dim+1, strides, is_edge);
	}
}

/* Return the real bin pointer and optionally, the pointers to discrete indices; if provided, indices must be of length hist->pdf.ndim */
static float4* histogram_real_bin_ptr(const Histogram *hist, float4 *indices[]) {
	int i;
	float4 *bins = BINS_PTR(hist);

	if(indices)
		memset(indices, 0, sizeof(float4*) * hist->pdf.ndim);

	for(i=0; i < hist->pdf.ndim; ++i) {
		if(dim_is_discrete(&hist->hdrs[i])) {
			if(indices)
				indices[i] = bins;

			bins += dim_index_size(&hist->hdrs[i]);
		}
	}

	return bins;
}

static void histogram_strides(const Histogram *hist, int32 strides[]) {
	int i;
	strides[hist->pdf.ndim-1] = 1;
	for(i=hist->pdf.ndim-2; i >= 0; --i)
		strides[i] = strides[i+1]*hist->hdrs[i+1].count;
}

int is_histogram(const PDF *pdf) {
	return pdf->vtbl->id == Histogram_vtbl.id;
}

static bool		 dim_is_discrete(const HistHead *dim) {
	return dim->type != DIM_UCON;
}

static int dim_index_size(const HistHead *dim) {
	return (dim_is_discrete(dim))? dim->count : 0;
}

static void histogram_steps(const Histogram *h, float4 steps[]) {
	int i;

	for(i = 0; i < h->pdf.ndim; ++i)
		steps[i] = dim_step(&h->hdrs[i]);
}

static float4 dim_step(const HistHead *dim) {
//	DEBUG("count is %d", dim->count);
	return (dim->maxv - dim->minv) / dim->count;
}

/*---------------------------------------------------------------------------*\
  API Implementation
\*---------------------------------------------------------------------------*/

float4 histogram_lower(PDF *_pdf, int dim) {
	return ((Histogram *) _pdf)->hdrs[dim].minv;
}

float4 histogram_upper(PDF *_pdf, int dim) {
	return ((Histogram *) _pdf)->hdrs[dim].maxv;
}

float4 histogram_pdf(PDF *_pdf, float4 *x) {
	TO_HISTOGRAM(pdf);
	int i, sdim=0, diag=0;
	float4 scale = 1.0, result=0., *bins;
	float4 *indices[_pdf->ndim];
	int32  strides[_pdf->ndim];
	
	bins = histogram_real_bin_ptr(pdf, indices);
	histogram_strides(pdf, strides);

	for(i=_pdf->ndim-1; i >= 0; --i) {
		int which;

		if(x[i] < pdf->hdrs[i].minv || x[i] > pdf->hdrs[i].maxv)
			return 0;                     /* Coordinate is out of range */

		if(!dim_is_discrete(&pdf->hdrs[i])) {// Find bin for continuous dimensions
			float4 step = dim_step(&pdf->hdrs[i]);
			which = (int) (x[i]-pdf->hdrs[i].minv)/step;
			scale *= step;

		} else {											/* Discrete; search index/update bins */
			for(which=0; which < pdf->hdrs[i].count; ++which)
				if(fabsf(x[i] - indices[i][which]) < TOLERANCE)
					break;

			if(which == pdf->hdrs[i].count)
				return 0.;                /* Value not found */
		}

		bins += which*strides[i];

		if(which==0) strides[i] = 0;
	}

	result = *bins;
	for(i=0; i < pdf->pdf.ndim; ++i) {
		if(strides[i]) {
			result -= bins[-strides[i]];
			diag += strides[i];
			sdim ++;
		}
	}

	if(sdim > 1)
		result += (sdim-1) * bins[-diag];

	return result;
}

float4 histogram_cdf(PDF *_pdf, float4 *x) {
	TO_HISTOGRAM(pdf);
	int stride = 1, index = 0, i;
	float4	*indices[pdf->pdf.ndim];
	float4 *bins;

	bins = histogram_real_bin_ptr(pdf, indices);

	for(i=_pdf->ndim-1; i >= 0; --i) {
		int which;
		HistHead *hdr = &pdf->hdrs[i];

		if(x[i] >= hdr->maxv)				      /* If above max, clip to last bin */
			which = hdr->count-1;
		else if(x[i] <= hdr->minv)        /* If below min value, CDF is zero */
			return 0.;
		else if(!dim_is_discrete(hdr)){	  /* If continuous, compute bin */
			which = (int) ceil((x[i]-hdr->minv)/dim_step(hdr)) - 1;
			if(which < 0) which = 0;				/* If x \in (minv, minv+step), fixup */
		} else {													/* If discrete, search index */
			for(which=0; which < hdr->count-1; ++which)
				if(x[i] < indices[i][which+1])/* Stop when next bin is larger */
					break;
		}

		index += which * stride;
		stride *= hdr->count;			/* Update stride for next dimension */
	}

	return bins[index];
}

float4 histogram_mass(PDF *_pdf) {
//	DEBUG("inside histogram_mass");
//	return 1.0;
	TO_HISTOGRAM(pdf);
	int i, count = 1;
	float4 *bins;
	float4	*indices[pdf->pdf.ndim];
	float4 mass = 0;
	bins = histogram_real_bin_ptr(pdf, indices);
	for(i = 0; i < pdf->pdf.ndim; i++) 
		count *= pdf->hdrs[i].count;
	
//	DEBUG("count : %d, pdf->pdf.ndim : %d", count, pdf->pdf.ndim);
	
	for(i = 0; i < count; i++)
		mass += bins[i];
		
	return mass;
}


PDF *histogram_sample(PDF *_pdf, int n) {
	/* histogram is already a sampled PDF */
	return _pdf;
}

/*---------------------------------------------------------------------------*\
  I/O
\*---------------------------------------------------------------------------*/

/*
 * Gets the category id for the given label; if not found, this automatically
 * inserts it into the catalog.
 */
static float4 get_catid(char* catname) {
	char sql[100];
	char sql2[100];
	int32 ret;
	bool isnull;
	int32 catid = 0;

	/* connect to the DB and execute the query */
	SPI_connect();
	snprintf(sql, 100, "SELECT catid FROM orion_catmap WHERE (catname = '%s' :: name)", catname);
	ret = SPI_exec(sql, 1);

	/* get the query result */
	if ((ret == SPI_OK_SELECT) &&(SPI_processed != 0))
	{
		TupleDesc tupdesc = SPI_tuptable->tupdesc;
		HeapTuple tuple = SPI_tuptable->vals[0];
		catid = DatumGetInt32( SPI_getbinval(tuple, tupdesc, 1, &isnull) );		
	}
	else {
		snprintf(sql2, 100, "INSERT INTO orion_catmap (catname) VALUES ('%s')", catname);

		ret = SPI_exec(sql2, 2);
		ret = SPI_exec(sql, 1);

		/* get the query result */
		if ((ret == SPI_OK_SELECT) &&(SPI_processed > 0))
		{
			TupleDesc tupdesc = SPI_tuptable->tupdesc;
			HeapTuple tuple = SPI_tuptable->vals[0];
			catid = DatumGetInt32( SPI_getbinval(tuple, tupdesc, 1, &isnull) );
		}
	}

	/* close the DB connection */
	SPI_finish();
	return (float4) catid;
}

/*
 * Gets the label for the given catagory id; throws an error if not found.
 */
static char* get_catname(float4 catid) {
	char sql[100];
	int32 ret;
	bool isnull;
	char* label = NULL;

	/* connect to the DB and execute the query */
	SPI_connect();
	snprintf(sql, 100, "SELECT catname FROM orion_catmap WHERE catid = %d", (int32) catid);
	ret = SPI_exec(sql, 1);

	/* get the query result */
	if ((ret == SPI_OK_SELECT) &&(SPI_processed > 0))
	{
		TupleDesc tupdesc = SPI_tuptable->tupdesc;
		HeapTuple tuple = SPI_tuptable->vals[0];
		label = (char *) SPI_getbinval(tuple, tupdesc, 1, &isnull);
//		printf("catname found : %s\n", label);
//		label = DatumGetCString( SPI_getbinval(tuple, tupdesc, 1, &isnull) );
	}
	else {
		ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE), errmsg("category id :%d doesn't exist", (int32) catid)));
	}

	/* close the DB connection */
	SPI_finish();
	return label;
}

PDF *histogram_parse(char *str, char *dimtypes) {
	char *pos;
	int32 i;
	int32 size;
	Histogram *hist;
	float4 *bins;
	char name[64];
	int32 dims = 0;
	int32 vals = 1;
	int32 nbin = 1;

	/* count delimiters to calculate the expected size */
	for (pos = str; *pos != '\0' && *pos != ':'; pos++) {
		switch(*pos) {
		case '{': dims++; vals += 4; break;  /* udis header adds four extra vals */
		case '[': dims++; vals ++; break;    /* ucon header adds an extra val */
		case ',': vals++; break;
		}
	}

	/* count commas after the colon to get the number of bins */
	for (; *pos != '\0'; pos++) {
		if (*pos == ',') {
			nbin++;
		}
	}
	size = sizeof(PDF) + (vals + nbin) * sizeof(float4);

	/* allocate and initialize the pdf */
	hist = palloc0(size);
	hist->pdf.vtbl = &Histogram_vtbl;
	hist->pdf.ndim = dims;
	hist->pdf.size = size;
	bins = BINS_PTR(hist);

	/* scan the histogram headers */
	pos = str;
	skipSpace(pos);
	for (i = 0; i < dims; i++) {
		HistHead *hdr = &hist->hdrs[i];

		pos = strpbrk(pos, "{[");
		if(!pos)										 /* Failed to find header, abort */
			return NULL;

		hdr->type = (*pos=='{')? DIM_UDIS : DIM_UCON;

		if (hdr->type == DIM_UCON) {

			skipDelim(pos, '[');
			pos = scanFloat4(&hdr->minv, pos, PDF_NAME, str);
			skipDelim(pos, ',');
			pos = scanFloat4(&hdr->maxv, pos, PDF_NAME, str);
			skipDelim(pos, ',');
			pos = scanInt32((int32*) &hdr->count, pos, PDF_NAME, str);
			skipDelim(pos, ']');

		} else { /* discrete or categorical */

			hdr->minv = FLT_MAX;
			hdr->maxv = FLT_MIN;
			hdr->count = 0;

			switch(pos[-1]) {
			case 'n': hdr->type = DIM_UNOM; break;
			case 'o': hdr->type = DIM_UORD; break;
			}

			skipDelim(pos, '{');
			while (*pos != '\0' && *pos != '}') {
				if (hdr->type == DIM_UDIS) {
					pos = scanFloat4(bins, pos, PDF_NAME, str);
				} else {
					pos = scanString(name, pos, PDF_NAME, str);
					*bins = get_catid(name);
				}
				skipDelim(pos, ',');
				if (*pos < hdr->minv) {
					hdr->minv = *bins;
				}
				if (*pos > hdr->maxv) {
					hdr->maxv = *bins;
				}
				hdr->count++;
				bins++;
			}

			skipDelim(pos, '}');
		} /* else */

		dimtypes[i] = hdr->type;
	}

	/* validate the number of bins */
	if (nbin != histogram_bincnt(hist)) {
		err_pdf_hdr(ERR_BINCNT, str);
	}

	/* scan the bins and the tail */
	skipDelim(pos, ':');
	for (i = 0; i < nbin; i++) {
		pos = scanFloat4(&bins[i], pos, PDF_NAME, str);
		skipDelim(pos, ',');
	}
	checkExtra(pos, str);

	return (PDF *) hist;
} /* histogram_parse_c(df) */

/* This version expects probability densities */
PDF *histogram_parse_d(char *str, char *dimtypes) {
//	DEBUG("inside histogram_parse_d");
	Histogram *hist;
	hist = (Histogram *) histogram_parse(str, dimtypes);
/*	{
		uint8 is_edge[hist->pdf.ndim];
		int32 strides[hist->pdf.ndim];
		float4 *bins = histogram_real_bin_ptr(hist, NULL);

		histogram_strides(hist, strides);
		histogram_to_cdf_representation(hist, bins, 0, strides, is_edge);
	}
*/
	return (PDF*) hist;
} /* histogram_parse_p(df) */

char *histogram_print(PDF *_pdf) {
	int32 nums;
	int32 size;
	char *str;
	char *pos;
	int32 i, j;
	float4 *bins;
	TO_HISTOGRAM(pdf);

	/* output buffer: 64/value + 3/num_delims + 4/dimen_delims */
	nums = (_pdf->size - sizeof(PDF)) / sizeof(float4);
	size = nums * 67 + _pdf->ndim * 4;
	str = palloc(size);
	pos = str;
	bins = BINS_PTR(pdf);

	/* print histogram headers */
	for (i = 0; i < _pdf->ndim; i++) {
		HistHead *hdr = &pdf->hdrs[i];
		if (i > 0) {
			pos += sprintf(pos, " , ");
		}

		if (hdr->type == DIM_UCON) {
			pos += sprintf(pos, "[%g , %g , %d]", hdr->minv, hdr->maxv, hdr->count);
			nums -= 4;
		} else if (hdr->type == DIM_UDIS) {
			pos += sprintf(pos, "{%g", *bins++);

			for (j=1; j < hdr->count; ++j)
				pos += sprintf(pos, " , %g", *bins++);

			*pos++ = '}';
			nums -= 4 + hdr->count;
		} else { /* categorical */
			pos += sprintf(pos, "%c{%s", 
				hdr->type==DIM_UNOM? 'n':'o', get_catname(*bins++));

			for (j=1; j < hdr->count; ++j)
				pos += sprintf(pos, " , %s", get_catname(*bins++));

			*pos++ = '}';
			nums -= 4 + hdr->count;
		}
	}

	/* print the bins and the tail */
	pos += sprintf(pos, ": ");
	for (i = 0; i < nums; i++) {
		if (i > 0) {
			pos += sprintf(pos, " , ");
		}
		pos += sprintf(pos, "%g", bins[i]);
	}

	return str;
} /* histogram_print */
