/*
 * normal.c - 1D Gaussian distribution
 * 
 * Orion Project, http://orion.cs.purdue.edu/
 * Copyright (c) 2008 Purdue University
 */

#include "postgres.h"
#include "pdf.h"
#include "pdfs_io.h"
#include "uncertain.h"
#include "errors.h"
#include <math.h>

/*---------------------------------------------------------------------------*\
  Prototypes & VTable
\*---------------------------------------------------------------------------*/

typedef struct Normal {
  PDF     pdf;
  float4  mean;
  float4  stddev;
} Normal;

#define TO_NORMAL(x) Normal *x = (Normal *) _ ## x;

static float4 normal_lower(PDF*, int);
static float4 normal_upper(PDF*, int);
static float4 normal_pdf(PDF*, float4*);
static float4 normal_cdf(PDF*, float4*);
static float4 normal_mass(PDF*);
static PDF*   normal_sample(PDF*, int n);
static PDF*   normal_parse(char*, char*);
static char*  normal_print(PDF*);

#define PDF_NAME "norm"

const PDF_vtable Normal_vtbl = {
  .name = PDF_NAME,
  .id = 'n',

  .lower = normal_lower,
  .upper = normal_upper,
  .pdf = normal_pdf,
  .cdf = normal_cdf,
  .mass = normal_mass,
  .sample = normal_sample,

  .parse = normal_parse,
  .print = normal_print
};

static float4 normal_lower(PDF *_pdf, int dim) {
  TO_NORMAL(pdf);
  return pdf->mean - 3*pdf->stddev;
}

static float4 normal_upper(PDF *_pdf, int dim) {
  TO_NORMAL(pdf);
  return pdf->mean + 3*pdf->stddev;
}

static float4 normal_pdf(PDF *_pdf, float4 *x) {
  float4 m, s;
  float4 prob = 0;
  TO_NORMAL(pdf);

  /* check the sampling bounds */
  if (*x >= normal_lower(_pdf, 0) && *x <= normal_upper(_pdf, 0)) {
    /* evaluate the closed-form expression */
    m = pdf->mean;
    s = pdf->stddev;
    prob = (1.0/(s * sqrt(2.0 * M_PI))) * exp(-1.0 * sqr(*x - m) / (2.0 * sqr(s)));
  }

  return prob;
}

static float4 normal_cdf(PDF *_pdf, float4 *x) {
  float4 m, s;
  float4 prob;
  TO_NORMAL(pdf);

  /* check the sampling bounds */
  if (*x < normal_lower(_pdf, 0))
    prob = 0;

  else if (*x > normal_upper(_pdf, 0))
    prob = 1;

  else {
    /* evaluate the closed-form expression */
    m = pdf->mean;
    s = pdf->stddev;
    prob = 0.5 * (1 + erf( (*x - m) / (s * M_SQRT2)));
  }

  return prob;
}

static float4 normal_mass(PDF *_pdf) {
  return 1.0;
}

static PDF *normal_sample(PDF *pdf, int n) {
	int i;
  float4 bins[n], step, x;
  HistHead hdr = {.count = n, .type = DIM_UCON};

  hdr.minv = normal_lower(pdf, 0);
  hdr.maxv = normal_upper(pdf, 0);
	step = (hdr.maxv-hdr.minv) / n;

	for(i=0, x=hdr.minv+step/2; i < n; ++i, x += step)
//		bins[i] = normal_cdf(pdf, &x);
		bins[i] = normal_pdf(pdf, &x) * step;

  return (PDF *) histogram_init(1, &hdr, bins);
}

/*---------------------------------------------------------------------------*\
  I/O
\*---------------------------------------------------------------------------*/

static PDF *normal_parse(char *str, char *dimtypes) {
	Normal *pdf;
	char *pos = str;

	/* validate the dimension type */
	dimtypes[0] = DIM_UCON;

	/* allocate and initialize the pdf */
	pdf = palloc(sizeof(Normal));
	pdf->pdf.vtbl = &Normal_vtbl;
	pdf->pdf.ndim = 1;
	pdf->pdf.size = sizeof(Normal);

	/* parse the input string */
	skipSpace(pos);
	pos = scanFloat4(&pdf->mean, pos, PDF_NAME, str);
	skipDelim(pos, ',')
	pos = scanFloat4(&pdf->stddev, pos, PDF_NAME, str);
	checkExtra(pos, str);

	if(pdf->stddev > 0)
		return (PDF *) pdf;
	else
		return NULL;
}

static char *normal_print(PDF *_pdf) {
	char *str = palloc(28);
	TO_NORMAL(pdf);
	sprintf(str, "%g , %g", pdf->mean, pdf->stddev);
	return str;
}
