/*
 * pdf.h - base structures for probability distribution functions
 * 
 * Each uncertain data item contains one or more of the following PDF
 * structures, which in turn represent the data item's possible values.
 * 
 * Orion Project, http://orion.cs.purdue.edu/
 * Copyright (c) 2008 Purdue University
 */

#ifndef PDF_H_
#define PDF_H_

/*---------------------------------------------------------------------------*\
  Internal API for PDFs
\*---------------------------------------------------------------------------*/

#define PDF_lower(pdf, dim) ((pdf)->vtbl->lower((pdf), (dim)))
#define PDF_upper(pdf, dim) ((pdf)->vtbl->upper((pdf), (dim)))
#define PDF_pdf(_pdf, x)    ((_pdf)->vtbl->pdf((_pdf), (x)))
#define PDF_cdf(pdf, x)     ((pdf)->vtbl->cdf((pdf), (x)))
#define PDF_mass(pdf)     ((pdf)->vtbl->mass(pdf))
#define PDF_sample(pdf,n)     ((pdf)->vtbl->sample((pdf), (n)))
#define PDF_print(pdf)      ((pdf)->vtbl->print(pdf))

struct PDF_vtable;

/* PDF header structure. */
typedef struct PDF {
  const struct PDF_vtable *vtbl;

  uint32 history_id;
  uint32 ndim;  /* number of dimensions */
  uint32 size;  /* size of the struct (including the header) */
} PDF;

/* Virtual function table. */
typedef struct PDF_vtable {
  const char *name;
  const char id;

  float4  (*lower)(PDF*, int);
  float4  (*upper)(PDF*, int);
  float4  (*pdf)(PDF*, float4*);
  float4  (*cdf)(PDF*, float4*);
  float4  (*mass)(PDF*);
  PDF*    (*sample)(PDF*, int n);

  PDF*    (*parse)(char*, char*);
  char*   (*print)(PDF*);
} PDF_vtable;

const PDF_vtable *pdf_name_to_vtbl(char *name);
const PDF_vtable *pdf_id_to_vtbl(char id);

#define sqr(x) ((x)*(x))
int ipow(int base, int exp);
//float PDF_mass(PDF* pdf);

/*---------------------------------------------------------------------------*\
  Built-in PDFs
\*---------------------------------------------------------------------------*/

/* Equi-width bounds for each histogram dimension. */
typedef struct HistHead {
	float4 minv;
	float4 maxv;
	uint32 count;
	uint32 type;
} HistHead;

/* Multidimensional histogram (generic structure for continuous pdfs). */
typedef struct Histogram {
	PDF pdf;
	HistHead hdrs[1];

	/* the bins represent the CDF, and are stored in row-major order */
	/* float4 bins[1]; */
} Histogram;

#define TOLERANCE (1e-5)
#define BINS_PTR(hist) ((float4*) &(hist)->hdrs[(hist)->pdf.ndim])

Histogram *histogram_init(int32 ndim, HistHead *hdrs, float4 *bins);
Histogram *histogram_from_pdf(PDF *pdf, int32 count);
void       histogram_apply_floor(Histogram *h, const uint8 npred, const uint8 types[], float4 *preds[]);
Histogram *histogram_marginalize(Histogram *h, const uint8 *dims, uint8 len);
Histogram *histogram_product(Histogram **in, uint8 nhist);
int        is_histogram(const PDF *pdf);

PDF				*delta_init(float4);
PDF				*zero_init(int ndim);

#endif /* PDF_H_ */
