/*
 * pdfs_io.c - common parsing functions for probability distributions
 * 
 * Orion Project, http://orion.cs.purdue.edu/
 * Copyright (c) 2008 Purdue University
 */

#include "postgres.h"
#include "pdf.h"
#include "pdfs_io.h"
#include "errors.h"

/*
 * Returns the position of dlm in str; raises an error if not found.
 */
char* findDelim(char dlm, char *str) {
	char *pos = str;
	while (*pos != dlm) {
		if (*pos == '\0') {
			err_pdf_delim(dlm, str);
		}
		pos++;
	}
	return pos;
}

/*
 * Returns the expected position of the next delimiter, given the type of the
 * value just scanned.
 */
static char *nextDelim(char *pos, bool isfloat) {

	/* scan past the optional sign */
	if (*pos == '-') {
		pos++;
	}

	/* special case: infinity */
	if (isfloat && pos[0] == 'i' && pos[1] == 'n' && pos[2] == 'f') {
		pos += 3;
	} else {

		/* scan past the required digits */
		while (isdigit((int) *pos)) {
			pos++;
		}

		/* handle floating point values */
		if (isfloat) {
			/* scan past the decimal point */
			if (*pos == '.') {
				pos++;
			}
			while (isdigit((int) *pos)) {
				pos++;
			}

			/* scan past the exponent */
			if (*pos == 'e' || *pos == 'E') {
				pos++;
				if (*pos == '+' || *pos == '-') {
					pos++;
					while (isdigit((int) *pos)) {
						pos++;
					}
				}
			}

		} /* isfloat */
	} /* infinity */

	/* delimiter should be right after any white space */
	skipSpace(pos);
	return pos;

} /* nextDelim */

/*
 * Parses a float4 at the given position and returns the next position.
 */
char *scanFloat4(float4 *dst, char *pos, const char *name, char *str) {
	if (*pos == '\0' || sscanf(pos, "%g", dst) != 1) {
		err_pdf_text(name, str, pos);
	}
	return nextDelim(pos, true);
}

/*
 * Parses an integer at the given position and returns the next position.
 */
char *scanInt32(int32 *dst, char *pos, const char *name, char *str) {
	if (*pos == '\0' || sscanf(pos, "%d", dst) != 1) {
		err_pdf_text(name, str, pos);
	}
	return nextDelim(pos, false);
}

/*
 * Parses a string at the given position and returns the next position.
 * (dst may be at most 64 bytes - 63 usable characters plus terminator)
 */
char *scanString(char *dst, char *pos, const char *name, char *str) {
	char *del;
	char *end;
	int len;

	/* find the next delimiter */
	for (del = pos; *del != ',' && *del != '}'; del++) {
		if (*del == '\0') {
			err_pdf_text(name, str, pos);
		}
	}

	/* find the last non-space */
	for (end = del - 1; *end == ' '; end--) {
		if (end <= pos) {
			err_pdf_text(name, str, pos);
		}
	}

	/* check the length and copy the string */
	len = end - pos + 1;
	if (len > 63) {
		err_pdf_extra(str, pos);
	}
	strncpy(dst, pos, len);
	dst[len] = '\0';

	return del;
}

/*
 * Raises an error if pos is not at the end of str.
 */
void checkExtra(char *pos, char *str) {
	if (*pos != '\0') {
		err_pdf_extra(str, pos);
	}
}
