/*
 * poisson.c - distribution over a single point
 * 
 * Orion Project, http://orion.cs.purdue.edu/
 * Copyright (c) 2008 Purdue University
 */

#include "postgres.h"
#include "pdf.h"
#include "pdfs_io.h"
#include "uncertain.h"
#include "errors.h"
#include <math.h>

/*---------------------------------------------------------------------------*\
	Prototypes & VTable
\*---------------------------------------------------------------------------*/

typedef struct Poisson {
	PDF     pdf;
	float4  lambda;
} Poisson;

#define LAMBDA(pdf) (((Poisson *) (pdf))->lambda)

static float4 poisson_lower(PDF*, int);
static float4 poisson_upper(PDF*, int);
static float4 poisson_pdf(PDF*, float4*);
static float4 poisson_cdf(PDF*, float4*);
static float4 poisson_mass(PDF*);
static PDF*   poisson_sample(PDF*, int n);
static PDF*   poisson_parse(char*, char*);
static char*  poisson_print(PDF*);

#define PDF_NAME "pois"

const PDF_vtable Poisson_vtbl = {
	.name = PDF_NAME,
	.id = 'p',

	.lower = poisson_lower,
	.upper = poisson_upper,
	.pdf = poisson_pdf,
	.cdf = poisson_cdf,
	.mass = poisson_mass,
	.sample = poisson_sample,

	.parse = poisson_parse,
	.print = poisson_print,
};

static PDF*   poisson_init(float4 x) {
	Poisson *pdf = palloc(sizeof(Poisson));

	pdf->pdf.vtbl = &Poisson_vtbl;
	pdf->pdf.ndim = 1;
	pdf->pdf.size = sizeof(Poisson);
	pdf->lambda = x;

	return (PDF *) pdf;
}


static float4 poisson_lower(PDF *pdf, int dim) {
  float lambda = LAMBDA(pdf);
	return Max(floor(lambda - 3 * sqrt(lambda)), 0);
}

static float4 poisson_upper(PDF *pdf, int dim) {
  float lambda = LAMBDA(pdf);
	return Max(ceil(lambda + 3 * sqrt(lambda)), 4);
}

static float factorial(int k) {
	float f = 1.;

	for(; k; --k)
		f *= k;

	return f;
}

static float4 poisson_pdf(PDF *pdf, float4 *v) {
	float k = round(*v);

	if(fabs(k-*v) < TOLERANCE) {
		float lambda = LAMBDA(pdf);
		return pow(lambda, k)  * exp(-lambda) / factorial(k);
	} else
		return 0.;
}

static float4 poisson_cdf(PDF *_pdf, float4 *v) {
	float lambda = LAMBDA(_pdf), pdf = exp(-lambda), cdf = pdf;
	int i;

	if(*v < 0) return 0.;

	for(i=1; i <= *v; i ++)
		cdf += (pdf *= lambda / i);

	return cdf;
}

static float4 poisson_mass(PDF* pdf){
	return 1.0;
}

static PDF*   poisson_sample(PDF *_pdf, int n) {
	HistHead hdr = {.type=DIM_UDIS};
	Histogram *h;
	float4 *bins, lambda = ((Poisson*) _pdf)->lambda, pdf=exp(-lambda), cdf=pdf;
	int i;

	hdr.minv = PDF_lower(_pdf, 0);
	hdr.maxv = PDF_upper(_pdf, 0);
	hdr.count = 1 + (int) hdr.maxv - (int) hdr.minv;
	h = histogram_init(1, &hdr, NULL);

	bins = BINS_PTR(h);

	for(i=0; i < hdr.count; ++i)
		*bins++ = hdr.minv + i;

	for(i=1; i <= hdr.minv; ++i)
		cdf += (pdf *= lambda / i);

//	*bins++ = cdf; rohitj
	*bins++ = pdf;
	for(i=1; i < hdr.count; ++i)
		//*bins++ = (cdf += (pdf *= lambda / (i+hdr.minv))); rohitj
		*bins++ = (pdf *= lambda / (i+hdr.minv));

	return (PDF *) h;
}

/*---------------------------------------------------------------------------*\
	I/O
	\*---------------------------------------------------------------------------*/

static PDF* poisson_parse(char *str, char *dimtypes) {
	char *pos = str;
	float4 value;

	/* validate the dimension type */
	dimtypes[0] = DIM_UDIS;

	/* parse the input string */
	skipSpace(pos);
	pos = scanFloat4(&value, pos, PDF_NAME, str);
	checkExtra(pos, str);

	return poisson_init(value);
}

static char* poisson_print(PDF *pdf) {
	char *str = palloc(13);
	sprintf(str, "%g", ((Poisson*) pdf)->lambda);
	return str;
}
