/*
 * pti_uncertain : Probability Threshold Indexing (PTI) over the UNCERTAIN type.
 *
 * Orion Project, Purdue University
 * http://orion.cs.purdue.edu/
 */

#include "postgres.h"
#include "access/gist.h"
#include "access/skey.h"
#include "executor/executor.h"
#include "uncertain.h"

#define DEBUG

#define DIVISIONS 10

/*
 * Structure for sorting entries within a page.
 */
typedef struct
{
  OffsetNumber i;
  Datum key;

} SortEntry;

/*
 * Mbrx : An entry of x-bounds at 0.0 (the MBR), 0.1, 0.2, 0.3, and 0.4.
 */
typedef struct {
  UncHead hdr;
  float xbounds[10];
} Mbrx;

float get_0_bound(UncHead *input, bool left);
float get_x_bound(UncHead *input, bool left, float x_bound, bool useCached);
Mbrx *newMbrx(void);
void unionEntry(Datum *punion, Datum key);

bool mbrx_consistent(Mbrx *mbrx, float4 value, float4 threshold, StrategyNumber strategy);
bool item_consistent(UncHead *ukey, float4 value, float4 threshold, StrategyNumber strategy);

UncHead *make_copy(UncHead *orig);
int compare_x0(const void *v1, const void *v2);

PG_FUNCTION_INFO_V1(pti_consistent);
Datum pti_consistent(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(pti_union);
Datum pti_union(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(pti_compress);
Datum pti_compress(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(pti_decompress);
Datum pti_decompress(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(pti_penalty);
Datum pti_penalty(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(pti_picksplit);
Datum pti_picksplit(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(pti_same);
Datum pti_same(PG_FUNCTION_ARGS);

/******************************************************************************\
 * Debugging Macros
\******************************************************************************/

/*
 * Generates an internal error, halting the executor and rolling back.
 */
#define ABORT(MESSAGE) \
  fflush(stdout); \
  ereport(ERROR, (errcode(ERRCODE_INTERNAL_ERROR), errmsg(MESSAGE)))

#ifdef DEBUG

/*
 * Macros for printing formatted debug output.
 */
#define TRACE(FORMAT) \
  printf("  >> "); \
  printf(FORMAT); \
  fflush(stdout)
#define TRACEF(FORMAT, ARGS...) \
  printf("  >> "); \
  printf(FORMAT, ARGS); \
  fflush(stdout)

void printMbrx(Mbrx *mbrx, const char* title, ...);

/*
 * Prints the contents of an Mbrx entry along with formatted title.
 */
void printMbrx(Mbrx *mbrx, const char* title, ...)
{
  int i;

  // print the header
  if (title != NULL)
  {
    // message with variable arguments
    va_list ap;
    va_start(ap, title);
    vprintf(title, ap);
    va_end(ap);
  }
  else
  {
    // default message
    TRACEF("X-bounds at %p:\n", mbrx);
  }

  // print each x-bound pair
  for (i = 0; i < 5; i++)
  {
    printf("  0.%d : left = %.2f , right = %.2f\n",
      i, mbrx->xbounds[i], mbrx->xbounds[9-i]);
  }
  printf("\n");
  fflush(stdout);

} // void printMbrx(Mbrx *mbrx, const char* title, ...)

#else // DEBUG not defined

#define TRACE(FORMAT) ((void) 0)
#define TRACEF(FORMAT, ARGS...) ((void) 0)

#define printMbrx(mbrx, title, ...) ((void) 0)

#endif // DEBUG


/******************************************************************************\
 * PTI Support Functions
\******************************************************************************/


/***************** Function to calculate x-bounds ***************/

/* This function returns the point where the prob. of input being left (right if variable left=false)
 of the returned point is exactly equal to x_bound i.e. returns the left x_bound or right x_bound
 depending on the value of variable left, it does not check the value of x_bound for errors, so make
 sure you pass a value <= 0.5; count is the number of divisions to use for numerical integration */

extern char *histogram_print(PDF *_pdf) ;

float get_x_bound(UncHead *input, bool left, float x_bound, bool useCached) {
  int2 dims[] = {0}; // Needed for c_margin
  static Histogram *hist;
  static float4 total_mass;
  int i;
  float4 minv, maxv, interval;
  uint32 count;
  float4 *cdf;
  float4 *point; // for discrete
  float retVal;
  
  Uncertain *item;
  UncHead *copy;
  
  if (input->nldim == 0) {
    i = x_bound * 10;
	if (left)
	  return ((Mbrx *) input)->xbounds[i];
	else
	  return ((Mbrx *) input)->xbounds[9-i];
  }
  
  copy = make_copy(input);
  
  if (!useCached) {
  	item = inflate(copy);
  	total_mass = c_mass(item);
	hist = ((Histogram *) c_reduce(c_margin(item, dims, 1), DIVISIONS)->pdfs[0]);
  }

  if (!left)
    x_bound = total_mass - x_bound;
  
  if (hist->hdrs[0].type == DIM_UCON) { // Continuous
	minv = hist->hdrs[0].minv;
	maxv = hist->hdrs[0].maxv;
	count = hist->hdrs[0].count;
	interval = (maxv - minv) / count;
	cdf = (float4*) &(hist)->hdrs[(hist)->pdf.ndim];

	if (x_bound <= 0) {
	  pfree(copy);
	  return minv;
	} else if (x_bound >= total_mass) {
	  pfree(copy);
	  return maxv;
	}

	for (i=0; i < count; i++) {
	  if (cdf[i] > x_bound) break;
	}
	  
	i--; /* We got the interval in which the x_bound falls, now find the point which
	  	    corresponds to this x_bound */

	// This looks cryptic, but draw a simple figure to find out what is going on here
	if (cdf[i+1] == cdf[i]) { // highly unlikely
	  retVal = (left ? (minv + interval*(i+1)) : (minv + interval*i)); // Get a tight bound
	  pfree(copy);
	  return retVal;
	} else { // interpolate
	  retVal = minv + (interval * i) + (x_bound - cdf[i]) * interval / (cdf[i+1] - cdf[i]);
	  pfree(copy);
	  return retVal;
	}
  } else if ((hist->hdrs[0].type == DIM_UDIS) ||
		  	 (hist->hdrs[0].type == DIM_UORD)) { // Ordered discrete
	  //histogram_print(hist);
  	  minv = hist->hdrs[0].minv;
  	  maxv = hist->hdrs[0].maxv;
  	  count = hist->hdrs[0].count;
  	  //interval = (maxv - minv) / count;
  	  point = (float4*) &(hist)->hdrs[(hist)->pdf.ndim];
  	  cdf = point + count;

  	  if (x_bound <= 0) {
  	  	pfree(copy);
  	    return minv;
  	  }
  	  else if (x_bound >= 1) {
  	  	pfree(copy);
  	    return maxv;
  	  }
  	  
  	  for (i=0; i < count; i++) {
  		  if (cdf[i] >= x_bound) break;
  	  }

	  retVal = point[i];
	  pfree(copy);
  	  return retVal;
  } else { // Unordered discrete or other
	ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
					errmsg("Can't build a PTI index for type: %c",
							hist->hdrs[0].type)));
	return -1; // To make compiler warning go away
  }
}

// Optimization
float get_0_bound(UncHead *input, bool left) {
  	float retVal;
  	UncHead *copy;
  	Uncertain *item;
  	
  	if (input->nldim == 0) {
    	if (left)
	  		return ((Mbrx *) input)->xbounds[0];
		else
	  		return ((Mbrx *) input)->xbounds[9];
  	}
  	
  	copy = make_copy(input);
  	item = inflate(copy);
  	retVal = (left ? c_lower(item, 0) : c_upper(item, 0));
  	pfree(copy);
  	return retVal;	
}

/*
 * Creates and initializes a new Mbrx entry.
 */
Mbrx *newMbrx(void)
{
  Mbrx *ret;
  int i;

  // init the header
  ret = (Mbrx *) palloc(sizeof(Mbrx));
  ret->hdr.vl_len_ = sizeof(Mbrx);
  ret->hdr.nldim = 0;

  // init the x-bounds
  for (i = 0; i < 5; i++)
  {
    ret->xbounds[i] = FLT_MAX; // left bound
    ret->xbounds[9-i] = -FLT_MAX; // corresponding right bound
  }
  return ret;

} // Mbrx *newMbrx(void)

/*
 * Helper function to union an entry into its new Mbrx.
 */
void unionEntry(Datum *punion, Datum key)
{
  Mbrx *mbrx;
  UncHead *cur;
  int i;
  float xbound;
  bool useCached = false;

  // create the union mbrx, if needed
  if ( ! DatumGetPointer(*punion) )
  {
    *punion = PointerGetDatum(newMbrx());
  }
  mbrx = (Mbrx *) DatumGetPointer(*punion);
  cur = (UncHead *) DatumGetPointer(PG_DETOAST_DATUM(key));

//printMbrx(mbrx, "\nmbrx:\n");
//printf("cur type = %c\n", cur->type);

  // union the uncertain into the mbrx
  for (i = 0; i < 5; i++)
  {
    // adjust the unioned left x-bound, if needed
    xbound = get_x_bound(cur, TRUE, (float) i / 10, useCached);
    useCached = true;
    if (mbrx->xbounds[i] > xbound)
    {
      mbrx->xbounds[i] = xbound;
    }

//printf("  0.%d  left = %f", i, xbound);

    // adjust the unioned right x-bound, if needed
    xbound = get_x_bound(cur, FALSE, (float) i / 10, useCached);
    if (mbrx->xbounds[9-i] < xbound)
    {
      mbrx->xbounds[9-i] = xbound;
    }

//printf("  right = %f\n", xbound);

  } // for i

} // void unionEntry(Datum *punion, Datum key, int N)

/*
 * Returns FALSE if it's safe to prune the current search path, given the
 * current Mbrx entry and the query predicate; TRUE otherwise.
 */
bool mbrx_consistent(Mbrx *mbrx, float4 value, float4 threshold, StrategyNumber strategy)
{
  float lx, rx, a, b;
  int i;
  bool retval = TRUE;

printMbrx(mbrx, "Consistent with mbrx?\n");

  // handle the given operation
  switch (strategy)
  {
    case 1:  // <?  (uncertain, float4)
      // calculate the x-bound to use, based on the system threshold
      i = (int) floorf(threshold * 10);
      if (i > 4)
      {
        // maximum left x-bound
        i = 4;
      }

      // get the x-bound and predicate values
      lx = mbrx->xbounds[i];
      b = value;

//TRACEF("t = %.2f, i = %d, lx = %.2f, b = %.2f\n", t, i, lx, b);

      // prune this path if the interval doesn't intersect the x-bound
      if (b < lx)
      {
        retval = FALSE;
      }
      break;

    case 2:  // >?  (uncertain, float4)

      // calculate the x-bound to use, based on the system threshold
      i = 9 - (int) floorf(threshold * 10);
      if (i < 5)
      {
        // maximum right x-bound
        i = 5;
      }

      // get the x-bound and predicate values
      rx = mbrx->xbounds[i];
      a = value;

//TRACEF("t = %.2f, i = %d, rx = %.2f, a = %.2f\n", t, i, rx, a);

      // prune this path if the interval doesn't intersect the x-bound
      if (a > rx)
      {
        retval = FALSE;
      }
      break;
  } // switch (strategy)

  // return the resulting value
  return retval;

} // bool mbrx_consistent(Mbrx *mbrx, void *query, StrategyNumber strategy)

/*
 * Returns FALSE if the given uncertain key value does not pass the given
 * query predicate; TRUE otherwise.
 */
bool item_consistent(UncHead *ukey, float4 value, float4 threshold, StrategyNumber strategy)
{
  float prob;
  int2 dims[] = {0}; // Needed for c_margin
  UncHead *copy = make_copy(ukey);
  Uncertain *item = c_margin(inflate(ukey), dims, 1);
  float4 minv[] = {c_lower(item, 0)};
  float4 maxv[] = {value};
  
  // use the comparison function
  prob = c_prob(item, minv, maxv); // item <% value
  if (prob < 0) prob = 0; // TODO: HACK!

  // invert the probability, if applicable
  if (strategy == 2) {
    prob = c_mass(item) - prob;
  }

  pfree(copy);
  // final test: does it pass the threshold?
  return (prob >= threshold);
} // bool item_consistent(Uncertain *ukey, void *query, StrategyNumber strategy)

/*
 * Make a defensive copy 
 */
UncHead *make_copy(UncHead *orig) {
	UncHead *ret = (UncHead *) palloc(VARSIZE(orig));
	memcpy(ret, orig, VARSIZE(orig));
	return ret;
}

/*
 * Comparison function for sorting entries within a page.
 */
int compare_x0(const void *v1, const void *v2)
{
  UncHead *u1 = (UncHead *) DatumGetPointer( PG_DETOAST_DATUM(((SortEntry *) v1)->key ));
  UncHead *u2 = (UncHead *) DatumGetPointer( PG_DETOAST_DATUM(((SortEntry *) v2)->key ));
  float l1, l2;
	
  // get the left and right bounds of the entries
  l1 = get_0_bound(u1, TRUE);
//  float r1 = get_x_bound(u1, FALSE, 0, N);
  l2 = get_0_bound(u2, TRUE);
//  float r2 = get_x_bound(u2, FALSE, 0, N);

  // for now, simply compare left boundaries
  return l1 - l2;
} // int compare_x0(const void *v1, const void *v2)

/******************************************************************************\
 * PTI operator class methods for GiST
\******************************************************************************/

/*
 * Given a predicate p on a tree page, and a user query, q, this method will
 * return false if it is certain that both p and q cannot be true for a given
 * data item.
 */
Datum pti_consistent(PG_FUNCTION_ARGS)
{
  GISTENTRY *entry = (GISTENTRY *) PG_GETARG_POINTER(0);
  HeapTupleHeader t = PG_GETARG_HEAPTUPLEHEADER(1);
  //void *query = PG_GETARG_POINTER(1);
  StrategyNumber strategy = (StrategyNumber) PG_GETARG_UINT16(2);
  bool retval;
  bool isnull;
  float4 value, threshold;
  Datum temp;
  
TRACEF("*** pti_consistent() called; oper = %d , leaf = %d\n", strategy, GIST_LEAF(entry));

  // Extract the threshold and value from query
  temp = GetAttributeByName(t, "value", &isnull);
  if (isnull) {
		ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
						errmsg("value can't be NULL")));
  } else
	  value = DatumGetFloat4(temp);
  
  temp = GetAttributeByName(t, "threshold", &isnull);
  if (isnull) {
		ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
						errmsg("threshold can't be NULL")));
  } else
	  threshold = DatumGetFloat4(temp);

    // call the applicable helper function
  if (GIST_LEAF(entry))
  {
    retval = item_consistent((UncHead *) DatumGetPointer(PG_DETOAST_DATUM(entry->key)),
    		value, threshold, strategy);
  }
  else
  {
    retval = mbrx_consistent((Mbrx *) DatumGetPointer(PG_DETOAST_DATUM(entry->key)),
    		value, threshold, strategy);
  }

TRACEF("Returning %s\n\n", retval ? "TRUE" : "FALSE");
//ABORT("Stoping here");

  // return the resulting value
  PG_RETURN_BOOL(retval);

} // Datum pti_consistent(PG_FUNCTION_ARGS)

/*
 * This method consolidates information in the tree. Given a set of entries,
 * this function generates a new predicate that is true for all the entries.
 *
 * In other words, calculates the new set of x-bounds for the given node.
 */
Datum pti_union(PG_FUNCTION_ARGS)
{
  GistEntryVector *entryvec = (GistEntryVector *) PG_GETARG_POINTER(0);
  int *sizep = (int *) PG_GETARG_POINTER(1);
  int i;

  // initialize the union result
  Mbrx *mbrx = NULL;
  *sizep = sizeof(Mbrx);

//TRACEF("*** pti_union() called, n = %d, 0: %c, 1: %c\n", entryvec->n,
//    ((Uncertain *)entryvec->vector[0].key)->type,
//    ((Uncertain *)entryvec->vector[1].key)->type);

  // find the minimum x-bounds for all entries
  for (i = 0; i < entryvec->n; i++)
  {
    unionEntry( (Datum *) &mbrx , entryvec->vector[i].key);
  }

//printMbrx(mbrx, "\nUnion mbrx:\n");
//ABORT("Stopping here");

  // return the resulting entry
  PG_RETURN_POINTER(mbrx);

} // Datum pti_union(PG_FUNCTION_ARGS)

/*
 * Converts the data item into a format suitable for physical storage in an
 * index page.
 */
Datum pti_compress(PG_FUNCTION_ARGS)
{
  // does nothing for now; storing data as is
  PG_RETURN_POINTER( PG_GETARG_POINTER(0) );

} // Datum pti_compress(PG_FUNCTION_ARGS)

/*
 * The reverse of the compress method. Converts the index representation of
 * the data item into a format that can be manipulated by the database.
 */
Datum pti_decompress(PG_FUNCTION_ARGS)
{
//  TRACE("*** pti_decompress() called\n");

  // does nothing for now; storing data as is
  PG_RETURN_POINTER( PG_GETARG_POINTER(0) );

} // Datum pti_decompress(PG_FUNCTION_ARGS)

/*
 * Returns a value indicating the "cost" of inserting the new entry into a
 * particular branch of the tree. items will be inserted down the path of
 * least penalty in the tree.
 */
Datum pti_penalty(PG_FUNCTION_ARGS)
{
  GISTENTRY *origentry = (GISTENTRY *) PG_GETARG_POINTER(0);
  GISTENTRY *newentry = (GISTENTRY *) PG_GETARG_POINTER(1);
  float *result = (float *) PG_GETARG_POINTER(2);
  float orgl, orgr, newl, newr;

  // get the left and right bounds of the entries
  orgl = get_0_bound((UncHead *) DatumGetPointer(PG_DETOAST_DATUM(origentry->key)), TRUE);
  orgr = get_0_bound((UncHead *) DatumGetPointer(PG_DETOAST_DATUM(origentry->key)), FALSE);
  newl = get_0_bound((UncHead *) DatumGetPointer(PG_DETOAST_DATUM(newentry->key)), TRUE);
  newr = get_0_bound((UncHead *) DatumGetPointer(PG_DETOAST_DATUM(newentry->key)), FALSE);

  // calculate the potential increase of the MBR (0.0 x-bound)
  *result = 0;
  if (newl < orgl)
  {
    *result += orgl - newl;
  }
  if (newr > orgr)
  {
    *result += newl - orgl;
  }

//  TRACEF("*** pti_penalty() called; cost = %.2f\n", *result);

  // return the resulting cost
  PG_RETURN_POINTER(result);

} // Datum pti_penalty(PG_FUNCTION_ARGS)

/*
 * When a page split is necessary, this function decides which entries on the
 * page are to stay on the old page, and which are to move to the new page.
 */
Datum pti_picksplit(PG_FUNCTION_ARGS)
{
  GistEntryVector *entryvec = (GistEntryVector *) PG_GETARG_POINTER(0);
  GIST_SPLITVEC *splitvec = (GIST_SPLITVEC *) PG_GETARG_POINTER(1);

  OffsetNumber i, maxoff = entryvec->n - 1;
  int nbytes = (maxoff + 2) * sizeof(OffsetNumber);

  // first sort all the entries to split
  SortEntry *sorted = (SortEntry *) palloc(entryvec->n * sizeof(SortEntry));
  for (i = FirstOffsetNumber; i <= maxoff; i = OffsetNumberNext(i))
  {
    sorted[i].i = i;
    sorted[i].key = entryvec->vector[i].key;
  }
  qsort((void *) &sorted[FirstOffsetNumber], maxoff - FirstOffsetNumber + 1,
      sizeof(SortEntry), compare_x0);

  // initialize the output vector
  splitvec->spl_left = (OffsetNumber *) palloc(nbytes);
  splitvec->spl_right = (OffsetNumber *) palloc(nbytes);
  splitvec->spl_ldatum = PointerGetDatum(0);
  splitvec->spl_rdatum = PointerGetDatum(0);
  splitvec->spl_nleft = 0;
  splitvec->spl_nright = 0;

  // for now, split right down the middle of the sorted list
  for (i = FirstOffsetNumber; i <= maxoff; i = OffsetNumberNext(i))
  {
    if (i <= (maxoff - FirstOffsetNumber + 1) / 2)
    {
      unionEntry( & splitvec->spl_ldatum , sorted[i].key);
      splitvec->spl_left[splitvec->spl_nleft] = sorted[i].i;
      splitvec->spl_nleft++;
    }
    else
    {
      unionEntry( & splitvec->spl_rdatum , sorted[i].key);
      splitvec->spl_right[splitvec->spl_nright] = sorted[i].i;
      splitvec->spl_nright++;
    }
  }

//printMbrx((Mbrx *) DatumGetPointer(splitvec->spl_ldatum), "\nLeft mbrx:\n");
//printMbrx((Mbrx *) DatumGetPointer(splitvec->spl_rdatum), "Right mbrx:\n");

//  TRACEF("*** pti_picksplit() called; leaf = %d , %d in left , %d in right\n",
//      entryvec->vector[0].leafkey == 0, splitvec->spl_nleft, splitvec->spl_nright);
  PG_RETURN_POINTER(splitvec);

} // Datum pti_picksplit(PG_FUNCTION_ARGS)

/*
 * Returns true if two entries are identical, false otherwise.
 */
Datum pti_same(PG_FUNCTION_ARGS)
{
  UncHead *u1 = (UncHead *) DatumGetPointer(PG_DETOAST_DATUM(PG_GETARG_POINTER(0)));
  UncHead *u2 = (UncHead *) DatumGetPointer(PG_DETOAST_DATUM(PG_GETARG_POINTER(1)));
  bool *result = (bool *) PG_GETARG_POINTER(2);

  if (u1 && u2)
  {
    // if the headers match, compare the data bytes
    *result =
      (VARSIZE(u1) == VARSIZE(u2)) &&
      (memcmp(u1, u2, VARSIZE(u1)) == 0)
      ? TRUE : FALSE;
  }
  else
  {
    // special case: both are NULL
    *result =
      (u1 == NULL) &&
      (u2 == NULL)
      ? TRUE : FALSE;
  }

//  TRACEF("*** pti_same() called; result = %s\n", *result ? "TRUE" : "FALSE");

  // return the resulting equality
  PG_RETURN_POINTER(result);

} // Datum pti_same(PG_FUNCTION_ARGS)
