/*
 * dependency_set.c - Utility functions for rewriter module
 *
 * Orion Project, http://orion.cs.purdue.edu/
 */

#include <string.h>

#include "dependency_set.h"
#include "debug.h"


/************** Begin Function prototypes *********************/
void processAttr(int attr_no, DependencySet *dependency_set, int size);
void updateDependency(char *begin_attr_name, char *end_attr_name,
                      DependencySet *dependency_set, int size,
                      int stored_in, int dimension);
bool matches(char *str1, char *str2, int number_chars);
/************** End Function prototypes *********************/

const char *ATTR_SEPARATOR = "__";
const char NOT_STORED_HERE_END_MARKER = '*';

/* For debugging purposes */
void printDependencySet(DependencySet *d, int size) {
  int i;
  for (i=0; i < size; i++) {
    DEBUG("Attr name = %s, stored in %d at %d",
          d[i].attr_name, d[i].stored_in, d[i].dimension);
  }
}

/* Returns true if the first number_chars of str1 and str2 matches. Additionally
   the char after the last matching character of str1 should be
   NOT_STORED_HERE_END_MARKER. str2 will contain atleast number_chars.
   There are no such guarrantees about str1*/
bool matches(char *str1, char *str2, int number_chars) {
  int i;

  for (i = 0; i < number_chars; i++) {
    if ((str1[i] == '\0') || (str1[i] != str2[i]))
      return false;
  }

  if (str1[number_chars] == NOT_STORED_HERE_END_MARKER)
    return true;
  else
    return false;
}

/* Updates the dependency information of attribute name specified by the
   string between begin_attr_name (inclusive) and end_attr_name (exclusive) to
   attr_no and dimension. If end_attr_name is null, just match till the end of
   the string.
 */
void updateDependency(char *begin_attr_name, char *end_attr_name,
                      DependencySet *dependency_set, int size,
                      int stored_in, int dimension) {
  int i;
  int attr_name_size = (end_attr_name == NULL) ? strlen(begin_attr_name) : (end_attr_name - begin_attr_name);

  for (i = 0; i < size; i++) {
    if (matches(dependency_set[i].attr_name, begin_attr_name, attr_name_size)) {
      // Update
      dependency_set[i].stored_in = stored_in;
      dependency_set[i].dimension = dimension;
      dependency_set[i].attr_name[attr_name_size] = '\0';
    }
  }
}


/* Processes attr_no. If the attr is a jointly distributed attribute (i.e.
   "...__....__....." updates the dependency information of all the attributes
   stored in attr_no */
void processAttr(int attr_no, DependencySet *dependency_set, int size) {
  int dimension = 0;
  char *attr_name = dependency_set[attr_no].attr_name;
  char *begin_attr_name = attr_name,
       *end_attr_name = strstr(begin_attr_name, ATTR_SEPARATOR);

  if (end_attr_name == NULL) // Not jointly distributed
    return;

  while (end_attr_name != NULL) {
    begin_attr_name = end_attr_name + strlen(ATTR_SEPARATOR);
    end_attr_name = strstr(begin_attr_name, ATTR_SEPARATOR);

    updateDependency(begin_attr_name, end_attr_name, dependency_set, size, attr_no, ++dimension);
  }

  /* Update the current attribute name */
  attr_name = strstr(attr_name, ATTR_SEPARATOR);
  attr_name[0] = '\0';
  /* We could have freed the extra space allocated. As this is temporary
     storage, doing a pfree followed by another palloc is not worth the
     trouble. */
}


/* Main function */
int getDependencySet(const Query * const qry, DependencySet *output[]) {
  ListCell *l;
  TargetEntry *attr;
  int attr_size = qry->targetList->length;
  DependencySet *dependency_set;
  int i;

  *output = (DependencySet *) palloc( attr_size * sizeof(DependencySet));
  dependency_set = *output;

  /* Copy the qry targetList into dependency_set */
  i = 0;
  foreach(l, qry->targetList) {
    attr = ((TargetEntry *) lfirst(l));
    dependency_set[i].attr_name = palloc(strlen(attr->resname) + 1);
    strcpy(dependency_set[i].attr_name, attr->resname);
    dependency_set[i].stored_in = i;
    dependency_set[i].dimension = 0;
    ++i;
  }

  /* Process dependency_set to figure out all the dependencies */
  for (i = 0; i < attr_size; i++) {
    processAttr(i, dependency_set, attr_size);
  }

  printDependencySet(dependency_set, attr_size);
  return attr_size;
}
