/*
 * dependent.c - support for the DEPENDENT keyword and dependency sets
 * 
 * Orion Project, http://orion.cs.purdue.edu/
 * Copyright (c) 2008 Purdue University
 */

#include "postgres.h"
#include "orion_bki.h"

#include "catalog/namespace.h"
#include "executor/spi.h"
#include "utils/syscache.h"

/*---------------------------------------------------------------------------*\
  SPI Functions
\*---------------------------------------------------------------------------*/

/*
 * Gets the next available dependency set identifier for the given relation.
 * 
 * The following query didn't work because of transaction isolation:
 * SELECT COALESCE(max(depid),0) FROM orion_depend WHERE relid = %d
 */
static int32 next_depid()
{
	char sql[100];
	int32 ret;
	bool isnull;
	int32 depid = 0;

	/* connect to the DB and execute the query */
	SPI_connect();
	snprintf(sql, 100, "SELECT nextval('orion_depend_seq')::integer");
	ret = SPI_execute(sql, true, 1);

	/* get the query result */
	if (ret == SPI_OK_SELECT)
	{
		TupleDesc tupdesc = SPI_tuptable->tupdesc;
		HeapTuple tuple = SPI_tuptable->vals[0];
		depid = DatumGetInt32( SPI_getbinval(tuple, tupdesc, 1, &isnull) );
	}

	/* close the DB connection */
	SPI_finish();
	return depid;

} /* next_depid */

/*
 * Inserts the given dependency set tuple into the catalog.
 */
static void depend_ins(Oid relid, int16 attnum, int32 depid, int16 dimnum)
{
	char sql[100];
	int32 ret;

	/* connect to the DB and execute the query */
	SPI_connect();
	snprintf(sql, 100,
			"INSERT INTO orion_depend VALUES (%d, %d, %d, %d)",
			relid, attnum, dimnum, depid);
	ret = SPI_execute(sql, false, 1);

	/* close the DB connection */
	SPI_finish();

} /* depend_ins */

/*---------------------------------------------------------------------------*\
  Backend Functions
\*---------------------------------------------------------------------------*/

/*
 * Creates a dependency set for the given relation; this is the final step of
 * CREATE TABLE when it includes a DEPENDENT constraint.
 */
void orion_dependent(RangeVar *heapRelation, List *attributeList)
{
	ListCell *cell;
	int32 depid, i;
	int16 *attnums = palloc(2 * list_length(attributeList));
	Oid relid = RangeVarGetRelid(heapRelation, false);
	int32 attn = 0;

	/* get the attnum of each attribute (see ComputeIndexAttrs in indexcmds.c) */
	foreach(cell, attributeList)
	{
		HeapTuple atttuple;
		Form_pg_attribute attform;
		IndexElem *attribute = (IndexElem *) lfirst(cell);

		/* save the column's attnum */
		atttuple = SearchSysCacheAttName(relid, attribute->name);
		attform = (Form_pg_attribute) GETSTRUCT(atttuple);
		attnums[attn] = attform->attnum;
		ReleaseSysCache(atttuple);
		attn++;
	}

	/* insert the dependency set into the catalog */
	depid = next_depid();
	for (i = 0; i < attn; i++)
	{
		depend_ins(relid, attnums[i], depid, i + 1);
	}

} /* orion_dependent */
