/*
 * util.c - Utility functions for rewriter module
 * 
 * Orion Project, http://orion.cs.purdue.edu/
 * Copyright (c) 2008 Purdue University
 */

#include "postgres.h"
#include "nodes/print.h"

#include "utils.h"
#include "utils/builtins.h"
#include "uncertain.h"
#include "oid_functions.h"
#include "debug.h"
#include "errors.h"

/*Solaris does not have vasprintf, so here is an implementation*/
int myvasprintf(char **ret, const char *format, va_list args) {
	int count = 0;
	va_list copy;
	va_copy(copy, args);

	/* Make sure it is determinate, despite manuals indicating otherwise */
	*ret = 0;

	count = vsnprintf(NULL, 0, format, args);
	if (count >= 0) {
		char* buffer = malloc(count + 1);
		if (buffer != NULL) {
			count = vsnprintf(buffer, count + 1, format, copy);
			if (count < 0)
				free(buffer);
			else
				*ret = buffer;
		}
	}
	va_end(args);  // Each va_start() or va_copy() needs a va_end()

	return count;
}

/*Solaris does not have asprintf, so here is an implementation*/
int myasprintf(char **ret, const char *format, ...) {
	int count = 0;
	va_list args;
	va_start(args, format);
	count = myvasprintf(ret, format, args);
	va_end(args);
	return(count);
}


/* Get Oid of typename */
Oid getTypeOid(char *typename) {
	char sql[100];
	int32 ret;
	bool isnull;
	Oid ans = (Oid) 0;

	/* connect to the DB */
	SPI_connect();

	snprintf(sql, 100, "SELECT oid FROM pg_type WHERE typname = lower('%s')", typename);

	ret = SPI_execute(sql, true, 1);

	if (SPI_processed == 0) {
		char *errorHint = palloc(100);
		snprintf(errorHint, 100, "getTypeOid -  %s is not a valid type name",
				typename);
		err_custom("Failed retrieving type information from db", errorHint);
		pfree(errorHint);
	}

	if (ret == SPI_OK_SELECT)
	{
		TupleDesc tupdesc = SPI_tuptable->tupdesc;
		HeapTuple tuple = SPI_tuptable->vals[0];
		ans = DatumGetObjectId( SPI_getbinval(tuple, tupdesc, 1, &isnull) );
	}

	/* close the DB connection */
	SPI_finish();
	
	return ans;
}

/* Get Oid of Func */
Oid getFuncOid(char *func) {
	char sql[100];
	int32 ret;
	bool isnull;
	Oid ans = (Oid) 0;

	/* connect to the DB */
	SPI_connect();
	
	snprintf(sql, 100, "SELECT oid FROM pg_proc WHERE proname = lower('%s')", func);
	
	ret = SPI_execute(sql, true, 1);
	
	if (SPI_processed == 0) {
		char *errorHint = palloc(100);
		snprintf(errorHint, 100, "getFuncOid -  %s is not a valid function name",
				func);
		err_custom("Failed retrieving type information from db", errorHint);
		pfree(errorHint);
	}
	
	
	if (ret == SPI_OK_SELECT)
	{
		TupleDesc tupdesc = SPI_tuptable->tupdesc;
		HeapTuple tuple = SPI_tuptable->vals[0];
		ans = DatumGetObjectId( SPI_getbinval(tuple, tupdesc, 1, &isnull) );
	}
	
	/* close the DB connection */
	SPI_finish();

	return ans;
}

Oid getVarType(Oid relid, AttrNumber attrNum) {
	char sql[100];
	int32 ret;
	bool isnull;
	Oid ans = (Oid) 0;

	/* connect to the DB */
	SPI_connect();
	
	snprintf(sql, 100, "select atttypid from pg_attribute where attrelid = %d and attnum = %d", relid, attrNum);
	
	ret = SPI_execute(sql, true, 1);
	if (ret == SPI_OK_SELECT)
	{
		TupleDesc tupdesc = SPI_tuptable->tupdesc;
		HeapTuple tuple = SPI_tuptable->vals[0];
		ans = DatumGetObjectId( SPI_getbinval(tuple, tupdesc, 1, &isnull) );
	}
	/* close the DB connection */
	SPI_finish();
	
	return ans;	
}

/* Return the Oid of func that can cast initialOid to float4 */
Oid getConvertFloat4Oid(Oid initialOid) {
	char sql[200];
	int32 ret;
	bool isnull;
	Oid ans = (Oid) 0;

	/* connect to the DB */
	SPI_connect();
	
	snprintf(sql, 200, "SELECT oid FROM pg_proc WHERE proname = 'float4' and proargtypes[0] = %d", initialOid);
	
	ret = SPI_execute(sql, true, 1);
	
	if (SPI_processed == 0)
		ans = InvalidOid;
	else {
		if (ret == SPI_OK_SELECT) {
			TupleDesc tupdesc = SPI_tuptable->tupdesc;
			HeapTuple tuple = SPI_tuptable->vals[0];
			ans = DatumGetObjectId( SPI_getbinval(tuple, tupdesc, 1, &isnull) );
		} else
			ans = InvalidOid;
	}
	
	/* close the DB connection */
	SPI_finish();

	return ans;
}

/* Check if varType is one of UCON/UDIS/UNOM/UORD */
bool isUncertainType(Oid varType) {
	Oid ucon_oid = getUConOid();
	Oid udis_oid = getUDisOid();
	Oid unom_oid = getUNomOid();
	Oid uord_oid = getUOrdOid();

	if ( (varType == ucon_oid) ||
		 (varType == udis_oid) ||
		 (varType == unom_oid) ||
		 (varType == uord_oid))
		return true;
	else
		return false;
}


/* Check if relid, attnum is part of a dependency set. Updates dimnum (dimension number) and depid (Dependency id) and depdim(dimension of dependency pdf) if true. */
bool isJointlyDistributed(Oid relid, AttrNumber attnum, int16 *dimnum, int32 *depid, int64 *depdim) {
	char sql[300];
	int32 ret;
	bool isnull;

	SPI_connect();
	snprintf(sql, 300, "SELECT temp.dimnum,temp.depid,count(temp2.dimnum) FROM orion_depend as temp, orion_depend as temp2 WHERE temp.relid=%d AND temp.attnum=%d AND temp.depid = temp2.depid GROUP BY temp.dimnum, temp.depid ", relid, attnum);
	ret = SPI_execute(sql, true, 1);

	if (SPI_processed == 0) { /* No rows returned */
		SPI_finish();
		return false;
	}

	/* We got a dependency set, get the dimnum and depid */
	if (ret == SPI_OK_SELECT) {
		TupleDesc tupdesc = SPI_tuptable->tupdesc;
		HeapTuple tuple = SPI_tuptable->vals[0];
		*dimnum = DatumGetInt16( SPI_getbinval(tuple, tupdesc, 1, &isnull) );
		*depid = DatumGetInt32( SPI_getbinval(tuple, tupdesc, 2, &isnull) );
		*depdim = DatumGetInt64( SPI_getbinval(tuple, tupdesc, 3, &isnull) );
	}

	/* close the DB connection */
	SPI_finish();	
	return true;
}

/* Get AttrNum of depid -- the lowest attrnum corresponding to the depid */
AttrNumber getAttrNum(int32 depid) {
	char sql[100];
	int32 ret;
	bool isnull;
	AttrNumber ans = (AttrNumber) 0;

	/* connect to the DB */
	SPI_connect();
	
	snprintf(sql, 100, "SELECT attnum FROM orion_depend WHERE depid = %d ORDER BY attnum LIMIT 1", depid);
	
	ret = SPI_execute(sql, true, 1);
	if (ret == SPI_OK_SELECT)
	{
		TupleDesc tupdesc = SPI_tuptable->tupdesc;
		HeapTuple tuple = SPI_tuptable->vals[0];
		ans = DatumGetInt16( SPI_getbinval(tuple, tupdesc, 1, &isnull) );
	}
	
	/* close the DB connection */
	SPI_finish();

	return ans;
}	

/* Returns the const (either integer or numeric) stored in Const */
char *getConstValue(Const *cExpr) {
	char *ans = NULL;
	
	if (cExpr->consttype == getInt4Oid()) {
		ans = palloc(100);
		snprintf(ans, 100, "%d", DatumGetInt32(cExpr->constvalue));
	} else if (cExpr->consttype == getNumericOid())
		ans = DatumGetCString(DirectFunctionCall1(numeric_out,cExpr->constvalue));
	else		
		elog(ERROR, "Orion Internal error: Not implemented Yet.");
		
	return ans;
}	

/* Get the "type string" (i.e. "ccc") of depid */
char *getTypeString(Oid relid, int32 depid) {
	char sql[200];
	int32 ret;
	bool isnull;
	
	int size = 0, i;
	int *temp = NULL;
	char *ans;
	Oid varType;

	/* connect to the DB */
	SPI_connect();
	
	snprintf(sql, 200, "SELECT attnum FROM orion_depend WHERE depid = %d ORDER BY attnum", depid);

	ret = SPI_execute(sql, true, 0);
	if (ret == SPI_OK_SELECT) {
		TupleDesc tupdesc = SPI_tuptable->tupdesc;
		size = SPI_processed;
		temp = malloc(sizeof(int) * size);				
		for(i = 0; i < size; i++) {
			HeapTuple tuple = SPI_tuptable->vals[i];
			temp[i] = (AttrNumber) DatumGetInt16(SPI_getbinval(tuple, tupdesc, 1, &isnull));
		}
	}
	
	SPI_finish();
	
	/* Make the type string */
	ans = palloc(sizeof(char) * size + 1);
	for(i = 0; i < size; i++) {
		varType = getVarType(relid, temp[i]); 
		if (varType == getUConOid()) ans[i] = DIM_UCON;
		else if (varType == getUDisOid()) ans[i] = DIM_UDIS; 
		else if (varType == getUNomOid()) ans[i] = DIM_UNOM; 
		else ans[i] = DIM_UORD; /* Not checking for any error case, but it should be one of these */
	}
	ans[size] = '\0';
	
	free(temp);
	return ans;	
}

/* Orion pprint for printing nodes during debugging. This is similar to 
 * nodes/print.c's pprint. msg is something that is printed before the 
 * query node (obj) */
void orion_debug_pprint(char *msg, void *obj) {
	char	   *s;
	char	   *f;

	DEBUG("%s", msg);
	
	s = nodeToString(obj);
	f = pretty_format_node_dump(s);
	pfree(s);
	DEBUG("%s", f);
	fflush(stdout);
	pfree(f);
}

#define WITH_QUERY(body, cols, sql_fmt, ...) { \
	char *sql; \
	int32 ret; \
	myasprintf(&sql, sql_fmt, ##__VA_ARGS__); \
	SPI_connect(); \
	ret = SPI_execute(sql, true, 0); \
	if(ret == SPI_OK_SELECT) { \
		int rownum; \
		for(rownum=0; rownum < SPI_processed; rownum++) { \
			Datum vals[(cols)]; \
			bool isnull[(cols)]; \
			HeapTuple tuple = SPI_tuptable->vals[rownum]; \
			{ \
			int i; \
			for(i=0; i < (cols); ++i) \
				vals[i] = SPI_getbinval(tuple, SPI_tuptable->tupdesc, i+1, &isnull[i]); \
			} \
			{body} \
		} \
	} \
	SPI_finish(); \
	free(sql); \
}


#undef DEBUG
#define DEBUG(fmt, ...) do{fprintf(stderr, fmt, ##__VA_ARGS__); fflush(stderr); }while(0)

static void orion_debug_query_pprint_plain(Query*);

static void orion_debug_system_pprint(Node *n) {
	char *s;
	s = nodeToString(n);
	DPRINTF("%s", s);
	pfree(s);
}
	
static void orion_debug_rte_pprint(RangeTblEntry *rte) {
	switch(rte->rtekind) {
		case RTE_RELATION: 
			WITH_QUERY(
				DEBUG("%s", NameStr(*DatumGetName(vals[0])));
				, 1, "SELECT relname FROM pg_class WHERE oid=%i", rte->relid);
			break;

		case RTE_SUBQUERY:
			DEBUG("(");
			orion_debug_query_pprint_plain(rte->subquery);
			DEBUG(")");
			break;

		default:
			DEBUG("(");
			orion_debug_system_pprint((Node*) rte);
			DEBUG(")");
			break;
	}
	if(rte->alias)
		DEBUG(" AS %s", rte->alias->aliasname);
	
}

static void orion_debug_const_pprint(Const *c) {
	if(c->constisnull) {
		DEBUG("NULL");
		return;
	} else {
		Oid					foid;
		char				*output;
		WITH_QUERY(
			foid = DatumGetObjectId(vals[0]);
		, 1,  "SELECT typoutput FROM pg_type WHERE oid=%i", c->consttype);

		output = OidOutputFunctionCall(foid, c->constvalue);
		DEBUG("%s", output);
		pfree(output);
	}
}

static List *case_expr_stack = NIL;

static void orion_debug_expr_pprint(Expr *expr, List *rtable) {


	if(IsA(expr, Var)) {
		Var *var = (Var*) expr;
		RangeTblEntry *rte = NULL;

		/* In case the references range table entry is out of range, don't crash */
		if(var->varno < 1 || var->varno > list_length(rtable))
			DEBUG("%i.", var->varno-1);
		else {
			rte = list_nth(rtable, var->varno-1);
			DEBUG("%s.", rte->eref->aliasname);
		}

		if(var->varattno == 0)
			DEBUG("* ");
		else if(var->varattno < 0 || rte == NULL) /*Gracefully handle weird values*/
			DEBUG("%i ", var->varattno);
		else
			DEBUG("%s", strVal(list_nth(rte->eref->colnames, var->varattno-1)));

	} else if(IsA(expr, FuncExpr)) {
		FuncExpr *fnexp = (FuncExpr*) expr;
		ListCell *cell;

		WITH_QUERY(
			DEBUG("%s(", NameStr(*DatumGetName(vals[0])));
			, 1,  "SELECT proname FROM pg_proc WHERE oid=%i", fnexp->funcid);

		foreach(cell, fnexp->args) {
			if(cell != list_head(fnexp->args))
				DEBUG(", ");

			orion_debug_expr_pprint((Expr*) lfirst(cell), rtable);
		}

		DEBUG(")");

	} else if(IsA(expr, Const)) {
		orion_debug_const_pprint((Const*) expr);

	} else if(IsA(expr, RelabelType)) {
		RelabelType *rt = (RelabelType *) expr;
		orion_debug_expr_pprint(rt->arg, rtable);
		DEBUG("::");
		WITH_QUERY(
			DEBUG("%s", NameStr(*DatumGetName(vals[0])));
			, 1, "SELECT typname FROM pg_type WHERE oid=%i", rt->resulttype);

	} else if(IsA(expr, CaseExpr)) {
		CaseExpr *ce = (CaseExpr *) expr;
		ListCell *cell;

		DEBUG("CASE ");

		if(ce->arg) {
			orion_debug_expr_pprint(ce->arg, rtable);
			DEBUG(" ");

			case_expr_stack = lappend(case_expr_stack, ce->arg);
		}

		foreach(cell, ce->args) {
			CaseWhen *cw = (CaseWhen*) lfirst(cell);
			DEBUG("WHEN ");
			orion_debug_expr_pprint(cw->expr, rtable);
			DEBUG(" THEN ");
			orion_debug_expr_pprint(cw->result, rtable);
			DEBUG(" ");
		}

		if(ce->defresult) {
			DEBUG("ELSE ");
			orion_debug_expr_pprint(ce->defresult, rtable);
			DEBUG(" ");
		}

		if(ce->arg)
			case_expr_stack = list_truncate(case_expr_stack, list_length(case_expr_stack)-1);

		DEBUG("END");

	} else if(IsA(expr, OpExpr)) {
		OpExpr *oe = (OpExpr *) expr;
		char op[10];

		WITH_QUERY(
			strncpy(op, NameStr(*DatumGetName(vals[0])), sizeof(op));
			, 1, "SELECT oprname FROM pg_operator WHERE oid=%i", oe->opno);
		op[sizeof(op)-1] = '\0';

		if(list_length(oe->args) == 1) {
			DEBUG("%s", op);
			orion_debug_expr_pprint((Expr*) linitial(oe->args), rtable);
		} else if(list_length(oe->args) == 2) {
			orion_debug_expr_pprint((Expr*) linitial(oe->args), rtable);
			DEBUG("%s", op);
			orion_debug_expr_pprint((Expr*) lsecond(oe->args), rtable);
		} else {
			ListCell *cell;
			DEBUG("%s", op);
			DEBUG("(");
			foreach(cell, oe->args) {
				orion_debug_expr_pprint((Expr*) lfirst(cell), rtable);
				DEBUG(", ");
			}
			DEBUG(")");
		}

	} else if(IsA(expr, CaseTestExpr)) {
		orion_debug_expr_pprint((Expr*) llast(case_expr_stack), rtable);

	} else if(IsA(expr, BoolExpr)) {
		BoolExpr *be = (BoolExpr *) expr;
		ListCell *cell;
		const char *opstr = NULL;

		switch(be->boolop) {
		case AND_EXPR: opstr = " AND "; break;
		case OR_EXPR: opstr = " OR "; break;
		case NOT_EXPR: opstr = " NOT "; 
			DEBUG("NOT ");
			break;
		}

		foreach(cell, be->args) {
			if(cell != list_head(be->args))
				DEBUG("%s", opstr);
			orion_debug_expr_pprint(lfirst(cell), rtable);
		}
	
	} else
		orion_debug_system_pprint((Node*) expr);
}
	
static void orion_debug_query_pprint_plain(Query *q) {
	ListCell *cell;

	/* Print command */
	switch(q->commandType) {
		case CMD_UNKNOWN: DEBUG("UNKNOWN "); break;
		case CMD_SELECT: DEBUG("SELECT "); break;
		case CMD_UPDATE: DEBUG("UPDATE "); break;
		case CMD_INSERT: DEBUG("INSERT "); break;
		case CMD_DELETE: DEBUG("DELETE "); break;
		case CMD_UTILITY: DEBUG("UTILITY "); break;
		case CMD_NOTHING: DEBUG("NOTHING "); break;
	}

	/* Print targetList */
	foreach(cell, q->targetList) {
		TargetEntry *te = lfirst(cell);

		if(cell != list_head(q->targetList))
			DEBUG(", ");

		orion_debug_expr_pprint(te->expr, q->rtable);
		if(te->resname)
			DEBUG(" AS %s", te->resname);
		if(te->resjunk)
			DEBUG("/*junk*/");
	}

	/* Print intoClause */

	/* Print resultRelation */
	if(q->resultRelation) {
		switch(q->commandType) {
			case CMD_DELETE: DEBUG(" FROM "); break;
			case CMD_INSERT: DEBUG(" INTO "); break;
			default: break;
		}
		orion_debug_rte_pprint(
			(RangeTblEntry*) list_nth(q->rtable, q->resultRelation-1));
	}

	/* Print range table */
	foreach(cell, q->rtable) {

		RangeTblEntry *rte = (RangeTblEntry *) lfirst(cell);

		if(list_head(q->rtable) != cell)
			DEBUG(", ");
		else
			DEBUG(" FROM ");

		orion_debug_rte_pprint(rte);
	}

	if(q->jointree && q->jointree->quals) {
		DEBUG(" WHERE ");
		orion_debug_expr_pprint((Expr*) q->jointree->quals, q->rtable);
	}

	if(q->limitCount) {
		DEBUG(" LIMIT ");
		orion_debug_expr_pprint((Expr*) q->limitCount, q->rtable);
	}

	if(q->limitOffset) {
		DEBUG(" OFFSET ");
		orion_debug_expr_pprint((Expr*) q->limitOffset, q->rtable);
	}
}

void orion_debug_query_pprint(char *msg, Query *q) {
//		DEBUG("\n");
//			orion_debug_pprint(msg, q);
//	DEBUG("\n");
	DEBUG("%s\n", msg);

	orion_debug_query_pprint_plain(q);
	DEBUG(";\n");
}
