/*
 * u_floor.c - implements the "floor" pdf operation
 * 
 * Orion Project, http://orion.cs.purdue.edu/
 * Copyright (c) 2008 Purdue University
 */

#include "postgres.h"
#include "utils/array.h"
#include "fmgr.h"
#include "uncertain.h"
#include "debug.h"
#include "errors.h"


/* Builds a map from logical dimension numbers to physical dimension numbers */
static void build_map(Uncertain *in, uint8 *map) {
	uint8 pdf_dimbase[in->npdf];
	int i;
	
	/* Build a map from (pdfnum,0) -> physical dim */
	pdf_dimbase[0] = 0;
	for(i=1; i < in->npdf; ++i)
		pdf_dimbase[i] = pdf_dimbase[i-1] + in->pdfs[i-1]->ndim;
	
	/* Now (pdfnum, i) -> dimbase(pdfnum) + i */
	for(i=0; i < in->nldim; ++i)
		map[i] = pdf_dimbase[in->diminfo[i].pdfnum] + in->diminfo[i].pdfdim;
	
	/* Constant goes after last physical dimension */
	map[in->nldim] = in->npdim;
}

/* Returns false if PDF is completely floored, true otherwise */
/* TODO: Handle predicate types */
static bool floor_pred_truncate(Uncertain *in, float4 *preds, uint8 *types, int *npred) {
	float4 upper[in->nldim];
	float4 lower[in->nldim];
	enum	 {partial, none, full} effect[*npred];
	int32	 i, dim, newpreds=0;
	float4 *pred_temp = preds;
	uint8  *types_temp = types;

	for(i=0; i < in->nldim; ++i) {
		upper[i] = PDF_upper(in->pdfs[in->diminfo[i].pdfnum], in->diminfo[i].pdfdim);
		lower[i] = PDF_lower(in->pdfs[in->diminfo[i].pdfnum], in->diminfo[i].pdfdim);
	}

	for(i=0; i < *npred; ++i) {
		float4 min = 0., max = 0., thresh = pred_temp[in->nldim];

		for(dim=0; dim < in->nldim; ++dim) {
			if(pred_temp[dim] < 0.) {
				min += upper[dim] * pred_temp[dim];
				max += lower[dim] * pred_temp[dim];
			} else if(pred_temp[dim] > 0.) {
				max += upper[dim] * pred_temp[dim];
				min += lower[dim] * pred_temp[dim];
			}
		}


#define TEST_INEQ(OP) { \
		bool min_t = (min OP thresh); \
		bool max_t = (max OP thresh); \
		effect[i] = (min_t == max_t)? (min_t? none : full) : partial; \
		break; \
		}
		
		switch(types[i]) {
		case U_LT:  TEST_INEQ(<);
		case U_LTE: TEST_INEQ(<=);
		case U_GT:  TEST_INEQ(>);
		case U_GTE: TEST_INEQ(>=);
		case U_EQ:
			if(thresh > max || thresh < min)
				effect[i] = full;
			else if(max == min && thresh == max)
				effect[i] = none;
			else
				effect[i] = partial;
			break;

		case U_NEQ:
			if(thresh > max || thresh < min)
				effect[i] = none;
			else if(max == min && thresh == max)
				effect[i] = full;
			else
				effect[i] = partial;
			break;
		default:
			err_bad_case("u_floor", "floor predicate type", types[i]);
		}

		if(effect[i] == none) { /* Nothing floored by this disjunction */
			*npred = 0;
			return TRUE;
		} else if(effect[i] == partial) {
			newpreds ++;
		}

		pred_temp += in->nldim+1;
	}

	/* Copy only predicates that cause partial flooring */
	pred_temp = preds;
	types_temp = types;
	for(i=0; i < *npred; ++i, preds += in->nldim+1, ++types) {
		if(effect[i] == partial) {		/* Keep this predicate */
			if(pred_temp != preds)
				memcpy(pred_temp, preds, sizeof(float4) * (in->nldim+1));
			pred_temp += in->nldim + 1;
			*types_temp++ = *types;
		}
	}

	*npred = newpreds;

	return newpreds > 0;
}

static Floors *floor_floors(Uncertain *in, float4 *preds, uint8 *types, int npred) {
	unsigned int inconj, i=0, j;
	Floors		*fin,		*fout;				/* In/out Floors structs */
	int			inpred,		onpred;				/* In/out pred count */
	float4 		*ipreds,	*opreds;			/* In/out predicates */
	uint8		*itypes, *otypes;				/* In/out predicate types */
	uint8		l2p[in->nldim+1];				/* Logical->Physical map */
	
	build_map(in, l2p);

	fin = in->floors;
	inpred = FLOOR_NPRED(fin);
	inconj = fin? fin->nconj : 0;
	onpred = npred + inpred;

	fout = palloc0(FLOOR_SIZE(inconj+1, onpred, in->npdim));

	fout->nconj = inconj+1;						/* Must add one conj */

	memcpy(fout->conjs, fin->conjs,				/* Copy old conjs[] */
		sizeof(fout->conjs)*inconj);
	fout->conjs[inconj] = inpred;				/* New conj base is old npred */
	fout->conjs[inconj+1] = onpred;				/* Set npred */
	
	ipreds = FLOOR_PREDS(fin);					/* Note this may be NULL */
	opreds = FLOOR_PREDS(fout);

	itypes = FLOOR_TYPES(fin);				/* Note that this might also be NULL */
	otypes = FLOOR_TYPES(fout);

	memcpy(opreds, ipreds,						/* Copy old predicates */
		sizeof(float4)*(in->npdim+1)*inpred);

	memcpy(otypes, itypes, sizeof(uint8)*inpred);   /* Copy old types */

	opreds += (in->npdim+1)*inpred;
	otypes += inpred;
	for(i=0; i < npred; ++i){					/* Copy new floors */
		for(j=0; j < in->nldim+1; ++j) 			/* nldim+1 b/c of constant */
			opreds[l2p[j]] += *preds++;
		opreds += in->npdim+1;					/* Advance to next pred */
		*otypes++ = *types++;
	}
	
	DEBUGASSERT(opreds == &FLOOR_PREDS(fout)[(in->npdim+1)*onpred]);
	return fout;
}

Uncertain *c_floor(Uncertain *in, float4 *preds, uint8 *types, int npred) {
	unsigned int size;
	Uncertain *out;
	bool is_zero;
	float4 *orig_preds = preds;
	uint8 *orig_types = types;

	if(npred == 0)
		return in;

	/* Copy preds and types so that they can be mangled by floor_pred_truncate */
	size = sizeof(float4) * npred * (in->nldim+1);
	preds = palloc(size);
	memcpy(preds, orig_preds, size);

	size = sizeof(uint8) * npred;
	types = palloc(size);
	memcpy(types, orig_types, size);

	/* Filter out floors that have no effect */
	//rohitj is_zero = !floor_pred_truncate(in, preds, types, &npred);
	
	is_zero = (1 < 0); 

	if(is_zero) {
//		DEBUG("is_zero");
	}
	else {
//		DEBUG("is_zero = false");
	}

	/* Allocate and copy data to new Uncertain structure */
	size= sizeof(Uncertain)+sizeof(void*)*(in->npdf-1);
	out = (Uncertain*) palloc(size);
	memcpy(out, in, size);

	if(npred > 0) {
//		DEBUG("npred > 0");
		out->floors = floor_floors(in, preds, types, npred);
	} else {
		out->floors = NULL;

		if(is_zero) {
			int i;
			for(i=0; i < out->npdf; ++i) {
				out->pdfs[i] = zero_init(in->pdfs[i]->ndim);
				out->pdfs[i]->history_id = in->pdfs[i]->history_id;
			}
		}
	}

//	DEBUG("after c_floor : %s", c_print(out));
	return out;
}


/*---------------------------------------------------------------------------*\
  Postgres V1 Function Prototypes
\*---------------------------------------------------------------------------*/

Datum u_floor(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(u_floor);

Datum u_floor_ex(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(u_floor_ex);

/*---------------------------------------------------------------------------*\
  SQL Interface (Wrappers)
\*---------------------------------------------------------------------------*/

Datum u_floor(PG_FUNCTION_ARGS)
{
	Uncertain *u = PG_GETARG_UNCERTAIN(0);
	ArrayType *arr = PG_GETARG_ARRAYTYPE_P(1);
	float4 *preds = (float4 *) ARR_DATA_PTR(arr);
	uint8  *types;
	int npred = ARR_DIMS(arr)[0];

	if(ARR_NDIM(arr) != 2)
		err_arg_value("u_floor", "floor array dimension", ARR_NDIM(arr));
	
	if(u->nldim+1 != ARR_DIMS(arr)[1])
		err_arg_value("u_floor", "floor array columns", ARR_DIMS(arr)[1]);

	types = palloc(npred);
	memset(types, U_GT, npred);

	PG_RETURN_UNCERTAIN(c_floor(u, preds, types, npred));
}

Datum u_floor_ex(PG_FUNCTION_ARGS)
{
	Uncertain *u = PG_GETARG_UNCERTAIN(0);
	ArrayType *preds_arr = PG_GETARG_ARRAYTYPE_P(2);
	ArrayType *types_arr = PG_GETARG_ARRAYTYPE_P(1);
	float4 *preds = (float4 *) ARR_DATA_PTR(preds_arr);
	uint16 *types2 = (uint16 *) ARR_DATA_PTR(types_arr);
	uint8 *types1;
	int i;

	int npred = ARR_DIMS(preds_arr)[0];

	if(ARR_NDIM(preds_arr) != 2)
		err_arg_value("u_floor", "floor array dimension", ARR_NDIM(preds_arr));
	
	if(u->nldim+1 != ARR_DIMS(preds_arr)[1])
		err_arg_value("u_floor", "floor array columns", ARR_DIMS(preds_arr)[1]);

	if(ARR_NDIM(types_arr) != 1)
		err_arg_value("u_floor", "types array dimensions", ARR_NDIM(types_arr));
	
	if(npred != ARR_DIMS(types_arr)[0])
		err_arg_value("u_floor", "elements in types does not equal columns of floor array", 0);

	types1 = palloc(npred);
	for(i=0; i < npred; ++i) {
		if(types2[i] > 5)
			err_arg_value("u_floor", "predicate type", types2[i]);
		types1[i] = types2[i];
	}
	
	PG_RETURN_UNCERTAIN(c_floor(u, preds, types1, npred));
}
