/*
 * u_margin.c - implements the "marginalize" pdf operation
 * 
 * Orion Project, http://orion.cs.purdue.edu/
 * Copyright (c) 2008 Purdue University
 */

#include "postgres.h"
#include "fmgr.h"
#include "utils/array.h"
#include "uncertain.h"
#include "debug.h"

static DimInfo * diminfo_margin(Uncertain *u, int2 *dims, int len, uint8 *map,
                                uint8 maplen) {
	DimInfo *dst, *src;
	int i;
	uint8 invmap[u->npdf];
	
	dst = palloc(len * sizeof(DimInfo));
	src = u->diminfo;

	for(i=0; i < maplen; ++i)
		invmap[map[i]] = i;
	
	for(i=0; i < len; ++i) {
		memcpy(&dst[i], &src[dims[i]], sizeof(DimInfo));
		dst[i].pdfnum = invmap[dst[i].pdfnum];
	}
	
	return dst;
}

static int find_removable_pdfs(Uncertain *src, int2 *dims, int len,
                                uint8 *map) {
	bool needed[src->npdf];
	int i, num_needed = 0;

	memset(needed, 0, sizeof(needed));
	for(i = 0; i < len; ++i)
		needed[src->diminfo[dims[i]].pdfnum] = true;

	if (src->floors) { /* Check floors */
		float4 *preds = FLOOR_PREDS(src->floors);
		const int npred = FLOOR_NPRED(src->floors);
		int pred;
		uint8 pdfnum[src->npdim];
		uncertain_dim2pdf(src, pdfnum);

		for(pred = 0; pred < npred; ++pred) {
			int dim;
			for(dim = 0; dim < src->npdim; ++dim) {
				if (*preds++) needed[pdfnum[dim]] = true;
			}
			preds++;
		}
	}

	for(i = 0; i < src->npdf; ++i) {
		if(!needed[i] && PDF_mass(src->pdfs[i]) < 1) needed[i] = true;

		if(needed[i]) {
			*map++ = i;
			num_needed++;
		}
	}

	return num_needed;
}

static Floors *marginalize_floors(Uncertain *src, Uncertain *dst,uint8 *pdfmap){
	Floors *fin = src->floors, *fout;
	float4 *predsin = FLOOR_PREDS(fin), *predsout;
	const int npred = FLOOR_NPRED(fin);
	int i, j, dim = 0;
	uint8 dim_map[dst->npdim];
	uint8 src_pdfbase[src->npdf];

	fout = (Floors*) palloc(FLOOR_SIZE(FLOOR_NCONJ(fin), FLOOR_NPRED(fin),
	                                   dst->npdim));
	fout->nconj = fin->nconj;
	memcpy(fout->conjs, fin->conjs, sizeof(*fin->conjs) * (fin->nconj+1));
	memcpy(FLOOR_TYPES(fout), FLOOR_TYPES(fin), sizeof(uint8) * npred);
	predsout = FLOOR_PREDS(fout);

	/* Build map from input dimensions to output dimensions */
	uncertain_pdfbase(src, src_pdfbase);
	for(i = 0; i < dst->npdf; ++i)
		for(j = 0; j < dst->npdf; ++j, ++dim)
			dim_map[dim] = src_pdfbase[pdfmap[i]] + j;

	/* Copy coefficients */
	for(i = 0; i < npred; ++i) {
		for(dim = 0; dim < dst->npdim; ++dim) {
			*predsout++ = predsin[dim_map[dim]];
		}
		*predsout++ = predsin[src->npdim];
		predsin += src->npdim + 1;
	}

	return fout;
}

Uncertain *c_margin(Uncertain *src, int2 *dims, int len) {
//	DEBUG("inside c_margin");
	Uncertain *dst;
	int size, out_npdf;
	uint8 map[src->npdf];

	/* Check for identity mapping; do nothing if found */
	if(len == src->nldim) {
		int i;
		for(i = 0; i < len; ++i)
		  if(dims[i] != i)
			  break;
		if (i == len)
		  return src;
	}

	out_npdf = find_removable_pdfs(src, dims, len, map);
	size = sizeof(Uncertain) + sizeof(void*)*(out_npdf-1);
	dst = palloc(size);
	memcpy(dst, src, size);

	dst->nldim = len;
	dst->npdf = out_npdf;
	dst->diminfo = diminfo_margin(src, dims, len, map, out_npdf);

	if (dst->npdf != src->npdf) {
		int pdf;
		dst->npdim = 0;
		for (pdf = 0; pdf < out_npdf; ++pdf) {  /* Fixup pdf pointers */
			dst->pdfs[pdf] = src->pdfs[map[pdf]];
			dst->npdim += dst->pdfs[pdf]->ndim;		/* and recompute npdim */
		}
		if (src->floors)
			dst->floors = marginalize_floors(src, dst, map);
	}
	
	return dst;
}


/*---------------------------------------------------------------------------*\
  Postgres V1 Function Prototypes
\*---------------------------------------------------------------------------*/

Datum u_margin(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(u_margin);

/*---------------------------------------------------------------------------*\
  SQL Interface (Wrappers)
\*---------------------------------------------------------------------------*/


Datum u_margin(PG_FUNCTION_ARGS)
{
	Uncertain *u = PG_GETARG_UNCERTAIN(0);
	ArrayType *arr = PG_GETARG_ARRAYTYPE_P(1);
	int2 *dims = (int2 *) ARR_DATA_PTR(arr);
	int len = ARR_DIMS(arr)[0];

	DEBUGASSERT(ARR_NDIM(arr) == 1);
	
	PG_RETURN_UNCERTAIN(c_margin(u, dims, len));
}
