/*
 * u_product.c - implements the "product" pdf operation
 * 
 * Orion Project, http://orion.cs.purdue.edu/
 * Copyright (c) 2008 Purdue University
 */

#include "postgres.h"
#include "fmgr.h"
#include "uncertain.h"
#include "debug.h"

static uint8 *build_dim_map(Uncertain *src, Uncertain *dst, uint8 *pdfmap) {
	uint8 *dimmap = palloc(src->npdim+1);
	uint8 dimbase[dst->npdf];
	int i;
	int pdfnum = 0;
	int dim = 0;

	/* Determine the dimension number for dimension 0 of each pdf in dst */
	dimbase[0] = 0;
	for(i=1; i < dst->npdf; ++i)
		dimbase[i] = dimbase[i-1] + dst->pdfs[i-1]->ndim;

	/* Loop through physical dimensions of src, mapping them to (src_pdf,dim) 
	 * pairs that are then mapped to (dst_pdf,dim) and finally to dst_pdim */
	for(i=0; i < src->npdim; ++i) {
		dimmap[i] = dimbase[pdfmap[pdfnum]] + dim;
		if(++dim == src->pdfs[pdfnum]->ndim) {
			dim = 0;
			++pdfnum;
		}
	}
	
	dimmap[src->npdim] = dst->npdim;	/* map the constant dimension */
	
	return dimmap;
}


static void floor_product_partial(Floors * fout, uint8 npdim, Uncertain *src, uint32 conjbase, uint32 predbase, uint8 *map) {
	Floors  *fsrc = src->floors;
	int     src_npred = FLOOR_NPRED(fsrc);
	float4  *opred = FLOOR_PREDS(fout),
	        *spred = FLOOR_PREDS(fsrc);
	uint8   *otypes = FLOOR_TYPES(fout),
	        *stypes = FLOOR_TYPES(fsrc);
	int conj, pred, dim;

	for(conj=0; conj < FLOOR_NCONJ(fsrc); ++conj)
		fout->conjs[conj+conjbase] = fsrc->conjs[conj]+predbase;

	opred += (npdim + 1) * predbase;
	otypes += predbase;
	for(pred=0; pred < src_npred; ++pred, opred += npdim+1) {
		*otypes++ = *stypes++;
		for(dim=0; dim < src->npdim+1; ++dim)
			opred[map[dim]] = *spred++;
	}
}

static DimInfo * dim_product(Uncertain *a, uint8 *amap, Uncertain *b, uint8 *bmap, int nldim) {
	DimInfo *out; 
	int i;
	
	out = palloc(sizeof(DimInfo) * nldim);
	
	memcpy(out, a->diminfo, sizeof(DimInfo) * a->nldim);
	memcpy(out+a->nldim, b->diminfo, sizeof(DimInfo) * b->nldim);
	
	/* Fixup pdfnums */
	for(i=0; i < a->nldim; ++i)
		out[i].pdfnum = amap[a->diminfo[i].pdfnum];
	
	for(i=0; i < b->nldim; ++i)
		out[i+a->nldim].pdfnum = bmap[b->diminfo[i].pdfnum];
	
	return out;
}

static Floors *floor_product(Uncertain *a, uint8 *amap, Uncertain *b, uint8 *bmap, Uncertain *c) {
	Floors  *fa = a->floors, 
			*fb = b->floors,
			*fout;
	uint32  nconj = FLOOR_NCONJ(fa) + FLOOR_NCONJ(fb);
	uint32  npred = FLOOR_NPRED(fa) + FLOOR_NPRED(fb);
	uint8 *dimmap;

	if(npred == 0)
		return NULL;

	fout = palloc0(FLOOR_SIZE(nconj, npred, c->npdim));
	fout->nconj = nconj;
	fout->conjs[nconj] = npred;

	dimmap = build_dim_map(a, c, amap);
	floor_product_partial(fout, c->npdim, a, 0, 0, dimmap);
	pfree(dimmap);

	dimmap = build_dim_map(b, c, bmap);
	floor_product_partial(fout, c->npdim, b, FLOOR_NCONJ(fa), FLOOR_NPRED(fa), dimmap);
	pfree(dimmap);

	return fout;
}

static void pdfdata_product(Uncertain *a, uint8 *amap, Uncertain *b, uint8 *bmap, PDF **pdfs) {
	int i;

	for(i=0; i < a->npdf; ++i)
		pdfs[amap[i]] = a->pdfs[i];
	
	for(i=0; i < b->npdf; ++i)
		if(pdfs[bmap[i]] == NULL || pdfs[bmap[i]]->vtbl->id != '0')
			pdfs[bmap[i]] = b->pdfs[i];
}

static int npdim_product(Uncertain *c) {
	int npdim, i;
	
	npdim = 0;
	for(i=0; i < c->npdf; ++i)
		npdim += c->pdfs[i]->ndim;
	
	return npdim;
}

#define cmp(a,b) (((a)==(b))? 0: ((a)>(b))? 1:-1)
static inline int history_compare(PDF *a, PDF *b) {	
	return cmp(a->history_id, b->history_id);
}

/* Create the mapping from input PDFs in a and b and the output PDFs in c;
 * maintain the PDFs in order, sorted by history, to ease future products */
static void build_pdf_maps(Uncertain *a, uint8 *amap, Uncertain *b, uint8 *bmap) {
	int ai = 0;
	int bi = 0;
	int i = 0;

	for(; ai < a->npdf && bi < b->npdf; ++i) {
		int c = history_compare(a->pdfs[ai], b->pdfs[bi]);

		if(c == 0 || c == -1)   /* a <= b, Map a */
			amap[ai++] = i;

		if(c == 0 || c == 1)   /* a >= b, Map b */
			bmap[bi++] = i;
	}

	for(; ai < a->npdf; ++ai, ++i)
		amap[ai] = i;
	
	for(; bi < b->npdf; ++bi, ++i)
		bmap[bi] = i;
}
	
Uncertain *c_product(Uncertain *a, Uncertain *b) {
	Uncertain *c;
	uint8 amap[a->npdf], bmap[b->npdf];
	uint8 npdf;

	build_pdf_maps(a, amap, b, bmap);
	npdf = Max(amap[a->npdf-1], bmap[b->npdf-1]) + 1;

	c = (Uncertain *) palloc0(sizeof(Uncertain) + sizeof(void*)*(npdf-1)); 
	
	c->nldim = a->nldim + b->nldim;
	pdfdata_product(a, amap, b, bmap, c->pdfs);
	/* npdim set below */
	c->npdf = npdf;
	c->diminfo = dim_product(a, amap, b, bmap, c->nldim);
	c->npdim = npdim_product(c);
	c->floors = floor_product(a, amap, b, bmap, c);
	
	return c;
}
        

/*---------------------------------------------------------------------------*\
  Postgres V1 Function Prototypes
\*---------------------------------------------------------------------------*/

Datum u_product(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(u_product);

/*---------------------------------------------------------------------------*\
  SQL Interface (Wrappers)
\*---------------------------------------------------------------------------*/


Datum u_product(PG_FUNCTION_ARGS)
{
	Uncertain *args[] = {PG_GETARG_UNCERTAIN(0), PG_GETARG_UNCERTAIN(1)};

        PG_RETURN_UNCERTAIN( c_product(args[0], args[1]) );
}
