/*
 * uncertain.c - internal operations on uncertain data types
 * 
 * Orion Project, http://orion.cs.purdue.edu/
 * Copyright (c) 2008 Purdue University
 */

#include "postgres.h"
#include "fmgr.h"
#include "uncertain.h"
#include "errors.h"
#include "debug.h"
#include "executor/spi.h"
#include "utils/snapmgr.h"

/*
 * Sets the varlena header; required as of PostgreSQL 8.3.
 */
#ifndef SET_VARSIZE
#define SET_VARSIZE(PTR, len) PTR->vl_len_ = len
#endif

/*
 * As of 8.4, we need to declare our own.
 */
Snapshot ActiveSnapshot = InvalidSnapshot;

/*---------------------------------------------------------------------------*\
  Size Functions
\*---------------------------------------------------------------------------*/

/*
 * Gets the total size of the given floors, in bytes.
 */
uint32 floors_size(Uncertain *p)
{
	Floors *f = p->floors;

	return f==NULL? 0 : FLOOR_SIZE(f->nconj, FLOOR_NPRED(f), p->npdim);
}

/*---------------------------------------------------------------------------*\
  Serialization
\*---------------------------------------------------------------------------*/

/*
 * Serializes the given uncertain data for subsequent storage on disk.
 */
UncHead *deflate(Uncertain *data)
{
	int32 i;
	UncHead *head;
	char *pos;
	int32 len;

	/* calculate the total varlena size */
	int32 size = sizeof(UncHead);
	int32 d_len = data->nldim * sizeof(DimInfo);
	int32 f_len = floors_size(data);
	size += d_len;
	size += f_len;
	for (i = 0; i < data->npdf; i++)
	{
		size += data->pdfs[i]->size;
	}

	/* allocate buffer and initialize header */
	head = palloc0(size);
	SET_VARSIZE(head, size);
	head->nldim = data->nldim;
	head->npdim = data->npdim;
	head->npdf = data->npdf;
	head->f_len = f_len;

	/* copy each structure into the buffer */
	pos = (char *) head + sizeof(UncHead);
	memcpy(pos, data->diminfo, d_len);
	pos += d_len;
	if (data->floors != NULL)
	{
		memcpy(pos, data->floors, f_len);
		pos += f_len;
	}

	for (i = 0; i < data->npdf; i++)
	{
		len = data->pdfs[i]->size;
		memcpy(pos, data->pdfs[i], len);
		memset(pos, 0, sizeof(data->pdfs[i]->vtbl));
		*pos = data->pdfs[i]->vtbl->id; /* persist the id, not the address */
		pos += len;
	}

	/* return the resulting structure */
	return head;

} /* deflate */

/*
 * Deserializes the recently detoasted uncertain data for in-memory use.
 */
Uncertain *inflate(UncHead *head)
{
	int32 size;
	Uncertain *data;
	char *pos;
	int32 i;

	/* calculate the size of the in-memory structure */
	size = sizeof(Uncertain);
	size += (head->npdf - 1) * sizeof(void *);

	/* allocate and initialize the wrapper */
	data = palloc0(size);
	data->nldim = head->nldim;
	data->npdim = head->npdim;
	data->npdf = head->npdf;

	/* point back into the original buffer */
	pos = (char *) head + sizeof(UncHead);
	data->diminfo = (DimInfo *) pos;
	pos += head->nldim * sizeof(DimInfo);

	/* set the optional floors pointer */
	if (head->f_len > 0)
	{
		data->floors = (Floors *) pos;
		pos += head->f_len;
	}

	/* set the array of pdf pointers */
	for (i = 0; i < head->npdf; i++)
	{
		data->pdfs[i] = (PDF *) pos;
		if (pos[1] == 0 && pos[2] == 0 && pos[3] == 0)  /* Ensure PDF is deflated */
			data->pdfs[i]->vtbl = pdf_id_to_vtbl(*pos);
		pos += data->pdfs[i]->size;
	}

	/* return the resulting structure */
	return data;

} /* inflate */

/*---------------------------------------------------------------------------*\
  Constructor
\*---------------------------------------------------------------------------*/

/*
 * Allocates an Uncertain wrapper and initializes its header.
 */
Uncertain *initUncertain(uint8 ndim, uint8 npdf)
{
	int32 size;
	Uncertain *data;

	/* calculate the variable-length size */
	size = sizeof(Uncertain);
	size += (npdf - 1) * sizeof(void *);

	/* allocate and initialize the wrapper */
	data = palloc0(size);
	data->nldim = ndim;
	data->npdim = ndim;
	data->npdf = npdf;

	/* allocate supporting structures */
	data->diminfo = palloc0(ndim * sizeof(DimInfo));

	/* return the resulting structure */
	return data;

} /* initUncertain */

/*---------------------------------------------------------------------------*\
  Utility Functions
\*---------------------------------------------------------------------------*/

/*
 * Counts the number of dimensions stored in a specific pdf.
 */
uint8 dimCount(Uncertain *unc, uint8 pdfnum)
{
	int32 i;
	uint8 cnt = 0;
	for (i = 0; i < unc->nldim; i++)
	{
		if (unc->diminfo[i].pdfnum == pdfnum)
		{
			cnt++;
		}
	}
	return cnt;
}

/*
 * Returns an array of flags that specifies which dimensions to display as
 * floats (where true) and which to display as integers (where false).
 */
bool *uncFlags(Uncertain *unc)
{
	int32 i;
	int32 t;
	bool *floats = palloc(unc->nldim);

	for (i = 0; i < unc->nldim; i++)
	{
		/* numeric PDFs use floats, categorical PDFs use integers */
		t = unc->diminfo[i].dimtype;
		floats[i] = (t == DIM_UCON || t == DIM_UDIS);
	}
	return floats;
}

/*
 * Alternative version of uncFlags that takes the metadata directly.
 */
bool *dimFlags(uint8 ndim, uint8 *dimtypes)
{
	int32 i;
	bool *floats = palloc(ndim);

	for (i = 0; i < ndim; i++)
	{
		floats[i] = (dimtypes[i] == DIM_UCON || dimtypes[i] == DIM_UDIS);
	}
	return floats;
}

/*
 * Returns the external display name of the given uncertain data type.
 */
char *typeName(uint8 ndim, char *dimtypes)
{
	const int32 NAME_LEN = 5; /* 4 chars + comma */
	int32 len;
	char *str;
	char *pos;
	int32 dim;

	/* allocate the output buffer */
	len = NAME_LEN * ndim;
	str = palloc(len);
	pos = str;

	/* print each type name */
	for (dim = 0; dim < ndim; dim++)
	{
		/* get and copy the type name */
		switch (dimtypes[dim])
		{
		case DIM_UCON:
			strcpy(pos, "ucon");
			break;
		case DIM_UDIS:
			strcpy(pos, "udis");
			break;
		case DIM_UORD:
			strcpy(pos, "uord");
			break;
		case DIM_UNOM:
			strcpy(pos, "unom");
			break;
		default:
			err_bad_case("typeName", ERR_SQLTYPE, dimtypes[dim]);
			break;
		}

		/* append a comma and advance position */
		pos[NAME_LEN - 1] = ',';
		pos += NAME_LEN;
	}

	/* terminate and return */
	str[len - 1] = '\0';
	return str;

} /* typeName */

/*
 * Takes an array of length npdim and sets pdfnum[i] to the physical pdf
 * for physical dimension i
 */
void uncertain_dim2pdf(Uncertain *u, uint8 *pdfnum) {
  int pdf;

  for(pdf=0; pdf < u->npdf; ++pdf) {
    memset(pdfnum, pdf, u->pdfs[pdf]->ndim);
    pdfnum += u->pdfs[pdf]->ndim;
  }
}

/* 
 * Takes an array of length npdf and sets pdfbase[i] to the first physical
 * dimension in PDF i
 */
void uncertain_pdfbase(Uncertain *u, uint8 *pdfbase) {
  int pdf;

  pdfbase[0] = 0;
  for(pdf=1; pdf < u->npdf; ++pdf)
    pdfbase[pdf] = u->pdfs[pdf-1]->ndim + pdfbase[pdf-1];
}

/*
 * Returns the next available history identifier.
 */
int32 next_histid(void)
{
	int32 ret;
	bool isnull;
	int32 histid = 0;
	bool destroy_snapshot = (ActiveSnapshot == NULL);

	if(!ActiveSnapshot)
		ActiveSnapshot = GetTransactionSnapshot();
	
	/* connect to the DB and execute the query */
	SPI_connect();
	ret = SPI_execute("SELECT nextval('history_id_seq')::integer", true, 1);

	/* get the query result */
	if (ret == SPI_OK_SELECT)
	{
		TupleDesc tupdesc = SPI_tuptable->tupdesc;
		HeapTuple tuple = SPI_tuptable->vals[0];
		histid = DatumGetInt32( SPI_getbinval(tuple, tupdesc, 1, &isnull) );
	}

	/* close the DB connection */
	SPI_finish();

	if(destroy_snapshot)
		ActiveSnapshot = NULL;
	return histid;
} /* next_histid */
