/*
 * uncertain.h - internal representation of uncertain data
 * 
 * All uncertain types share the same generic data structure internally, which
 * not only can store univariate PDFs, but also handles arbitrary multivariate
 * distributions of compound types (i.e. multiple sql columns):
 * 
 *   <header>   - serialization header
 *   <dim info> - headers for each dimension
 *   <pdf info> - headers for each pdf
 *   <floors>   - optional floor data
 *   <history>  - optional history data
 *   <pdfs>     - actual pdf structures
 * 
 * This entire structure ultimately represents a single joint distribution.  If
 * multiple pdf structures are present, they are part of an implicit product
 * operation.  Note that the <dim info> and <pdf info> arrays have lengths ndim
 * and npdf, respectively.
 * 
 * Since many data members are variable-legnth, this serialized form stored on
 * disk is automatically inflated into a more convenient in-memory structure
 * for general use.
 * 
 * Orion Project, http://orion.cs.purdue.edu/
 * Copyright (c) 2008 Purdue University
 */

#include "pdfs/pdf.h"
#include "utils/snapshot.h"

#ifndef UNCERTAIN_H_
#define UNCERTAIN_H_

/*---------------------------------------------------------------------------*\
  Constants & Macros
\*---------------------------------------------------------------------------*/

/*
 * Unique type ids for each data type.
 */
#define DIM_UCON ((uint8) 'c')
#define DIM_UDIS ((uint8) 'd')
#define DIM_UORD ((uint8) 'o')
#define DIM_UNOM ((uint8) 'n')

/*
 * Postgres V1 Function Macros (that detoast and marshall automatically)
 */
#define PG_GETARG_UNCERTAIN(n) inflate((UncHead *) PG_GETARG_VARLENA_P(n))
#define PG_RETURN_UNCERTAIN(x) PG_RETURN_POINTER(deflate(x))

extern Snapshot ActiveSnapshot;

/*---------------------------------------------------------------------------*\
  Serialized Structures
\*---------------------------------------------------------------------------*/

/*
 * Header for storing uncertain data on disk in serialized form.
 */
typedef struct UncHead
{
	/* varlena header (do not touch directly!) */
	int32 vl_len_;

	/* number of dimensions (logical number of attributes) */
	unsigned nldim : 8;

	/* number of visible dimensions */
	unsigned npdim : 8;

	/* size of the optional floor data */
	unsigned f_len : 24;

	/* number of pdf structures (in implicit lazy product) */
	unsigned npdf : 8;

} UncHead;

/*
 * Packed header structure for each dimension.
 */
typedef struct DimInfo
{
	/* external data type of the dimension */
	uint8 dimtype;

	/* index of the pdf storing the dimension */
	uint8 pdfnum;

	/* index of the dimension within the pdf */
	uint8 pdfdim;

} DimInfo;

#define FLOOR_NCONJ(floor) ((floor)? (floor)->nconj : 0)
#define FLOOR_NPRED(floor) ((floor)? (floor)->conjs[(floor)->nconj]:0)
#define FLOOR_SIZE(nconj, npred, npdim) (sizeof(Floors)+(nconj)+(1+4*((npdim)+1))*npred)
#define FLOOR_TYPES(floor) ((floor)? (uint8 *) (&(floor)->conjs[(floor)->nconj+1]):NULL)
#define FLOOR_PREDS(floor) ((floor)? (float4 *) (&FLOOR_TYPES(floor)[FLOOR_NPRED(floor)]) : NULL)

#define U_NEQ 0x0
#define U_EQ 0x1
#define U_LT 0x2
#define U_LTE 0x3
#define U_GT 0x4
#define U_GTE 0x5

typedef struct Floors
{
	/* Number of conjunctions */
	uint8   nconj;

	/* Indicies of first predicate in each conjunction */
	uint8 conjs[1];
	
	/* conjs[nconj] = npred, the number of predicates */
	//uint8 npred;
	/* pred_types is a 1D array of size npred that holds predicate types */
	//uint8 pred_types[npred];
	/* preds[] is a 2D array of size (ndim+1)*npred */
	//float4 preds[1];

} Floors;

/*---------------------------------------------------------------------------*\
  In-Memory Version
\*---------------------------------------------------------------------------*/

/*
 * Top-level uncertain data structure.
 */
typedef struct Uncertain
{
	/* number of logical dimensions */
	uint8 nldim;

	/* number of physical dimensions (i.e., \sum_i pdfinfo[i].dimcnt) */
	uint8 npdim;

	/* number of pdf structures (in implicit lazy product) */
	uint8 npdf;

	/* array of dimension headers */
	DimInfo *diminfo;

	/* floored regions */
	Floors *floors;

	/* array of pdf pointers */
	PDF *pdfs[1];

} Uncertain;

/*---------------------------------------------------------------------------*\
  Function Prototypes
\*---------------------------------------------------------------------------*/

uint32 floors_size(Uncertain *u);
uint32 history_size(Uncertain *u);

UncHead *deflate(Uncertain *data);
Uncertain *inflate(UncHead *head);

Uncertain *initUncertain(uint8 ndim, uint8 npdf);
Uncertain *c_reduce(Uncertain *unc, int32 count);
char *c_print(Uncertain *u);

Uncertain *c_product(Uncertain *, Uncertain *);
Uncertain *c_margin(Uncertain *src, int2 *dims, int len);
Uncertain *c_floor(Uncertain *c, float4 *preds, uint8 *types, int npred);

float4 c_lower(Uncertain *unc, int32 dim);
float4 c_upper(Uncertain *unc, int32 dim);
float4 c_prob(Uncertain *unc, float4 *minv, float4 *maxv);
float4 c_mass(Uncertain *u);

uint8 dimCount(Uncertain *unc, uint8 pdfnum);
bool *uncFlags(Uncertain *unc);
bool *dimFlags(uint8 ndim, uint8 *dimtypes);
char *typeName(uint8 ndim, char *dimtypes);

/* Fill pdfnum[i] with the physical pdf for physical dimension i */
void uncertain_dim2pdf(Uncertain *u, uint8 *pdfnum); 
void uncertain_pdfbase(Uncertain *u, uint8 *pdfbase);

int32 next_histid(void);

#endif /* UNCERTAIN_H_ */
