/*
 * uncertain_api.c - SQL functions and operators for uncertain data types
 * 
 * TODO: extend API to handle multiple pdfs
 * 
 * Orion Project, http://orion.cs.purdue.edu/
 * Copyright (c) 2008 Purdue University
 */

#include "postgres.h"
#include "fmgr.h"
#include "utils/array.h"
#include "uncertain.h"
#include "pdfs/pdf.h"
#include "errors.h"
#include "debug.h"
#include <math.h>

static Uncertain *c_sample(Uncertain *, int);
static bool c_zero(Uncertain *);
/*---------------------------------------------------------------------------*\
  Inquiry Functions
\*---------------------------------------------------------------------------*/

/*
 * Gets the sampling lower bound of the given PDF.
 */
float4 c_lower(Uncertain *unc, int32 dim)
{
	DimInfo *d = &unc->diminfo[dim];
	return PDF_lower(unc->pdfs[d->pdfnum], d->pdfdim);
}

/*
 * Gets the sampling upper bound of the given PDF.
 */
float4 c_upper(Uncertain *unc, int32 dim)
{
	DimInfo *d = &unc->diminfo[dim];
	return PDF_upper(unc->pdfs[d->pdfnum], d->pdfdim);
}

/*
 * Given a point in logical space, converts it into a point in physical space.
 * If some dimension is assigned two values, takes the minimum for the
 * coordinate and returns true. Otherwise, returns false.
 *
 * l[nldim] = logical coordinates (input)
 * p[npdim] = physical coordinates (npdim) (output)
 * t[npdim] = false if dimension was marginalized, type of dimension otherwise
 * returns true if there was a conflict during conversion
 */
static bool point_realspace(Uncertain *u, float4 l[], float4 p[], uint8 t[]) {
	int i;
	uint8 pdfbase[u->npdf];
	bool conflict=false;

	memset(t, 0, sizeof(uint8) * u->npdim);
	uncertain_pdfbase(u, pdfbase);

	for(i=0; i < u->nldim; ++i) {
		int d = pdfbase[u->diminfo[i].pdfnum] + u->diminfo[i].pdfdim;

		if(t[d]) { /* Conflict? */
			conflict = fabs(p[d] - l[i]) > TOLERANCE;
			p[d] = Min(l[i], p[d]);
		} else {
			t[d] = u->diminfo[i].dimtype;
			p[d] = l[i];
		}
	}

	return conflict;
}

/*
 * Find the cumulative probability of the PDF at the given point
 */
static float4 c_cdf(Uncertain *u, float4 *lpnt) {
	if(c_zero(u))            /* Handle trivial case */
		return 0.;
	
	if(u->floors)            /* Attempt to remove floors without product */
	  u = c_sample(u, 10);
		
	if(u->floors) {          /* Resort to product if there are still floors */
		u = c_reduce(u, 10);
		return PDF_cdf(u->pdfs[0], lpnt);

	} else {                 /* Otherwise, multiply CDF in each dimension */
		int i, pdf, pdfdim;
		float4 ppnt[u->npdim];
		uint8  type[u->npdim];
		float4 p=1., *pnt = ppnt;

		point_realspace(u, lpnt, ppnt, type);

		/* Take upper limit for marginalized dimensions */
		for(i=0, pdf=0, pdfdim=0; i < u->npdim; ++i) {
			if(type[i] == 0)
				ppnt[i] = PDF_upper(u->pdfs[pdf], pdfdim);

			pdfdim = (pdfdim+1) % (u->pdfs[pdf]->ndim);
			if(pdfdim == 0) pdf ++;
		}

		for(i=0; i < u->npdf; ++i) {
			p *= PDF_cdf(u->pdfs[i], pnt);
			pnt += u->pdfs[i]->ndim;  /* Advance past consumed dimensions */
		}

		return p;
	}
}

/*
 * Find the probability density/mass function of the PDF at the given point
 */
static float4 c_pdf(Uncertain *u, float4 *lpnt) {
	float4 ppnt[u->npdim]; /* Point in physical coordinates */
	uint8  type[u->npdim]; /* Type of each dimension (or 0, if marginalized) */

	if(c_zero(u))
		return 0.;           /* Whole PDF is trivially floored */

	if(point_realspace(u, lpnt, ppnt, type))
		return 0.;           /* Conflict during conversion; pdf must be zero */

	u = c_reduce(u, 10);
	return PDF_pdf(u->pdfs[0], lpnt);
}

/*
 * Returns true if the probability of the given PDF is trivially zero
 */
static bool  c_zero(Uncertain *u) {
	int i;

	/* check degenerate condition */
	for (i = 0; i < u->npdf; i++) {
		if(u->pdfs[i]->vtbl->id == '0')
			return true;
	}

	return false;
}

/*
 * Returns the probability mass of the given PDF within the given range.
 */
float4 c_prob(Uncertain *unc, float4 *minv, float4 *maxv)
{
	float4 mincdf, maxcdf;
	
	if(c_zero(unc)) return 0.;

	mincdf = c_cdf(unc, minv);
	maxcdf = c_cdf(unc, maxv);
	return (maxcdf - mincdf);
}

/*
 * Returns the total probability mass of the given PDF.
 */
float4 c_mass(Uncertain *u)
{
//	DEBUG("inside c_mass");
	int32 i;
	float4 p = 1.0;
	
	if(c_zero(u)) return 0.;

	if(u->floors) {          /* Resort to product if there are still floors */
//		DEBUG("floor");
		u = c_reduce(u, 10);
		p = PDF_mass(u->pdfs[0]);
	} else {
//		DEBUG("no floor");
		for(i=0; i < u->npdf; ++i)
			p *= PDF_mass(u->pdfs[i]);
	}
	
	return p;
	/* get the cdf value at the highest point */
/*rohitj	for (i = 0; i < u->nldim; i++)
		pnt[i] = c_upper(u, i);
	
	return c_cdf(u, pnt);
	*/
}

/*---------------------------------------------------------------------------*\
  Sampling Functions
\*---------------------------------------------------------------------------*/

/*
 * Converts a certain value into an uncertain pdf.
 */
static Uncertain *c_real2hist(float4 val)
{
	Uncertain *data;

	/* simply create a histogram over the point */
	data = initUncertain(1, 1);
	data->diminfo[0].dimtype = DIM_UCON;
	data->pdfs[0] = delta_init(val);
	data->pdfs[0]->history_id = next_histid();
	return data;

} /* c_real2hist */

/*
 * Samples or enumerates the given distribution, returning a Histogram (or
 * Discrete) with the desired number of bins (or points).
 */

static int c_sample_find_conj_owner(Uncertain *unc, 
	int npred, float4 *preds, uint8 pdfnum[]) {
	int owner = -1, dim, pred;

	for(pred=0; pred < npred; ++pred) {
		for(dim=0; dim < unc->npdim; ++dim) {
			if(*preds++ != 0.) {
				if(owner != pdfnum[dim] && owner != -1)	/* Multiple owners */
					return -1;
				owner = pdfnum[dim];
			}
		}
		preds++; /* Skip constant */
	}

	return owner;
}

static Uncertain *c_sample(Uncertain *in, int32 count)
{
//	DEBUG("inside c_sample");
	int i, size;
	Uncertain *unc;

	size = sizeof(Uncertain) + sizeof(void*)*(in->npdf-1);
	unc = palloc(size);
	memcpy(unc, in, size);

	for(i=0; i < unc->npdf; ++i)
		unc->pdfs[i] = PDF_sample(unc->pdfs[i], count);

	return unc; //rohitj
	
	
	/* push floors that apply to only one PDF down into the histograms */
	if(unc->floors) {
		uint8  nremain=0, npremain=0;                // # PDF-spanning conjs/preds
		uint8  pdfnum[unc->npdim];                   // PDF owning each p-dimension
		uint8	 pdfbase[unc->npdf];                   // first p-dimension in ea. PDF
		uint8  conj=0;							                 // iterating variable
		float4 *preds = FLOOR_PREDS(unc->floors);    // predicates for current conj
		uint8  *types = FLOOR_TYPES(unc->floors);    // predicate types
		uint32 keep_conj[(unc->floors->nconj+31)/32];// bitmask of PDF-spanning conj

		uncertain_dim2pdf(unc, pdfnum);
		uncertain_pdfbase(unc, pdfbase);
		memset(keep_conj, 0, sizeof(keep_conj));

		for(conj=0;  conj < unc->floors->nconj; ++conj) {
			int owner, conj_preds;
			
			conj_preds = unc->floors->conjs[conj+1]-unc->floors->conjs[conj];
			owner = c_sample_find_conj_owner(unc, conj_preds, preds, pdfnum);

			if(owner == -1) {				/* Owned by more than one PDF */
				keep_conj[conj/32] |= 1<<(conj%32);
				nremain++;
				npremain += conj_preds;
			} else {								/* Owned by just one PDF */
				uint8 i;
				float4 *mypreds[conj_preds];
				float4 mypreddata[conj_preds][unc->pdfs[owner]->ndim+1];
				uint8 pdf_ndim = unc->pdfs[owner]->ndim;

				for(i=0; i < conj_preds; ++i) {	/* Build temp array of floors */
					mypreds[i] = mypreddata[i];
					memcpy(mypreds[i], &preds[(unc->npdim+1)*i], sizeof(float4)*pdf_ndim);
					mypreds[i][pdf_ndim] = preds[(unc->npdim+1)*(i+1)-1];
				}

				histogram_apply_floor((Histogram*) unc->pdfs[owner], conj_preds, types, mypreds);
			}

			preds += conj_preds * (unc->npdim+1); /*Move to base of next conjunction*/
			types += conj_preds;
		}

		/* Fixup floors array if any of the floors were applied */
		if(nremain > 0 && nremain < unc->floors->nconj) {
			int old_nconj = unc->floors->nconj;
			float4 *old_preds, *new_preds;
			uint8 *old_conjs, *new_conjs;
			uint8 *old_types, *new_types;
			Floors *new_floors = palloc(FLOOR_SIZE(nremain, npremain, unc->npdim));

			old_preds = FLOOR_PREDS(unc->floors);
			old_types = FLOOR_TYPES(unc->floors);
			old_conjs = unc->floors->conjs;

			new_floors->nconj  = nremain;
			new_floors->conjs[0] = 0;
			new_preds = FLOOR_PREDS(new_floors);
			new_types = FLOOR_TYPES(new_floors);
			new_conjs = new_floors->conjs;

			for(conj=0; conj < old_nconj; ++conj) {
				int npred_conj = unc->floors->conjs[conj+1]-unc->floors->conjs[conj];
				int size = sizeof(float4)*(unc->npdim+1)*npred_conj;

				if(keep_conj[conj/32] & (1<<(conj%32))) {
					memcpy(new_preds, old_preds, size);
					memcpy(new_types, old_types, npred_conj);
					new_preds += size;
					new_types += npred_conj;
					new_conjs[1] = new_conjs[0]+npred_conj;		/* Set base of next conj */
					++new_conjs;
				}
				old_preds += size;
				old_types += npred_conj;
			}

			unc->floors = new_floors;
		}	else if(nremain==0)			/* If no remaining conjunctions, */
			unc->floors = NULL;			/* Clear floors */
	}
		
	return unc;
} /* c_sample */


/*
 * Evaluates any implicit products and floored regions to produce a single
 * multigram PDF in canonical form.
 */
Uncertain *c_reduce(Uncertain *orig, int32 count)
{
//	DEBUG("inside c_reduce");
	Uncertain *ret, *u;
	Histogram *hist;
	int32 i;
	uint8 dims[orig->nldim];
	uint8 pdfbase[orig->npdf];

	DEBUGASSERT(c_print(orig));
	c_print(orig);
	u = c_sample(orig, count);
	DEBUGASSERT(c_print(u));
	c_print(u);
//	c_print(u);

	/* TODO: Use floors to find partition of PDF set, and product/floor within
	 * each partition before performing a full product/marginalize. */
	
	/* Reduce to one histogram */
	hist = histogram_product((Histogram **) u->pdfs, u->npdf);
//	DEBUG("histogram after product : %s", histogram_print(hist));

	/* Apply floors */
	if(u->floors) {
		float4 *preds = FLOOR_PREDS(u->floors);
		uint8  *types = FLOOR_TYPES(u->floors);
		int	conj=0, pred;

		/* Ensure that we're not mangling the input */
		if((PDF*) hist == orig->pdfs[0]) {
			hist = palloc(orig->pdfs[0]->size);
			memcpy(hist, orig->pdfs[0], orig->pdfs[0]->size);
		}

		for(conj=0; conj < u->floors->nconj; ++conj) {
			int conj_preds = u->floors->conjs[conj+1]-u->floors->conjs[conj];
			float4 *mypreds[conj_preds];

			/* Build temporary floor array */
			for(pred=0; pred < conj_preds; ++pred, preds+= u->npdim + 1)
				mypreds[pred] = preds;

			histogram_apply_floor(hist, conj_preds, types, mypreds);

			types += conj_preds;
		}
	}
//	DEBUG("histogram after flooring:%s", histogram_print(hist));
	/* Marginalize */
	uncertain_pdfbase(u, pdfbase);
	for(i=0; i < u->nldim; ++i)
		dims[i] = pdfbase[u->diminfo[i].pdfnum] + u->diminfo[i].pdfdim;
	hist = histogram_marginalize(hist, dims, u->nldim);
//	DEBUG("histogram after marginalize:%s", histogram_print(hist));
	
	/* initialize the reduced structure */
	ret = initUncertain(u->nldim, 1);
	ret->pdfs[0] = (PDF *) hist;
	for (i = 0; i < u->nldim; i++) {
		ret->diminfo[i].dimtype = u->diminfo[i].dimtype;
		ret->diminfo[i].pdfnum = 0;
		ret->diminfo[i].pdfdim = i;
	}
	
	/* TODO: propagate histories
	 * History cannot be propagated with current data structure design:
	 * a one-to-one correspondence between PDFs and history entries is
	 * required.
	 */

	/* return the resulting histogram */
//	c_print(ret);
	DEBUGASSERT(c_print(ret));
//	DEBUG("c_reduce done");
	return ret;
} /* c_reduce */

/*---------------------------------------------------------------------------*\
  Postgres V1 Function Prototypes
\*---------------------------------------------------------------------------*/

Datum u_lower(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(u_lower);

Datum u_upper(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(u_upper);

Datum u_pdf(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(u_pdf);

Datum u_cdf(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(u_cdf);

Datum u_prob(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(u_prob);

Datum u_mass(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(u_mass);

Datum u_real2hist(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(u_real2hist);

Datum u_sample(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(u_sample);

Datum u_reduce(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(u_reduce);

/*---------------------------------------------------------------------------*\
  SQL Interface (Wrappers)
\*---------------------------------------------------------------------------*/

Datum u_lower(PG_FUNCTION_ARGS)
{
	Uncertain *unc = PG_GETARG_UNCERTAIN(0);
	int32 dim = PG_GETARG_INT32(1);
	PG_RETURN_FLOAT4( c_lower(unc, dim) );
}

Datum u_upper(PG_FUNCTION_ARGS)
{
	Uncertain *unc = PG_GETARG_UNCERTAIN(0);
	int32 dim = PG_GETARG_INT32(1);
	PG_RETURN_FLOAT4( c_upper(unc, dim) );
}

Datum u_pdf(PG_FUNCTION_ARGS)
{
	Uncertain *unc = PG_GETARG_UNCERTAIN(0);
	ArrayType *arr = PG_GETARG_ARRAYTYPE_P(1);
	float4 *pnt = (float4 *) ARR_DATA_PTR(arr);
	PG_RETURN_FLOAT4( c_pdf(unc, pnt) );
}

Datum u_cdf(PG_FUNCTION_ARGS)
{
	Uncertain *unc = PG_GETARG_UNCERTAIN(0);
	ArrayType *arr = PG_GETARG_ARRAYTYPE_P(1);
	float4 *pnt = (float4 *) ARR_DATA_PTR(arr);
	PG_RETURN_FLOAT4( c_cdf(unc, pnt) );
}

Datum u_prob(PG_FUNCTION_ARGS)
{
	Uncertain *unc = PG_GETARG_UNCERTAIN(0);
	ArrayType *ar1 = PG_GETARG_ARRAYTYPE_P(1);
	ArrayType *ar2 = PG_GETARG_ARRAYTYPE_P(2);
	float4 *minv = (float4 *) ARR_DATA_PTR(ar1);
	float4 *maxv = (float4 *) ARR_DATA_PTR(ar2);
	PG_RETURN_FLOAT4( c_prob(unc, minv, maxv) );
}

Datum u_mass(PG_FUNCTION_ARGS)
{
	Uncertain *unc = PG_GETARG_UNCERTAIN(0);
	PG_RETURN_FLOAT4( c_mass(unc) );
}

Datum u_real2hist(PG_FUNCTION_ARGS)
{
	float4 val = PG_GETARG_FLOAT4(0);
	PG_RETURN_UNCERTAIN( c_real2hist(val) );
}

Datum u_sample(PG_FUNCTION_ARGS)
{
	Uncertain *unc = PG_GETARG_UNCERTAIN(0);
	int32 count = PG_GETARG_INT32(1);
	PG_RETURN_UNCERTAIN( c_sample(unc, count) );
}

Datum u_reduce(PG_FUNCTION_ARGS)
{
	Uncertain *unc = PG_GETARG_UNCERTAIN(0);
	int32 count = PG_GETARG_INT32(1);
	PG_RETURN_UNCERTAIN( c_reduce(unc, count) );
}
