/*
 * uncertain_io.c - input/output functions for uncertain data
 *
 * Relevant code examples:
 *   include/utils/array.h - multidimensional structures
 *   backend/utils/adt/numeric.c - data type modifiers
 *
 * Orion Project, http://orion.cs.purdue.edu/
 * Copyright (c) 2008 Purdue University
 */

#include "postgres.h"
#include "fmgr.h"
#include "uncertain.h"
#include "pdfs/pdf.h"
#include "pdfs/pdfs_io.h"
#include "errors.h"
#include "debug.h"
#include "nodes/pg_list.h"
#include <math.h>

// #include <math.h>

/* PDF name for implicit products. */
#define PRODUCT_NAME "prod"
#define PRODUCT_STRLEN 4
#define MARGIN_NAME "marg"

/* Strings for each floor predicate type */
static const char *FLOOR_TYPE_STR[]={"!=", "=", "<", "<=", ">", ">="};

/*---------------------------------------------------------------------------*\
  Uncertain Input
\*---------------------------------------------------------------------------*/

#define parse_err(fmt, ...) \
	ereport(ERROR, (errcode(ERRCODE_INVALID_TEXT_REPRESENTATION), \
		errmsg("%s:%d:\t" fmt, __FILE__, __LINE__, ##__VA_ARGS__)));




/* Substitute for strcasestr since solaris 10 does not have an implementation of strcasestr */
static char *mystrcasestr(char *haystack, char *needle) {
	int i, j;
	int m = strlen(needle);
	int n = strlen(haystack);
	/* Searching */
	for (j = 0; j <= (n - m); ++j) {
		for (i = 0; i < m && (toupper(needle[i]) == toupper(haystack[i + j])); ++i);
		if (i >= m)
			return needle + j;
	}
	return NULL;
}

/* Return a pointer to the first non-whitespace character in s */
static char *eat_white(char *s) {
	if(s)
		while(*s && isspace(*s))
			++s;
	return s;
}

/* Assume there are depth open parenthesis prior to s and return a pointer
 * to the parenthesis that matches the first one */
static char *match_parens(char *s, int depth) {
	if(!s) return NULL;

	goto first;													/* No increment on first go through */
	do {
		s++;
		first:
		if(*s == '(') depth++;
		else if(*s == ')') depth--;
	}while(depth > 0 && *s);

	return depth? NULL:s;
}

/* Act like strtok, but tokenize delim-delimited arguments while matching
 * parenthesis */
static char *strarg(char *start, char **end) {
	if(!start)
		start = *end;
	else
		*end = start;

	if(!start)
		return NULL;

	/* Argument ends when we hit the end of string or find a comma
	   in this level of nesting */
	while(!(*end==NULL || **end == ',')) {
		*end = strpbrk(*end, "(,");

		if(*end && **end == '(' && !(*end = match_parens(*end, 0))) /* Skip matching parens */
			parse_err("mismatched parenthesis in '%s'\n", start);
	}

	if(*end)
		*(*end)++ = 0;      /* Null-terminate and move to next arg if needed */

	return start;
}

/* Strip matched parentheses surrounding string s */
static char *strip_parens(char *s) {
	char *e;
	
	s = eat_white(s);
	e = s+strlen(s)-1;

	while(isspace(*e) && e>=s) *e-- = '\0'; /* Strip trailing whitespace */

	while(*s == '(' && *e == ')' && e >= s){
		s = eat_white(++s);
		do{ *e-- = '\0'; }while(isspace(*e) && e >= s);
	}

	return s;
}

/* Top-level parse function */
static Uncertain *_uncertain_parse(char *str, uint8 ndim, char *dimtypes);

/* Parse a projection function call, applying the specified projection
 * (marginalization) to the function's first argument
 * ERRATA: the function may not be able to determine the types of marginalized
 * dimensions, so parsing will fail incorrectly in this case.  Fixing this
 * requires changing how error checking during parsing is handled. */
static Uncertain *uncertain_proj_parse(char *str, uint8 ndim, char *dimtypes) {
	Uncertain *u = NULL;
	char *unc_str, *arg, *state;
	char inner_types[256];
	int2 margin_arg[ndim];
	int  i=0, maxdim = -1;

	memset(inner_types, 0, sizeof(inner_types));
	unc_str = strarg(str, &state);  /* Wait until schema is determined to parse */

	while((arg = strarg(NULL, &state)) && i < ndim) {
		long dim;
		char *end;
		arg = eat_white(arg);

		if(*arg == 'x') arg++;  /* Skip optional "x" */

		dim = strtol(arg, &end, 0);

		if(arg == end)
			parse_err("bad argument for %s: %s", MARGIN_NAME, arg);

		end = eat_white(end);
		if(!end || *end != '\0')
			parse_err("trailing garbage after dimension in %s: %s", MARGIN_NAME, arg);

		if(dim < 0 || dim >= 256)
			parse_err("dimension index out of range in %s: %s", MARGIN_NAME, arg);

		
		margin_arg[i] = dim;
		if(inner_types[dim] && inner_types[dim] != dimtypes[i])
			parse_err("two different types specified for the same underlying attribute %li in %s", dim, MARGIN_NAME);
		inner_types[dim] = dimtypes[i];
		maxdim = Max(dim, maxdim);

		i++;
	}

	if(arg)
		parse_err("maximum dim count expected in projection was %i, got %i", ndim, i);

	u = _uncertain_parse(unc_str, 255, inner_types);

	if(u->nldim <= maxdim)
		parse_err("attribute %i referenced in %s does not exist", maxdim, MARGIN_NAME);

	if(!u)
		parse_err("failed to parse first argument of %s: %s", MARGIN_NAME, unc_str);

	return c_margin(u, margin_arg, i);
}

/* Parse a product function call. */
static Uncertain *uncertain_prod_parse(char *str, uint8 ndim, char *dimtypes) {
	Uncertain *u = NULL;
	char *arg, *state;

	for(arg = strarg(str, &state); arg; arg = strarg(NULL, &state)) {
		Uncertain *ui = _uncertain_parse(eat_white(arg), ndim, dimtypes);

		if(!ui)
			parse_err("malformed string in product: %s", arg);

		ndim -= ui->nldim;
		dimtypes += ui->nldim;
		u = u? c_product(u, ui) : ui;
	}

	return u;
}

/* A struct to hold the parts of a clause together while it's being worked on */
typedef struct Clause {
	uint8  type;
	float4 limit;
	float4 coefs[1];
} Clause;

/* Given just one side of a constraint, update the clause */
static void uncertain_coefs_parse(Clause *clause, char *str, int ndim, float sgn) {
	enum { COEF, DIM, MULT, SGN } state = COEF;
	float4 coef=1;
	char *end;
	long dim;

	str = eat_white(str);

	if(*str == '-') {     /* Special case leading negation */
		coef = -1;
		str++;
	}

	while(*str || state != SGN) {
		if(state == COEF) {
			if(*str != 'x') {
				coef = strtof(str, &end) * coef;

				if(str == end)
					parse_err("expected a coefficient in floor: %s", str);
				str = end;
				state = MULT;
			} else
				state = DIM;
		}
		else if(state == DIM) {
			if(*str != 'x')
				parse_err("expected dim, got : %s", str);

			dim = strtol(str+1, &end, 10);
			if(str == end || dim < 0 || dim >= ndim)
				parse_err("bad dimension in floor: %s", str);

			clause->coefs[dim] += sgn * coef;
			str = end;
			state = SGN;
		}
		else if(state == SGN) {
			if(*str != '+' && *str != '-')
				parse_err("expected a plus or minus in floor, got: %s", str);
			coef = *str=='+'? 1:-1;

			str++;
			state = COEF;
		}
		else if(state == MULT) {
			if(*str != '*') { /* just a coefficient */
				clause->limit += -1 * sgn * coef;
				state = SGN;
			} else {          /* should see dim */
				str++;
				state = DIM;
			}
		}
				
		str = eat_white(str);
	}
}

/* Convert a clause to a string (for debugging) */
#if DEBUG_PRINT == 1
static char *Clause_tostr(Clause *clause, int ndim) {
	static char ret[512]="";
	char *p = ret;
	int i;

	for(i=0; i < ndim ; ++i)
		if(clause->coefs[i])
			p += sprintf(p, "%+g*x%i", clause->coefs[i], i);
		
	p += sprintf(p, " %s %g", FLOOR_TYPE_STR[clause->type], clause->limit);
		
	return ret;
}
#endif

/* Parse a clause into a Clause */
#define OP_CHARS "><=!"
static Clause *uncertain_clause_parse(Uncertain *u, char *str) {
	Clause	*clause = palloc0(sizeof(float4)*(u->nldim-1) + sizeof(Clause));
	char		*op, *left, *right;
	int			op_len, i;

	str = strip_parens(str);

	op = strpbrk(str, OP_CHARS);

	if(!op)
		parse_err("operand missing in floor string: %s", str);
	
	op_len = strspn(op, OP_CHARS);

	clause->type = 0xFF;
	for(i=0; i < sizeof(FLOOR_TYPE_STR)/sizeof(*FLOOR_TYPE_STR); ++i) {
		if(op_len == strlen(FLOOR_TYPE_STR[i])
			 && memcmp(FLOOR_TYPE_STR[i], op, op_len)==0){
			clause->type = i;
			break;
		}
	}

	if(clause->type == 0xFF)
		parse_err("bad or missing operator in floor string: %s", str);

	*op = '\0';     /* Null terminate left side */

	left = str;
	right = op + op_len;

	uncertain_coefs_parse(clause, left, u->nldim, 1);
	uncertain_coefs_parse(clause, right, u->nldim, -1);

	return clause;
}

/* Parse a disjunction as a collection of clauses, then apply it to u */
static Uncertain *uncertain_disj_parse(Uncertain *u, char *str) {
	char *end;
	List *clauses = NIL;  /* List of Clause */
	ListCell *cell;
	float4 *preds = NULL, *pred;
	uint8 *types = NULL, *type;
	int nclauses;

	str = strip_parens(str);

	while(str) {
		end = mystrcasestr(str, "or");

		if(end) {
			*end = 0;
			end += 2;
		}

		clauses = lappend(clauses, (void*) uncertain_clause_parse(u, str));
		str = end;
	}
		
	if(list_length(clauses) == 0)
		parse_err("Expecting a clause in conditions, but got %s", str);

	nclauses = list_length(clauses);

	/* Collect clauses into a single array */
	pred = preds = palloc(sizeof(float4) * (u->nldim+1) * nclauses);
	type = types = palloc(sizeof(uint8) * nclauses);

	foreach(cell, clauses) {
		Clause *clause = (Clause*) lfirst(cell);
		memcpy(pred, (float4*) clause->coefs, sizeof(float4) * u->nldim);
		pred[u->nldim] = clause->limit;

		*type++ = clause->type;
		pred += u->nldim+1;
	}

	list_free_deep(clauses);

	return c_floor(u, preds, types, nclauses);
}

/* Parse a floor by considering it one disjunction at a time.  The floor must be
 * in CNF */
static Uncertain *uncertain_floors_parse(Uncertain *u, char *str) {
	char *end;

	str = strip_parens(str);

	while(str) {
		end = mystrcasestr(str, "and");

		if(end) {
			*end = 0;
			end += 3;
		}

		u = uncertain_disj_parse(u, str);
		str = end;
	}
	
	return u;
}

/* Parse a pdf.  If the history is specified, use it, otherwise generate new
   history information. */
static Uncertain *uncertain_pdf_parse(const PDF_vtable *vtbl, uint32 *hist, char *args, uint8 ndim, char *dimtypes) {
	PDF				*pdf;
	char			rettypes[256];
	Uncertain *u = NULL;
	int				i;

	pdf = vtbl->parse(args, rettypes);
	if(!pdf)
		parse_err("bad arguments for '%s': %s", vtbl->name, args);

	if(pdf->ndim > ndim) {
		DEBUG("pdf->ndim : %d, ndim : %d", pdf->ndim, ndim);
		parse_err("too many dimensions on pdf %s(%s)", vtbl->name, args);
	}

	u = initUncertain(pdf->ndim, 1);
	for(i=0; i < pdf->ndim; ++i) {
		if(dimtypes[i] && dimtypes[i] != rettypes[i])
			parse_err("pdf %s(%s) has wrong type", vtbl->name, args);

		u->diminfo[i].dimtype = rettypes[i];
		u->diminfo[i].pdfnum = 0;
		u->diminfo[i].pdfdim = i;
	}

	pdf->vtbl = vtbl;
	pdf->history_id = hist? *hist : next_histid();
	u->pdfs[0] = pdf;

	return u;
}

/* Top-level parsing function.  Look for a function-call-esque construct,
   and then call other functions are necessary.  Will be called recursively
	 to handle products and projections.  Look for history or floor modifiers
	 hanging off PDFs as well. */
static Uncertain *_uncertain_parse(char *str, uint8 ndim, char *dimtypes) {
	char				*name = NULL, *args = NULL, *s;
	Uncertain		*u = NULL;
	const PDF_vtable	*vtbl;

	if(!str)
		return NULL;
	
	name = s = eat_white(str);
	s = strchr(s, '(');

	if(!s)
		return NULL;

	DEBUGASSERT(*s == '(');
		
	*s++ = 0;																				/* Null terminate name */
	args = eat_white(s);
	s = match_parens(s, 1);

	if(!s)
		parse_err("mismatched parentheses:  %s", args);

	DEBUGASSERT(*s == ')');

	*s++ = 0;																			/* Null terminate args */
	s = eat_white(s);

	if(strcmp(MARGIN_NAME, name) == 0) {
		u = uncertain_proj_parse(args, ndim, dimtypes);
	} else if(strcmp(PRODUCT_NAME, name) == 0) {
		u = uncertain_prod_parse(args, ndim, dimtypes);
	} else if((vtbl = pdf_name_to_vtbl(name))) {
		uint32  h;
		bool    has_history = false;

		if(*s == '#') {
			char *end;

			s++; /* skip # */
			h = strtol(s, &end, 0);

			if(end == s)
				parse_err("bad history id: %s", s);

			s = eat_white(end);
			has_history = true;
		}

		u = uncertain_pdf_parse(vtbl, has_history? &h:NULL, args, ndim, dimtypes);
	} else
		parse_err("unknown function '%s'", name);

	if(*s == 0)
		return u;
	else if(*s == '|')
		return uncertain_floors_parse(u, s+1);
	else {
		parse_err("trailing garbage after %s: %s", name, s);
		return NULL;
	}
}

/*---------------------------------------------------------------------------*\
  Uncertain Output
\*---------------------------------------------------------------------------*/

/*
 * Prints additional details about the given uncertain to the console.
 */
static void print_server(Uncertain *unc)
{
	int32 i;

	/* header data */
	UncHead *hdr = deflate(unc);
	printf("\n--------------------\n");
	printf("size  = %d\n", VARSIZE(hdr));
	printf("f_len = %d\n", hdr->f_len);

	/* dimension headers */
	printf("\nlogical dimensions (%d): type -> (pdf, dim)\n", hdr->nldim);
	for (i = 0; i < unc->nldim; i++)
	{
		printf("  %c -> (%i, %i)\n",
			unc->diminfo[i].dimtype,
			unc->diminfo[i].pdfnum,
			unc->diminfo[i].pdfdim
		);
	}

	/* pdf headers (and contents) */
	printf("\npdfs (%d)\n", hdr->npdf);
	printf("  %5s%6s%5s%15s  %s\n", "dims", "type", "size", "history", "blob");
	for(i=0; i < unc->npdf; ++i) {
		char *pdf_pars = PDF_print(unc->pdfs[i]);
		printf("  %5i     %c%5i%15d  %s(%s)\n",
			unc->pdfs[i]->ndim,
			unc->pdfs[i]->vtbl->id,
			unc->pdfs[i]->size,
			unc->pdfs[i]->history_id,
			unc->pdfs[i]->vtbl->name,
			pdf_pars
		);
		pfree(pdf_pars);
	}

	/* floors */
	if (unc->floors != NULL) {
		/* floor header */
		uint8 nconj = FLOOR_NCONJ(unc->floors);
		uint8 npred = FLOOR_NPRED(unc->floors);
		uint8 *conj = unc->floors->conjs+1;		/* Points to last pred
												 * in current conjunction */
		uint8 pred = 0;
		uint8 dim = 0;
		float4 *preds;
		uint8  *types;

		printf("\nfloors (%d conjunctions, %d predicates)\n      ", nconj, npred);

		/* preds matrix */
		types = FLOOR_TYPES(unc->floors);
		preds = FLOOR_PREDS(unc->floors);
		for(pred = 0; pred < npred; ++pred) {
			
			if(*conj == pred) {	 /* End of the current conjunction? */
				conj++;			 /* Move to next conjunction */
				printf(" &    "); /* Mark beginning of new conjunction with & */
			} else
				if(pred)		 /* Don't print '|' for first predicate */
					printf("   |  ");
				
			for(dim = 0; dim < hdr->npdim; ++dim) {
				printf("%g ", preds[(hdr->npdim+1)*pred+dim]);
			}
			printf("%s %g\n", FLOOR_TYPE_STR[*types++],
				preds[(hdr->npdim+1)*pred+dim]);
		}
	}
	printf("--------------------\n\n");

} /* print_server */

/*
 * Check to see if logical dimension map is identity
 */
static bool diminfo_is_identity(Uncertain *u) {
	int dimnum=0, pdfnum=0, pdfdim=0;

	if(u->nldim != u->npdim)
		return false;
	
	for(dimnum = 0; dimnum < u->nldim; ++dimnum) {
		if(u->diminfo[dimnum].pdfnum != pdfnum ||
		   u->diminfo[dimnum].pdfdim != pdfdim)
			return false;

		if(++pdfdim == u->pdfs[pdfnum]->ndim) {
			++pdfnum;
			pdfdim = 0;
		}
	}

	return true;
}

/* Prefixes the string in str with prefix; returns length of prefix*/
static int str_prefix(char *str, const char *prefix) {
	int pfxlen = strlen(prefix);
	memmove(str+pfxlen, str, strlen(str)+1);
	memcpy(str, prefix, pfxlen);
	return pfxlen;
}

/* prepend prefix and append suffix to str, returning count of chars added */
static int str_wrap(const char *prefix, char *str, const char *suffix) {
	int ret;

	ret = str_prefix(str, prefix);
	strcat(str, suffix);

	return ret + strlen(suffix);
}

/*
 * Returns a string version of the internal uncertain structure.
 */
static char *print_client(Uncertain *u)
{
	int32 i;
	static char str[65536] = "";
	char *p = str;


	/* Now pdfs (just print blob, since other stuff is redundant or variable) */
	for(i=0; i < u->npdf; ++i) {
		char *pdf_pars = PDF_print(u->pdfs[i]);
		p += sprintf(p,"%s%s(%s)#%i", i? ", ":"",
			u->pdfs[i]->vtbl->name, pdf_pars, u->pdfs[i]->history_id);
		pfree(pdf_pars);
	}

	if(u->npdf > 1) /* Wrap pdf description in a prod() if necessary */
		p += str_wrap("prod(", str, ")");

	/* Floors, if any */
	if(u->floors != NULL) {
		uint8 npred = FLOOR_NPRED(u->floors);
		uint8 *conj_end = u->floors->conjs+1;		/* Points to index of first pred
												 * of next conjunction */
		bool  needs_paren = *conj_end > 1 && u->floors->nconj > 1;

		uint8 pred = 0;
		uint8 dim = 0;
		float4 *preds;
		uint8  *types;

		p += sprintf(p, " | ");

		/* preds matrix */
    types = FLOOR_TYPES(u->floors);
		preds = FLOOR_PREDS(u->floors);

		if(needs_paren)
			p += sprintf(p, "(");

		for(pred = 0; pred < npred; ++pred) {
			int nonzero_coefs = 0;

			if(*conj_end == pred) {	 /* End of the current conjunction? */
				if(needs_paren) /* multiple clauses in disjunction? */
					p += sprintf(p, ")");

				needs_paren = conj_end[1] - conj_end[0] > 1 && u->floors->nconj > 1;
				conj_end++;

				p += sprintf(p, " and ");

				if(needs_paren) /* Multiple clauses in disjunction? */
					p += sprintf(p, "(");

			} else
				if(pred)		 /* Don't print '|' for first predicate */
					p += sprintf(p, " or ");

			for(dim = 0; dim < u->npdim; ++dim, preds++) {
				if(*preds != 0) {
					if(nonzero_coefs) p += sprintf(p, (*preds>0)? " + ":" - ");
					else if(*preds == -1) p += sprintf(p, "-");
					if(*preds != 1 && *preds != -1) /*take abs if plus/minus was output*/
						p += sprintf(p, "%g*", nonzero_coefs? *preds : fabs(*preds));
					p += sprintf(p, "x%i", dim);
					nonzero_coefs++;
				}
			}

			if(nonzero_coefs == 0) /* These floors should be getting filtered ... */
				p += sprintf(p, "0");

			p += sprintf(p, " %s %g", FLOOR_TYPE_STR[*types++], *preds++);
		}

		if(needs_paren)
			p += sprintf(p, ")");
	}

	if(!diminfo_is_identity(u)) {  /* Don't print project(...) for trivial maps */
		uint8 pdfbase[u->npdf];
		uncertain_pdfbase(u, pdfbase);

		/* Output dimension map in terms of physical dimension number */
		for(i=0; i < u->nldim; ++i)
			p += sprintf(p, ", x%i",
				pdfbase[u->diminfo[i].pdfnum] + u->diminfo[i].pdfdim);

		p += str_wrap(MARGIN_NAME "(", str, ")");
	}

	/* return the resulting string in the current memory context */
	p = palloc(strlen(str)+1);
	strcpy(p, str);
	return p;

} /* print_client */

/*
 * Prints the given uncertain to both the server and the client.
 */
char *c_print(Uncertain *u)
{
//	print_server(u);
	return print_client(u);
}

/*---------------------------------------------------------------------------*\
  Postgres V1 Function Prototypes
\*---------------------------------------------------------------------------*/

Datum uncertain_in(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(uncertain_in);

Datum uncertain_parse(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(uncertain_parse);

Datum uncertain_out(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(uncertain_out);

Datum uncertain_print(PG_FUNCTION_ARGS);
PG_FUNCTION_INFO_V1(uncertain_print);

/*---------------------------------------------------------------------------*\
  SQL Interface (Wrappers)
\*---------------------------------------------------------------------------*/

Datum uncertain_in(PG_FUNCTION_ARGS)
{
	char dimtypes[256];
	uint8 ndim = 255;
	char *str = PG_GETARG_CSTRING(0);

	memset(dimtypes, 0, sizeof(dimtypes));

	PG_RETURN_UNCERTAIN(_uncertain_parse(str, ndim, dimtypes));
}

Datum uncertain_parse(PG_FUNCTION_ARGS)
{
	char *dimtypes = PG_GETARG_CSTRING(0);
	uint8 ndim = strlen(dimtypes);
	char *str = PG_GETARG_CSTRING(1);
	Uncertain *u = _uncertain_parse(str, ndim, dimtypes);

	if(u->nldim != ndim)
		parse_err("pdf data has wrong number of dimensions");
	
	/* call the applicable parse function */
	PG_RETURN_UNCERTAIN(u);
}

Datum uncertain_out(PG_FUNCTION_ARGS)
{
	Uncertain *unc = PG_GETARG_UNCERTAIN(0);

	PG_RETURN_CSTRING(print_client(unc));
}

Datum uncertain_print(PG_FUNCTION_ARGS)
{
	Uncertain *unc = PG_GETARG_UNCERTAIN(0);
	PG_RETURN_CSTRING( c_print(unc) );
}
